/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef ASYNCHRONOUS_SIGNALS_SEQUENTIAL_SIGNAL_EXECUTOR_HPP
#define ASYNCHRONOUS_SIGNALS_SEQUENTIAL_SIGNAL_EXECUTOR_HPP 1

#include <asynchronous/signals/sequential_signal_executor_fwd.hpp>
#include <algorithm>
#include <list>
#include <map>
#include <string>
#include <type_traits>
#include <boost/asynchronous/diagnostics/diagnostic_item.hpp>
#include <asynchronous/signals/invoke.hpp>
#include <asynchronous/signals/type/combiner.hpp>
#include <asynchronous/signals/type/slot.hpp>

namespace asynchronous
{

  namespace signals
  {

    template < typename Result
	       , typename ... Args
	       , typename Combiner
	       , typename SignalTag
	       >
    class sequential_signal_executor <Result(Args...),Combiner,SignalTag>
    {
      using combiner_type = type::combiner_t<Result(Args...),Combiner>;
      using slot_type = type::slot_t<Result(Args...)>;

      using diagnostic_item = boost::asynchronous::diagnostic_item;
      using diagnostic_list = std::list<diagnostic_item>;
      using diagnostic_map = std::map<std::string,diagnostic_list>;

    public:
      sequential_signal_executor () = default;

      explicit sequential_signal_executor (combiner_type combiner)
	: combiner_(combiner)
      {}

    public:
      template < typename SlotIterator, typename ... T >
      auto operator () (SlotIterator begin, SlotIterator end, T&& ... t) const
      {
	return execute_sequentially(begin,end,std::forward<T>(t)...);
      }

    public:
      void clear_diagnostics () {}
      auto get_scheduler_diagnostics () const { return diagnostic_map(); }
      auto get_threadpool_diagnostics () const { return diagnostic_map(); }

    public:
      void set_combiner (combiner_type combiner) { combiner_ = combiner; }
      auto combiner () const { return combiner_; }

    private:
      template < typename SlotIterator, typename ... T >
      auto execute_sequentially (SlotIterator begin, SlotIterator end, T&& ... t) const
      {
	using result_is_void = typename std::is_void<Result>;
	return execute_sequentially(result_is_void(),begin,end,std::forward<T>(t)...);
      }

      template < typename SlotIterator, typename ... T >
      auto execute_sequentially (std::false_type, SlotIterator begin, SlotIterator end, T&& ... t) const
      {
	using std::for_each;

	auto combiner = combiner_;
	for_each(begin,end,
		 [&combiner,tuple=std::make_tuple(std::forward<T>(t)...)](slot_type slot)
		 {
		   combiner(invoke::apply(slot,tuple));
		 });
	return combiner.result();
      }

      template < typename SlotIterator, typename ... T >
      auto execute_sequentially (std::true_type, SlotIterator begin, SlotIterator end, T&& ... t) const
      {
	using std::for_each;

	auto combiner = combiner_;
	for_each(begin,end,
		 [&combiner,tuple=std::make_tuple(std::forward<T>(t)...)](slot_type slot)
		 {
		   invoke::apply(slot,tuple);
		   combiner();
		 });
	return combiner.result();
      }

    private:
      combiner_type combiner_;
    };

  } // namespace signals

} // namespace asynchronous

#endif
