/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef ASYNCHRONOUS_SIGNALS_SIGNAL_BASE_HPP
#define ASYNCHRONOUS_SIGNALS_SIGNAL_BASE_HPP 1

#include <cassert>
#include <asynchronous/signals/type/slot.hpp>
#include <observer_ptr.hpp>
#include <use_default.hpp>

namespace asynchronous
{

  namespace signals
  {

    namespace detail
    {

      template < typename Signature
		 , template < typename ... > class SlotCollection
		 >
      struct slot_container
      {
	using type = SlotCollection<type::slot_t<Signature>>;
      };

      template < typename Signature >
      struct slot_container< Signature,use_default >
      {
	using type = type::slot_container_t<Signature>;
      };

      template < typename Signature
		 , template < typename ... > class SlotCollection
		 >
      using slot_container_t = typename slot_container<Signature,SlotCollection>::type;

    } // namespace detail

    template < typename Signature
	       , template < typename ... > class SlotCollection
	       >
    class signal_base
    {
      using self = signal_base;

    public:
      using slot_type = type::slot_t<Signature>;

    private:
      using slot_container_type = detail::slot_container_t<Signature,SlotCollection>;

    protected:
      using slot_iterator = slot_iterator_t<typename slot_container_type::iterator>;
      using const_slot_iterator = type::const_slot_iterator_t<Signature>;

    public:
      using connection_type = slot_iterator;

    private:
      class tracked_connection_type
      {
	using pointer = observer_ptr<self>;

	tracked_connection_type (tracked_connection_type const&) = delete;
	auto& operator = (tracked_connection_type const&) = delete;
	auto& operator = (tracked_connection_type&&) = delete;

      public:
	tracked_connection_type (self& signal, connection_type connection)
	  : p_(make_observer(signal))
	  , connection_(connection)
	{}
	tracked_connection_type (tracked_connection_type&&) = default;
	~tracked_connection_type () { disconnect(); }

	void disconnect ()
	{
	  if (p_)
	    p_->disconnect(connection_);
	  p_.reset();
	}

      private:
	pointer p_;
	connection_type connection_;
      };

      static auto make_tracked_connection (self& signal, connection_type connection)
      { return tracked_connection_type(signal,connection); }

    public:
      auto empty () const { return num_slots() == 0; }

      auto num_slots () const
      {
	using std::distance;
	return distance(begin(),end());
      }

      template < typename Slot >
      connection_type connect (Slot slot)
      {
	// XXX<Signature>();
	// XXX<Slot>();
	slots_.insert(slots_.end(),slot);
	return --slots_.end();
      }

      template < typename Slot >
      auto tracked_connect (Slot slot)
      { return make_tracked_connection(*this,connect(slot)); }

      void disconnect (connection_type connection)
      { assert(*connection); *connection = nullptr; }

      void disconnect_all_slots ()
      { slots_.clear(); }

    protected:
      auto begin () { return make_slot_iterator(slots_.begin(),slots_.end()); }
      auto end () { return make_slot_iterator(slots_.end(),slots_.end()); }

      auto begin () const { return make_slot_iterator(slots_.begin(),slots_.end()); }
      auto end () const { return make_slot_iterator(slots_.end(),slots_.end()); }

    private:
      slot_container_type slots_;
    };

  } // namespace signals

} // namespace asynchronous

#endif
