/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef DATAFRAME_DATAFRAME_READER_HPP
#define DATAFRAME_DATAFRAME_READER_HPP 1

#include <dataframe/dataframe_reader_fwd.hpp>
#include <iosfwd>
#include <string>
#include <vector>
#include <boost/spirit/include/qi.hpp>
#include <boost/algorithm/string/find.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <dataframe/comment.hpp>
#include <dataframe/debug_parser.hpp>
#include <dataframe/delimiter.hpp>
#include <dataframe/missing_data.hpp>
#include <dataframe/skip.hpp>
#include <dataframe/skip_blank_lines.hpp>
#include <dataframe/trim_fields.hpp>
#include <dataframe/trim_whitespace.hpp>

namespace dataframe
{

  template < typename Policy >
  class dataframe_reader
  {
    using self = dataframe_reader;

  public:
    template < typename Dataframe >
    std::istream& operator () (std::istream& is, Dataframe& dataframe) const
    { return apply(is,dataframe); }

    template < typename Dataframe >
    static std::istream& apply (std::istream& is, Dataframe& dataframe)
    { return read(is,dataframe); }

    template < typename Dataframe >
    static std::istream& read (std::istream& is, Dataframe& dataframe)
    {
      using line_type = std::string;
      using parsed_line_type = std::vector<std::string>;
      using size_type = typename Dataframe::size_type;

      using boost::algorithm::find_first;

      auto comment_char = comment(is);
      auto debug_level = debug_parser(is);
      auto delimiter_character = delimiter(is);
      auto missing_data_patterns = missing_data(is);
      auto skip_lines = skip(is);
      auto skip_empty_lines = skip_blank_lines(is);
      auto trim_f = trim_fields(is);
      auto trim_ws = trim_whitespace(is);
      
      if (debug_level.value() > 1)
	std::cout << "comment:   '" << comment_char.value() << "'" << std::endl;
      if (debug_level.value() > 1)
	std::cout << "delimiter: '" << delimiter_character.value() << "'" << std::endl;
      if (debug_level.value() > 1)
	std::cout << "missing data: '" << missing_data_patterns.pattern() << "' -> '" << missing_data_patterns.replacement() << "'" << std::endl;

      line_type line;
      size_type line_number = 0;
      bool header_line = true;
      while (read_line(is,line))
	{
	  ++line_number;
	  if (debug_level)
	    std::cout << line_number << ". " << line << std::endl;
	  if (line_number > skip_lines.value())
	    {
	      if (comment_char)
		{
		  auto range = find_first(line,comment_char.pattern());
		  auto n = range.begin() - line.begin();
		  line.resize(n);
		}
	      if (!skip_empty_lines || !line.empty())
		{
		  if (debug_level.value() > 2)
		    std::cout << line_number << ". parsing line: " << line << std::endl;
		  parsed_line_type parsed_line;
		  parse_line(line,parsed_line,delimiter_character,debug_level);
		  if (header_line)
		    handle_header(dataframe,self::trim_ws(parsed_line,trim_f,trim_ws),debug_level);
		  else
		    handle_data(dataframe,self::trim_ws(parsed_line,trim_f,trim_ws),missing_data_patterns,line_number,debug_level);
		  header_line = false;
		}
	    }
	}
      return is;
    }

  private:
    template < typename Line >
    static bool read_line (std::istream& is, Line& line)
    {
      using std::getline;
      using std::make_pair;
      line.clear();
      getline(is,line);
      auto eof = is.eof() && line.empty();
      return !eof;
    }

    template < typename Range, typename ParsedLine, typename Delimiter, typename Debug >
    static bool
    parse_line (Range& range, ParsedLine& parsed_line, Delimiter delimiter, Debug debug)
    { return parse_line(range.begin(),range.end(),parsed_line,delimiter,debug); }
    
    template < typename Iterator, typename ParsedLine, typename Delimiter, typename Debug >
    static bool
    parse_line (Iterator first, Iterator last,
		ParsedLine& parsed_line, Delimiter delimiter, Debug debug)
    {
      using boost::spirit::qi::char_;
      using boost::spirit::qi::alnum;
      using boost::spirit::ascii::space;

      using boost::spirit::qi::parse;

      bool r = parse(first,last,*(char_-delimiter.value()) % char_(delimiter.value()),parsed_line);
      if (first != last)
        return false;
      return r;
    }

    template < typename Dataframe, typename ParsedLine, typename Debug >
    static void handle_header (Dataframe& dataframe, ParsedLine line, Debug debug)
    {
      if (debug)
	{
	  std::cout << "header (" << line.size() << "): ";
	  bool is_first = true;
	  for (auto x : line)
	    {
	      if (!is_first)
		std::cout << ", ";
	      std::cout << x;
	      is_first = false;
	    }
	  std::cout << std::endl;
	}
      dataframe.headers_ = line;
      dataframe.data_.resize(dataframe.headers_.size());
    }
    
    template < typename Dataframe, typename ParsedLine, typename MissingDataPatterns, typename LineNumber, typename Debug >
    static void
    handle_data (Dataframe& dataframe, ParsedLine line, MissingDataPatterns missing_data_patterns, LineNumber line_number, Debug debug)
    {
      if (line.size() == dataframe.fields())
	handle_data_enough_fields(dataframe,line,missing_data_patterns);
      else if (line.size() < dataframe.fields())
	Policy::missing_fields(dataframe.headers_,dataframe.data_,line,line_number);
      else if (line.size() > dataframe.fields())
	Policy::unexpected_fields(dataframe.headers_,dataframe.data_,line,line_number);
      else
	assert(false);
    }
    
    template < typename Dataframe, typename ParsedLine, typename MissingDataPatterns >
    static void handle_data_enough_fields (Dataframe& dataframe, ParsedLine line, MissingDataPatterns missing_data_patterns)
    {
      using size_type = typename Dataframe::size_type;
      assert(line.size() == dataframe.headers_.size());
      for (size_type j = 0; j < dataframe.headers_.size(); ++j)
	dataframe.data_[j].push_back(missing_data_patterns.transform(line[j]));
    }

    template < typename ParsedLine, typename TrimFields, typename TrimWhitespace >
    static ParsedLine trim_ws (ParsedLine line, TrimFields trim_fields, TrimWhitespace trim_whitespace)
    {
      using boost::algorithm::trim;
      if (trim_whitespace)
	{
	  for (auto& x : line)
	    trim(x);
	}
      if (trim_fields)
	{
	  while (!line.empty() && line.rbegin()->empty())
	    line.resize(line.size()-1);
	}
      return line;
    }
  };

} // namespace dataframe

#endif
