/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef DATAFRAME_DATAFRAME_FACTOR_HPP
#define DATAFRAME_DATAFRAME_FACTOR_HPP 1

#include <algorithm>
#include <functional>
#include <iosfwd>
#include <memory>
#include <stdexcept>
#include <string>
#include <type_traits>
#include <vector>
#include <boost/exception/exception.hpp>
#include <boost/operators.hpp>
#include <boost/optional.hpp>
#include <boost/throw_exception.hpp>
#include <dataframe/element_transform.hpp>

namespace dataframe
{

  template < typename Value = std::string
	     , typename Compare = std::less<Value>
	     , typename Container = std::vector<Value>
	     >
  class factor
    : boost::totally_ordered<factor<Value,Compare,Container>>
  {
    using self = factor;
    using container_type = Container;
    using iterator = typename container_type::iterator;
    using const_iterator = typename container_type::const_iterator;
    using pointer = std::shared_ptr<container_type>;
    using optional_iterator = boost::optional<const_iterator>;

    struct factor_value_not_found
      : std::logic_error
      , virtual boost::exception
    {
      factor_value_not_found ()
	  : std::logic_error("factor value not found")
      {}
    };

  public:
    using value_type = typename container_type::value_type;
    using size_type = typename container_type::size_type;
    using value_compare = Compare;

  public:
    factor () = default;
    factor (value_compare compare) : compare_(compare) {}

    template < typename C, typename Ct >
    factor (factor<Value,C,Ct> other)
    {
      if (!other.empty())
	{
	  insert(other.begin(),other.end());
	  value(other.value());
	}
    }

    template < typename C, typename Ct >
    factor (factor<Value,C,Ct> other, Compare compare)
      : compare_(compare)
    {
      if (!other.empty())
	{
	  insert(other.begin(),other.end());
	  value(other.value());
	}
    }

    template < typename C, typename Ct >
    auto& operator = (factor<Value,C,Ct> other)
    { return operator=(self(std::move(other))); }

    void swap (self& other)
    {
      using std::swap;
      swap(compare_,other.compare_);
      swap(p_,other.p_);
      swap(i_,other.i_);
    }

    friend void swap (self& left, self& right) { left.swap(right); }

  public:
    auto value_comp () const { return compare_; }

  public:
    operator value_type const& () const { return value(); }
    auto& value () const { return i_ ? **i_ : *begin(); }
    auto& value (value_type value)
    {
      i_ = find(value);
      if (*i_ == end())
	BOOST_THROW_EXCEPTION(factor_value_not_found());
      return *this;
    }
    auto index () const { using std::distance; return i_ ? distance(begin(),*i_) : 0; }

  public:
    auto empty () const { return container().empty(); }
    auto size () const { return container().size(); }

    auto begin () { return container().begin(); }
    auto end () { return container().end(); }

    auto begin () const { return container().begin(); }
    auto end () const { return container().end(); }

  public:
    template < typename Iterator >
    auto& insert (Iterator begin, Iterator end)
    {
      using element_transform = element_transform<decltype(*begin),value_type>;
      std::for_each(begin,end,[this](auto x){ insert(element_transform::apply(x)); });
      return *this;
    }

    auto insert (value_type value)
    {
      auto i = std::lower_bound(begin(),end(),value);
      if (i == end() || *i != value)
	i = container().insert(i,value);
      return i;
    }

  public:
    auto find (value_type value) const
    { return std::find(begin(),end(),value); }

  public:
    auto& merge () { return *this; }
    auto& merge (self& other)
    {
      auto this_value = value();
      auto other_value = other.value();
      insert(other.begin(),other.end());
      other.p_ = p_;
      value(this_value);
      other.value(other_value);
      return *this;
    }
    template < typename ... Factors >
    auto& merge (self& other, Factors& ... others)
    {
      merge(other);
      return merge(others...);
    }

  private:
    auto& container () { return static_cast<container_type&>(*p_); }
    auto& container () const { return static_cast<container_type const&>(*p_); }

  private:
    static auto initialize () { return std::make_shared<container_type>(); }

  private:
    value_compare compare_;
    pointer p_ = initialize();
    optional_iterator i_;
  };

  template < typename Value, typename Compare, typename Container, typename ... Factors >
  auto& merge (factor<Value,Compare,Container>& factor, Factors& ... factors)
  { return factor.merge(factors...); }

  template < typename Value, typename Compare, typename Container >
  auto operator == (factor<Value,Compare,Container> left,
		    factor<Value,Compare,Container> right)
  {
    return left.size() == right.size()
      && std::equal(left.begin(),left.end(),right.begin(),
		    [compare=left.value_comp()](auto left, auto right)
		    { return !compare(left,right) && !compare(right,left); })
      && left.value() == right.value();
  }

  template < typename Value, typename Compare, typename Container >
  auto operator < (factor<Value,Compare,Container> left,
		   factor<Value,Compare,Container> right)
  {
    return left.size() == right.size()
      && std::equal(left.begin(),left.end(),right.begin(),
		    [compare=left.value_comp()](auto left, auto right)
		    { return !compare(left,right) && !compare(right,left); })
      && left.value_comp()(left.value(),right.value());
  }

  template < typename Value, typename Compare, typename Container >
  auto& operator << (std::ostream& os, factor<Value,Compare,Container> f)
  {
    if (!f.empty())
      os << f.value();
    os << "[";
    bool print_comma = false;
    for (auto x : f)
      {
	if (print_comma)
	  os << ',';
	os << x;
	print_comma = true;
      }
    os << "]";
    return os;
  }

  template < typename T, typename Range >
  auto make_factor (Range&& range)
  {
    using factor_type = factor<T>;
    factor_type factor;
    factor.insert(range.begin(),range.end());
    return factor;
  }

  template < typename Range >
  auto make_factor (Range&& range)
  {
    using value_type = std::decay_t<decltype(*range.begin())>;
    return make_factor<value_type>(std::forward<Range>(range));
  }

  template < typename T, typename Range, typename Value >
  auto make_factor (Range&& range, Value&& value)
  {
    auto factor = make_factor<T>(std::forward<Range>(range));
    factor.value(std::forward<Value>(value));
    return factor;
  }

  template < typename Range, typename Value >
  auto make_factor (Range&& range, Value&& value)
  {
    auto factor = make_factor(std::forward<Range>(range));
    factor.value(std::forward<Value>(value));
    return factor;
  }

  template < typename Compare, typename V, typename C, typename Ct >
  auto reorder_factor (factor<V,C,Ct> other)
  {
    using factor_type = factor<V,Compare,Ct>;
    return factor_type(other);
  }

  template < typename V, typename C, typename Ct, typename Compare >
  auto reorder_factor (factor<V,C,Ct> other, Compare compare)
  {
    using factor_type = factor<V,Compare,Ct>;
    return factor_type(other,compare);
  }

} // namespace dataframe

#endif
