/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef DATAFRAME_MANIPULATOR_STORAGE_BASE_HPP
#define DATAFRAME_MANIPULATOR_STORAGE_BASE_HPP 1

#include <iostream>
#include <type_traits>

namespace dataframe
{

  namespace detail
  {

    template < typename Derived, typename T >
    struct explicit_conversion_to_bool
    {
      explicit operator bool () const { return derived().value(); }

    private:
      auto& derived () { return static_cast<Derived&>(*this); }
      auto& derived () const { return static_cast<Derived const&>(*this); }
    };

    template < typename Derived >
    struct explicit_conversion_to_bool< Derived,bool >
    {};

  } // namespace detail

  template < typename Derived
	     , typename T
	     , typename = std::integral_constant<bool,(sizeof(T) > sizeof(long))>
             >
  class manipulator_storage_policy_base;

  template < typename Derived, typename T >
  struct manipulator_storage_policy_base< Derived,T,std::false_type >
  {
  protected:
    static auto& reference (std::ios_base& ios_base){ return ios_base.iword(index()); }

  private:
    static int index ()
    {
      static int index = std::ios_base::xalloc();
      return index;
    }
  };

  template < typename Derived, typename T >
  struct manipulator_storage_policy_base< Derived,T,std::true_type >
  {
    using self = manipulator_storage_policy_base;

  protected:
    static auto& reference (std::ios_base& ios_base) { return *construct(ios_base); }

  private:
    static int index ()
    {
      static int index = std::ios_base::xalloc();
      return index;
    }

    static auto construct (std::ios_base& ios_base)
    { 
      auto& p = pointer_reference(ios_base);
      if (!p)
	{
	  p = new T();
	  ios_base.register_callback(&self::destruct,index());
	}
      return static_cast<T*>(p);
    }

    static void destruct (std::ios_base::event event, std::ios_base& ios_base, int index)
    {
      if (event == std::ios_base::erase_event)
	{
	  auto& p = pointer_reference(ios_base,index);
	  if (p)
	    delete(static_cast<T*>(p));
	  p = nullptr;
	}
    }

    static auto& pointer_reference (std::ios_base& ios_base)
    { return pointer_reference(ios_base,index()); }

    static auto& pointer_reference (std::ios_base& ios_base, int index)
    { return ios_base.pword(index); }
  };

  template < typename Derived, typename T >
  class default_manipulator_storage_policy
    : manipulator_storage_policy_base<Derived,T>
  {
    using base = manipulator_storage_policy_base<Derived,T>;

  public:
    static T initialize () { return T(); }
    static T initialize (std::ios_base& ios_base) { return base::reference(ios_base); }

    static void assign (std::ios_base& ios_base, T t) { base::reference(ios_base) = t; }
  };

  template < typename Derived
	     , typename T = bool
	     , typename Policy = default_manipulator_storage_policy<Derived,T>
	     >
  class manipulator_storage_base
    : public detail::explicit_conversion_to_bool<Derived,T>
  {
    using self = manipulator_storage_base;
    
  public:
    manipulator_storage_base () = default;
    explicit manipulator_storage_base (T t) : t_(t) {}
    explicit manipulator_storage_base (std::ios_base& ios_base)
      : t_(Policy::initialize(ios_base))
    {}

    explicit operator T () const { return value(); }
    T value () const { return t_; }

    friend std::istream& operator >> (std::istream& is, self const& b)
    { Policy::assign(is,b.t_); return is; }

    friend std::ostream& operator << (std::ostream& os, self const& b)
    { Policy::assign(os,b.t_); return os; }

  private:
    Derived& derived () { return static_cast<Derived&>(*this); }
    Derived const& derived () const { return static_cast<Derived const&>(*this); }

  private:
    T t_ = Policy::initialize();
  };

} // namespace dataframe

#endif
