/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef DATAFRAME_RANGE_PROXY_HPP
#define DATAFRAME_RANGE_PROXY_HPP 1

#include <type_traits>
#include <dataframe/element_transform_iterator.hpp>
#include <observer_ptr.hpp>

namespace dataframe
{

  namespace detail
  {

    template < typename Range, typename Value >
    class range_proxy_base
    {
    public:
      using size_type = typename Range::size_type;
      using value_type = typename Range::value_type;

    protected:
      using iterator_transform
      = element_transform_iterator<value_type,Value>;

    private:
      using range_iterator = typename Range::iterator;
      using const_range_iterator = typename Range::const_iterator;

    public:
      using iterator = typename iterator_transform::template result_t<range_iterator>;
      using const_iterator = typename iterator_transform::template result_t<const_range_iterator>;
    };
    
  } // namespace detail
  
  template < typename Range
	     , typename Value = typename Range::value_type
	     >
  class range_proxy
    : public detail::range_proxy_base<Range,Value>
  {
    using self = range_proxy;
    using base = detail::range_proxy_base<Range,Value>;
    using pointer = observer_ptr<Range>;

 public:
    range_proxy () = default;
    
    explicit range_proxy (Range& range)
      : p_(make_observer(range))
    {}

    self& operator = (Range const& range)
    {
      value() = range;
      return *this;
    }

    self& operator = (Range&& range)
    {
      value() = std::move(range);
      return *this;
    }

    bool empty () { return !p_ || value().empty(); }
    auto size () { return p_ ? value().size() : 0; }

    auto begin () { return base::iterator_transform::apply(p_->begin()); }
    auto end () { return base::iterator_transform::apply(p_->end()); }

    auto begin () const { return base::iterator_transform::apply(p_->begin()); }
    auto end () const { return base::iterator_transform::apply(p_->end()); }

    Range& value () const { return range(); }
    operator Range& () const { return range(); }

  protected:
    Range& range () const { return *p_; }

  private:
    pointer p_;
  };

  template < typename Range, typename Value >
  class range_proxy< Range const,Value >
    : public detail::range_proxy_base<Range const,Value>
  {
    using base = detail::range_proxy_base<Range const,Value>;
    using pointer = observer_ptr<Range const>;

  public:
    range_proxy () = default;
    
    explicit range_proxy (Range const& range)
      : p_(make_observer(range))
    {}

    bool empty () { return !p_ || value().empty(); }
    auto size () { return p_ ? value().size() : 0; }

    auto begin () const { return base::iterator_transform::apply(p_->begin()); }
    auto end () const { return base::iterator_transform::apply(p_->end()); }

    Range const& value () const { return range(); }
    operator Range const& () const { return range(); }

  protected:
    Range const& range () const { return *p_; }

  private:
    pointer p_;
  };

  template < typename Value, typename Range >
  auto make_range_proxy (Range& range)
  {
    using range_proxy_type = range_proxy<Range,Value>;
    return range_proxy_type(range);
  }

  template < typename Value, typename Range >
  auto make_range_proxy (Range const& range)
  {
    using range_proxy_type = range_proxy<Range const,Value>;
    return range_proxy_type(range);
  }

} // namespace dataframe

#endif
