/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef DEBUGGING_STACK_TRACE_HPP
#define DEBUGGING_STACK_TRACE_HPP 1

#include <algorithm>
#include <array>
#include <iosfwd>
#include <memory>
#include <string>
#include <vector>
#include <boost/core/demangle.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string.hpp>

#include <execinfo.h>		// XXX - stack traces

namespace debugging
{

  /// \brief Construct a stack trace
  /// \tparam Size The maximum number of stack frames to trace
  ///
  /// This class constructs a stack trace from the point at which a \c
  /// stack_trace object is constructed.  Necessarily, the first few
  /// stack frames correspond to member functions of this class and
  /// should likely be ignored.
  ///
  /// \bug The current implementation is built upon the \c backtrace()
  /// and \c backtrace_symbols() functions which may not be portable.
  ///
  template < std::size_t Size = 1024 >
  class stack_trace
  {
    using array_type = std::array<void*,Size>;

    struct stack_frame
    {
      stack_frame (std::string level, std::string name, std::string function, std::string address, std::string direction, std::string offset)
	: level_(level)
	, name_(name)
	, function_(function)
	, address_(address)
	, direction_(direction)
	, offset_(offset)
      {}

      auto demangled () const { using boost::core::demangle; return demangle(function_.c_str()); }

      friend std::ostream& operator << (std::ostream& os, stack_frame const& sf)
      { return os << sf.level_ << ": " << sf.demangled() << " [ " << sf.address_ << " " << sf.direction_ << " " << sf.offset_ << " ]"; }
	
    private:
      std::string level_;
      std::string name_;
      std::string function_;
      std::string address_;
      std::string direction_;
      std::string offset_;
    };

  public:
    stack_trace () { initialize(); }

    friend std::ostream& operator << (std::ostream& os, stack_trace const& st)
    {
      using std::for_each;
      os << "stack frame:" << std::endl;
      for_each(st.stack_frames_.begin(),st.stack_frames_.end(),
	       [&os](auto stack_frame){ os << stack_frame << std::endl; });
      return os;
    }

  private:
    void initialize ()
    {
      using unique_strings_ptr = std::unique_ptr<char*[],void(*)(char**)>;

      using boost::algorithm::is_any_of;
      using boost::algorithm::split;
      using boost::algorithm::token_compress_on;

      array_type stack_frames;
      auto n = backtrace(stack_frames.data(),stack_frames.size());
      unique_strings_ptr strs
	(backtrace_symbols(stack_frames.data(),n),[](char** x){ free(x); });
      for (int i = 0; i < n; ++i)
	{
	  using strings_type = std::vector<std::string>;
	  strings_type strings;
	  split(strings,strs[i],is_any_of(" "),token_compress_on);
	  assert(strings.size() == 6);
	  stack_frames_.push_back
	    (stack_frame(strings[0],strings[1],strings[3],strings[2],strings[4],strings[5]));
	}
    }

  private:
    std::vector<stack_frame> stack_frames_;
  };

} // namespace debugging

#endif
