/*
 * Copyright (c) 2017 Gregory Penn.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#ifndef GDAL_GEOGRAPHIC_SRS_HPP
#define GDAL_GEOGRAPHIC_SRS_HPP

#include <ogrsf_frmts.h>
#include <ogr_spatialref.h>
#include <boost/operators.hpp>
#include <ostream>
#include <gdal/smartptr.hpp>
#include <gdal/wgs84.hpp>

namespace gdal
{

  class geographic_srs
    : boost::equality_comparable<geographic_srs>
  {
    using self = geographic_srs;

  public:
    geographic_srs () = default;

    geographic_srs (smartptr<OGRSpatialReference> p): p_(p) {}

    geographic_srs (const char* wkt)
      : p_(from_wkt(wkt))
      {}

  private:
    template < typename DeleterTag >
    geographic_srs (OGRSpatialReference* p, DeleterTag tag)
      : geographic_srs(p, tag)
    {}

  public:
    auto get () const { return p_.get(); }

    explicit operator bool () { return static_cast<bool>(p_); }
    bool     operator ==   (self const& other) const { return p_->IsSame(other.p_.get()); }

    auto wkt (bool format_pretty = false) const
      {
        char* p;
        format_pretty ? p_->exportToPrettyWkt(&p) :  p_->exportToWkt(&p);
        std::shared_ptr<char> wkt(p, [](auto p) { CPLFree(p); });
        return(wkt);
      }


  private:
    smartptr<OGRSpatialReference> from_wkt (const char* wkt)
    {
      auto p_raw = static_cast<OGRSpatialReference*>(OSRNewSpatialReference(wkt));
      return smartptr<OGRSpatialReference>
        (p_raw
         , [p_raw] () { OGRSpatialReference::DestroySpatialReference(p_raw); }    // deleter
         , [p_raw] () { p_raw->Reference(); }                // increment
         , [p_raw] () { p_raw->Dereference(); }              // decrement
         , [p_raw] () { return p_raw->GetReferenceCount(); } // use_count
          );
    }

  private:
    smartptr<OGRSpatialReference> p_;
  };

  std::ostream& operator << (std::ostream& os, geographic_srs const& srs)
  {
    char* p;
    srs.get()->exportToPrettyWkt(&p);
    std::shared_ptr<char> wkt(p, [](auto p) { CPLFree(p); });

    return os << p;
  }

  auto make_geographic_srs (const int epsg)
  {
    using ptr = smartptr<OGRSpatialReference>;
    OGRSpatialReference* p_raw;
    p_raw = new OGRSpatialReference;
    p_raw->importFromEPSG(epsg);
    ptr p(p_raw
          , [p_raw] () { OGRSpatialReference::DestroySpatialReference(p_raw); }
          , [p_raw] () { p_raw->Reference(); }
          , [p_raw] () { p_raw->Dereference(); }
          , [p_raw] () { return p_raw->GetReferenceCount(); }
      );

    return geographic_srs(p);
  }

  geographic_srs make_wgs84 ()
  {
    using ptr = smartptr<OGRSpatialReference>;
    auto p_raw = OGRSpatialReference::GetWGS84SRS()->Clone();
    ptr p(p_raw
          , [p_raw] () { OGRSpatialReference::DestroySpatialReference(p_raw); }
          , [p_raw] () { p_raw->Reference(); }
          , [p_raw] () { p_raw->Dereference(); }
          , [p_raw] () { return p_raw->GetReferenceCount(); }
      );

    return geographic_srs(p);
  }


} // namespace gdal

#endif
