/*
 * Copyright (c) 2017 Gregory Penn.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#ifndef GDAL_POINT_HPP
#define GDAL_POINT_HPP

#include <array>
#include <ogr_geometry.h>
#include <iostream>
#include <ogr_spatialref.h>
#include <boost/operators.hpp>
#include <gdal/geographic_srs.hpp>
#include <gdal/grid_srs.hpp>
#include <gdal/wgs84.hpp>
#include <gdal/geometry_handle.hpp>
#include <mpl/enable_constructor.hpp>
#include <gdal/OGRCoordinateTransformation_handle.hpp>
// #include <gdal/geographic_point_transformation.hpp>

namespace gdal
{

  // SpatialReference must be either
  // geographic_srs or grid_srs
  template < typename SpatialReference  >
  class point;

  template <> class point < grid_srs >;

  template <>
  class point < geographic_srs >
    : boost::equality_comparable<point<geographic_srs>>
  {
    using self = point<geographic_srs>;

  public:
    point () = default;

    template < typename SRSArg >
      point (double x, double y, SRSArg arg)
      : point_(x, y), srs_(arg), p_(initialize(x,y)) {}

  public:
    auto x () const { return point_.getX(); }
    auto y () const { return point_.getY(); }
    std::array<double, 2> coords () { return { x(), y() }; }

    auto within (OGRGeometry* pGeom) { return point_.Within(pGeom); }
    auto within (geometry_handle h)  { return point_.Within(h.gdal_ptr().get()); }
    geographic_srs srs () const { return srs_; }

    auto p () const { return std::shared_ptr<OGRPoint>(p_); }

    // Returns true iff coordinates and srs match, else false.
    bool operator == (self other) const
    { return (point_.Equals(&other.point_) && (srs_ == other.srs_)); }

    auto transform_srs (geographic_srs const& target_srs) const
    {
      if (srs_ == target_srs) return *this;
      else
      {
        OGRCoordinateTransformation_handle transformer(srs_, target_srs);
        auto coords = transformer.transform(x(), y());

        return point<geographic_srs>(coords.first, coords.second, target_srs);
      }
    }

  private:
    using sharedptr_type = std::shared_ptr<OGRPoint>;
    sharedptr_type initialize (double x, double y)
    {
      auto pt = std::make_shared<OGRPoint>(x, y);
      if (srs_ && (srs_.get()->Validate() == OGRERR_NONE))
	pt->assignSpatialReference(srs_.get());
      return pt;
    }

  private:
    OGRPoint point_;
    geographic_srs srs_;
    sharedptr_type p_;
  };


  template <>
  class point < grid_srs >
    : boost::equality_comparable<point<grid_srs>>
  {
    using self = point<grid_srs>;

  public:
    point () = default;

    point (int x, int y): x_(x), y_(y) {}

    point (int x, int y, grid_srs grid)
      { initialize(x, y, grid); }

    template < typename Coords >
    point (Coords coords, grid_srs grid)
    { initialize(coords[0], coords[1], grid); }

    point (std::pair<int,int> coords, grid_srs grid)
    { initialize(coords.first, coords.second, grid); }

    bool operator == (self const& other) const
    {
      return (other.x_  == x_)
          && (other.y_  == y_)
          && (other.grid_ == grid_ );
    }

  public:
    auto x () const { return x_; }
    auto y () const { return y_; }
    auto grid () const { return grid_; }

  private:
    void initialize (int x, int y, grid_srs grid)
    {
      assert(x >= 0); assert(x <= grid.xmax());
      assert(y >= 0); assert(y <= grid.ymax());
      x_ = x; y_ = y; grid_ = grid;
    }

  private:
    grid_srs grid_;
    int x_;
    int y_;
  };

  template < typename SpatialReference >
  std::ostream& operator << (std::ostream& os, point<SpatialReference> const& point)
  { return os << "Cordinates (x, y): (" << point.x() << ", " << point.y() << ")"; }

} // namespace gdal

#endif
