/*
 * Copyright (c) 2017 Gregory Penn.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#ifndef GDAL_RASTERBAND_HPP
#define GDAL_RASTERBAND_HPP

#include <gdal.h>
#include "gdal/geographic_srs.hpp"
#include "gdal/grid_srs.hpp"
#include "gdal/point.hpp"

namespace gdal
{
        template < typename ValueType >
        auto gdal_datatype (ValueType)
        {
                auto size = sizeof(ValueType);

                std::string s;
                GDALDataType e;
                if (std::is_floating_point<ValueType>::value)
                { e = size==8 ? GDT_Float64 : GDT_Float32; }
                else
                { e = GDT_Int32; }

                return e;
        }

        template < typename ValueType = double, typename IntegralType = int >
        class rasterband
        {


        public:
                rasterband () = default;

                rasterband (smartptr<GDALRasterBand> p
                                   , geographic_srs srs
                                   , grid_srs grid)
                        : p_(p)
                        , srs_(srs)
                        , grid_(grid)
                        {}

                bool updatable ()
                        {
                                auto access = p_->GetAccess();
                                return access == GA_Update;
                        }


                template < typename Point >
                auto read (Point point) const
                        { return apply(point); }

                auto read (int x, int y) const
                        { return read(point<grid_srs>(x, y, grid_)); }

                auto apply (point<geographic_srs> const& pt) const
                        { return apply(pt.transform_srs(srs_).grid_point()); }


                auto apply (point<grid_srs> const& pt
                           , int buffer_xsize = 1
                           , int buffer_ysize = 1) const
                        {
                                ValueType result;
                                auto outcome = p_->RasterIO(GF_Read
                                                           , pt.x()
                                                           , pt.y()
                                                           , 1
                                                           , 1
                                                           , &result
                                                           , 1
                                                           , 1
                                                           , gdal_datatype(ValueType())
                                                           , sizeof(ValueType)
                                                           , sizeof(ValueType) * buffer_xsize
                                                           , nullptr);
                                if(outcome == CE_Failure) assert(false && "Unable to read rasterband");
                                return result;
                        }

                auto write (int x, int y, ValueType value, int buffer_xsize = 1, int buffer_ysize = 1)
                        { return write(point<grid_srs>(x, y, grid_), value, buffer_xsize, buffer_ysize); }

                auto write (point<grid_srs> pt, ValueType value, int buffer_xsize = 1, int buffer_ysize = 1)
                        {
                                p_->RasterIO(GF_Write
                                             , pt.x()
                                             , pt.y()
                                             , 1
                                             , 1
                                             , &value
                                             , buffer_xsize
                                             , buffer_ysize
                                             , gdal_datatype(ValueType())
                                             , sizeof(ValueType)
                                             , sizeof(ValueType) * buffer_xsize);
                        }

                auto bandp () { return p_; }
                auto srs   () { return srs_; }
                auto grid  () { return grid_; }

//        private:
        public:
                using coeff_array = std::array<double, 6>;
        private:
                smartptr<GDALRasterBand> p_;
                geographic_srs           srs_;
                grid_srs                grid_;
        };

} // namespace gdal
#endif
