/*
 * Copyright (c) 2017 Gregory Penn.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#ifndef GDAL_RASTERDATASET_HPP
#define GDAL_RASTERDATASET_HPP

#include <gdal/opengdal_readonly.hpp>
#include <gdal/opengdal_readwrite.hpp>
#include <gdal/closegdal.hpp>
#include <gdal/geographic_srs.hpp>
#include <gdal/grid_srs.hpp>
#include <gdal/rasterband.hpp>
#include <gdal_priv.h>
#include <gdal.h>

namespace gdal
{

  template < typename ValueType = double >
  class rasterdataset
  {
    using grid_type = grid_srs;
    using self = rasterdataset;

  public:

    rasterdataset () = default;

    rasterdataset (smartptr<GDALDataset> p): p_(p) {}

    rasterdataset (const char* datasource, bool allow_updates = true)
      : rasterdataset(rasterdataset::create_smartptr(datasource, allow_updates))
      {}

    template < typename array_type >
    rasterdataset (std::string name
            , int xsize
            , int ysize
            , int nbands
            , geographic_srs srs
            , array_type &affine_transform_coeffs)
      : rasterdataset(rasterdataset::create_smartptr(name, xsize, ysize, nbands, srs, affine_transform_coeffs))
      {}

  public:

    auto srs  () const { return geographic_srs(p_->GetProjectionRef()); }
    auto grid () const
      {
        return grid_srs(p_->GetRasterXSize()
                        , p_->GetRasterYSize());
      }

    auto p () const { return p_; }

    bool updatable ()
      {
        auto access = p_->GetAccess();
        return access == GA_Update;
      }

    auto band (int id) { return gdal::rasterband<ValueType>(bandptr(id), srs(), grid()); }

    auto use_count () const { return p_.use_count(); }

  private:

    auto bandptr (int id)
      {
        auto p_raw = p_->GetRasterBand(id);
        return smartptr<GDALRasterBand>(p_raw
                                        , p_
                                        , [] () {}
                                        , [] () {}
                                        , [] () {}
                                        , [] () { return 0; }
          );
      }

    static smartptr<GDALDataset> create_smartptr (const char* datasource, bool allow_updates)
      {
        GDALDataset* p_raw = allow_updates ? opengdal_readwrite(datasource)
          : opengdal_readonly(datasource);

        auto p = smartptr<GDALDataset>(p_raw
                                       , [p_raw] () { GDALClose(p_raw); }
                                       , [p_raw] () { p_raw->Reference(); }
                                       , [p_raw] () { p_raw->Dereference(); }
                                       , [p_raw] () { return p_raw->GetRefCount(); });
        return p;
      }

    template < typename array_type >
    static smartptr<GDALDataset> create_smartptr (std::string name
                                                  , int xsize
                                                  , int ysize
                                                  , int nbands
                                                  , geographic_srs srs
                                                  , array_type &affine_transform_coeffs)

      {
        gdalregistration::reg();
        GDALDriver *poDriver;
        GDALDataset *p_raw;
        poDriver = GetGDALDriverManager()->GetDriverByName("GTiff");
        char tiled[] = { "TILED=YES" };
        char* options[] = { tiled, nullptr };
        if (!poDriver) { printf("%s driver not found.", "GTiff"); assert(FALSE); }
        p_raw = poDriver->Create(name.data()
                                 , xsize
                                 , ysize
                                 , nbands
                                 , GDT_Float64
                                 , options);

        auto p = smartptr<GDALDataset>(p_raw
                                       , [p_raw] () { GDALClose(p_raw); }
                                       , [p_raw] () { p_raw->Reference(); }
                                       , [p_raw] () { p_raw->Dereference(); }
                                       , [p_raw] () { return p_raw->GetRefCount(); });

        p->SetProjection(srs.wkt().get());
        auto t = affine_transform_coeffs.data();
        p->SetGeoTransform( t );

        return p;
      }


  private:
    smartptr<GDALDataset> p_;

  };

} // namespace gdal

#endif
