/*
 * Copyright (c) 2017 Gregory Penn.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#ifndef GDAL_REF_MANAGER_HPP
#define GDAL_REF_MANAGER_HPP

#include <functional>
#include <utility>
#include <memory>

namespace gdal
{

  class ref_manager
  {
    using self = ref_manager;

  public:
    ref_manager () = default;

    ref_manager (std::function<void()> deleter
                 , std::function<void()> increment
                 , std::function<void()> decrement
                 , std::function<int ()> use_count)
      : deleter_(deleter)
      , increment_(increment)
      , decrement_(decrement)
      , use_count_(use_count)
      {}

    ref_manager (self const& other)
      : deleter_(other.deleter_)
      , increment_(other.increment_)
      , decrement_(other.decrement_)
      , use_count_(other.use_count_)
    { increment_(); }

    ref_manager (self&& other)
      : deleter_  (std::move(other.deleter_))
      , increment_(std::move(other.increment_))
      , decrement_(std::move(other.decrement_))
      , use_count_(std::move(other.use_count_))
    {}

    auto& operator = (self const& other)
    {
      self temp(other);
      swap(*this, temp);
      return *this;
    }

    auto& operator = (self&& other)
    {
      self temp(std::move(other));
      swap(*this, temp);
      return *this;
    }

    ~ref_manager ()
      {
        decrement_();
        if (use_count_() == 0) deleter_();
      }

    friend void swap (self& left, self& right)
      {
        using std::swap;
        swap(left.increment_, right.increment_);
        swap(left.decrement_, right.decrement_);
      }

    int use_count () const { return use_count_(); }

  private:
    std::function<void()> increment_ = [] () {};
    std::function<void()> decrement_ = [] () {};
    std::function<int ()> use_count_ = [] () { return 0; };
    std::function<void()> deleter_   = [] () {};
  };

} // namespace gdal

#endif
