/*
 * Copyright (c) 2017 Gregory Penn.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#ifndef GDAL_VECTORDATASET_HPP
#define GDAL_VECTORDATASET_HPP

#include <ogrsf_frmts.h>
#include <gdal/opengdal_readonly.hpp>
#include <gdal/opengdal_readwrite.hpp>
#include <gdal/closegdal.hpp>
#include <gdal/geographic_srs.hpp>
#include <gdal/vectorlayer.hpp>

namespace gdal
{

  class vectordataset
  {
    using self = vectordataset;
    using layer_ptr = smartptr<OGRLayer>;

  public:
    vectordataset () = default;

    vectordataset (smartptr<GDALDataset> p): p_(p) {}

    vectordataset (const char* datasource, bool allow_updates = true)
      : vectordataset(vectordataset::create_smartptr(datasource, allow_updates))
      {}

  public:
    auto nlayers () const { return p_->GetLayerCount(); }

  public:
    auto get () { return p_.get(); }

    auto layer (int index)
    {
      return vectorlayer(
        layer_ptr(p_->GetLayer(index)
                  , p_       // owner
                  , [] () {} // deleter
                  , [] () {} // increment
                  , [] () {} // decrement
                  , [] () { return 0; } // use_count
                  )
				);
    }

    auto layer (const char* name)
    {
      return vectorlayer(
                layer_ptr(
                  p_->GetLayerByName(name)
                  , p_       // owner
                  , [] () {} // deleter
                  , [] () {} // increment
                  , [] () {} // decrement
                  , [] () { return 0; } // use_count
                  )
				);
    }

    auto layer_from_sql (const char  *pszSQLCommand
			 , OGRGeometry *poSpatialFilter = NULL
			 , const char  *pszDialect = NULL)
    {
      OGRLayer* p_raw(p_->ExecuteSQL(pszSQLCommand, poSpatialFilter, pszDialect));

      return vectorlayer(
        layer_ptr(
          p_raw
          , p_ // owner
          , [this, p_raw] () { p_->ReleaseResultSet(p_raw); } // deleter
          , [p_raw] () { p_raw->Reference(); }
          , [p_raw] () { p_raw->Dereference(); }
          , [p_raw] () { return p_raw->GetRefCount(); }
          )
        );
    }

  private:
    static smartptr<GDALDataset> create_smartptr (const char* datasource, bool allow_updates)
      {
        GDALDataset* p_raw = allow_updates ? opengdal_readwrite(datasource)
          : opengdal_readonly(datasource);

        auto p = smartptr<GDALDataset>(p_raw
                                       , [p_raw] () { GDALClose(p_raw); }
                                       , [p_raw] () { p_raw->Reference(); }
                                       , [p_raw] () { p_raw->Dereference(); }
                                       , [p_raw] () { return p_raw->GetRefCount(); });
        return p;
      }

  private:
    smartptr<GDALDataset> p_;
  };

} // namespace gdal

#endif
