/*
 * Copyright (c) 2017 Gregory Penn.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#ifndef GDAL_VECTORLAYER_HPP
#define GDAL_VECTORLAYER_HPP

#include <ogrsf_frmts.h>
#include <gdal/geographic_srs.hpp>
#include <gdal/smartptr.hpp>
#include <gdal/layerfeature.hpp>
#include <gdal/geometry_handle.hpp>
#include <gdal/point.hpp>
#include <observer_ptr.hpp>

namespace gdal
{

  class vectorlayer
  {
    using dataset_ptr    = smartptr<GDALDataset>;
    using layer_ptr      = smartptr<OGRLayer>;
    using feature_ptr    = smartptr<OGRFeature>;
    using featuredef_ptr = smartptr<OGRFeatureDefn>;

  public:
    vectorlayer () = default;

    vectorlayer (smartptr<OGRLayer> p): p_(p) {}

    auto spatial_filter () const
    {
      auto filter_ptr = smartptr<OGRGeometry>(p_->GetSpatialFilter()  // raw pointer
                                              , p_                    // owner
                                              , [] () {}              // deleter
                                              , [] () {}              // increment
                                              , [] () {}              // decrement
                                              , [] () { return 0; }); // use_count
      return geometry_handle(filter_ptr);
    }

    auto& spatial_filter (OGRGeometry* geometryp)
    {
      p_->SetSpatialFilter(geometryp);
      return *this;
    }

    auto& spatial_filter (point<geographic_srs> pt)
    {
      p_->SetSpatialFilter(pt.p().get());
      return *this;
    }

    auto& spatial_filter (geometry_handle h)
    {
      p_->SetSpatialFilter(h.gdal_ptr().get());
      return *this;
    }

    auto clear_spatial_filter () { p_->SetSpatialFilter(NULL); }

    void reset_reading () { p_->ResetReading(); }

    int field_index (const char* name)
    { return p_->GetLayerDefn()->GetFieldIndex(name); }

    // Does not affect next_feature() position
    auto feature (int index)
    {
        auto p = smartptr<OGRFeature>
          (
            [this, index] () { return p_->GetFeature(index); }
            , p_
            );

      return layerfeature(p);
    }

    auto next_feature ()
    {
      return layerfeature(smartptr<OGRFeature>([this] () { return p_->GetNextFeature(); }, p_));
    }

    auto feature_count () { return p_->GetFeatureCount(); }

    auto srs ()
      {
        auto p_raw = p_->GetSpatialRef();

        smartptr<OGRSpatialReference>
          spatialrefp(
            p_raw
            , [] () {} // deleter
            , [p_raw] () { p_raw->Reference(); } // increment
            , [p_raw] () { p_raw->Dereference(); } // decrement
            , [p_raw] () { return p_raw->GetReferenceCount(); } // use_count
            );

        return geographic_srs(spatialrefp);
      }

    auto p () const { return p_; }


  private:
    layer_ptr      p_;
  };

} // namespace gdal

#endif
