/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_DISTRIBUTION_MCMC_MCMC_SIGNAL_POLICY_HPP
#define GRAPH_MODEL_DISTRIBUTION_MCMC_MCMC_SIGNAL_POLICY_HPP 1

#include <graph_model/signals/signal.hpp>
#include <graph_model/signals/tags.hpp>
#include <graph_model/type/signal.hpp>
#include <mpl/enable_constructor.hpp>
#include <use_default.hpp>

namespace graph_model
{

  namespace MCMC
  {

    class mcmc_signal_policy
    {
      using self = mcmc_signal_policy;

      // using signal_type = boost::signals2::signal<void()>;
      // using signal_type = signals::signal<void()>;
      using iteration_signal_type
      = type::signal_t<void(),use_default<>,signals::tag::iteration>;
      using accept_signal_type
      = type::signal_t<void(),use_default<>,signals::tag::accept>;
      using reject_signal_type
      = type::signal_t<void(),use_default<>,signals::tag::reject>;

    public:
      template < typename ... Args
		 , typename = mpl::enable_constructor_t<self,Args...>
		 >
      explicit mcmc_signal_policy (Args&& ... args)
	: iteration_signal_(std::forward<Args>(args)...)
	, accept_signal_(std::forward<Args>(args)...)
	, reject_signal_(std::forward<Args>(args)...)
      {}

    public:
      template < typename SignalTag, typename Slot >
      auto&& connect_signal (Slot&& slot) &&
      { connect<SignalTag>(std::forward<Slot>(slot)); return std::move(*this); }

    public:
      template < typename SignalTag, typename Slot >
      auto connect (Slot&& slot) { return connect(std::forward<Slot>(slot),SignalTag()); }

      template < typename SignalTag >
      void emit () { emit(SignalTag()); }

      template < typename SignalTag >
      void emit () const { emit(SignalTag()); }

    private:
      template < typename Slot >
      auto connect (Slot&& slot, signals::tag::iteration)
      { return iteration_signal_.connect(std::forward<Slot>(slot)); }

      template < typename Slot >
      auto connect (Slot&& slot, signals::tag::accept)
      { return accept_signal_.connect(std::forward<Slot>(slot)); }

      template < typename Slot >
      auto connect (Slot&& slot, signals::tag::reject)
      { return reject_signal_.connect(std::forward<Slot>(slot)); }

    private:
      void emit (signals::tag::iteration)
      { return iteration_signal_(); }

      void emit (signals::tag::accept)
      { return accept_signal_(); }

      void emit (signals::tag::reject)
      { return reject_signal_(); }

    private:
      void emit (signals::tag::iteration) const
      { return iteration_signal_(); }

      void emit (signals::tag::accept) const
      { return accept_signal_(); }

      void emit (signals::tag::reject) const
      { return reject_signal_(); }

    private:
      iteration_signal_type iteration_signal_;
      accept_signal_type accept_signal_;
      reject_signal_type reject_signal_;
    };

    template < typename ... Args >
    auto make_mcmc_signal_policy (Args&& ... args)
    {
      return mcmc_signal_policy(std::forward<Args>(args)...);
    }

  } // namespace MCMC

} // namespace graph_model

#endif
