/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_MCMC_RANDOM_VARIABLE_PROXY_COLLECTION_HPP
#define GRAPH_MODEL_MCMC_RANDOM_VARIABLE_PROXY_COLLECTION_HPP 1

#include <algorithm>
#include <vector>
#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>
#include <graph_model/MCMC/random_variable_proxy.hpp>
#include <graph_model/vertex_and_edge_list_graph.hpp>
#include <logging/channel_log.hpp>
#include <logging/global_log.hpp>

namespace graph_model
{

  namespace MCMC
  {

    class random_variable_proxy_collection
    {
      using rv_proxy_type = random_variable_proxy;
      using collection_type = std::vector<rv_proxy_type>;
      using log_type = logging::channel_log;

    public:
      void clear () { collection_.clear(); }

      void push_back (rv_proxy_type proxy) { collection_.push_back(proxy); }

      void use_current_variate ()
      {
	using namespace boost::log::trivial;
	using std::for_each;
	GLOBAL_LOG(log_,trace) << "use_current_variate()";
	for_each(collection_.begin(),collection_.end(),
		 [](auto& proxy){ proxy.use_current_variate(); });
	GLOBAL_LOG(log_,trace) << "use_current_variate() completed";
      }

      void use_proposed_variate ()
      {
	using namespace boost::log::trivial;
	using std::for_each;
	GLOBAL_LOG(log_,trace) << "use_proposed_variate()";
	for_each(collection_.begin(),collection_.end(),
		 [](auto& proxy){ proxy.use_proposed_variate(); });
	GLOBAL_LOG(log_,trace) << "use_proposed_variate() completed";
      }

      void accept ()
      {
	using namespace boost::log::trivial;
	using std::for_each;
	GLOBAL_LOG(log_,trace) << "accept()";
	for_each(collection_.begin(),collection_.end(),
		 [](auto& proxy){ proxy.accept(); });
	GLOBAL_LOG(log_,trace) << "accept() completed";
      }

      void reject ()
      {
	using namespace boost::log::trivial;
	using std::for_each;
	GLOBAL_LOG(log_,trace) << "reject()";
	for_each(collection_.begin(),collection_.end(),
		 [](auto& proxy){ proxy.reject(); });
	GLOBAL_LOG(log_,trace) << "reject() completed";
      }

      void commit ()
      {
	using namespace boost::log::trivial;
	using std::for_each;
	GLOBAL_LOG(log_,trace) << "commit()";
	for_each(collection_.begin(),collection_.end(),
		 [](auto& proxy){ proxy.commit(); });
	GLOBAL_LOG(log_,trace) << "commit() completed";
      }

      void graph_callback (vertex_and_edge_list_graph& graph)
      {
	using namespace boost::log::trivial;
	using std::for_each;
	GLOBAL_LOG(log_,trace) << "graph_callback()";
	for_each(collection_.begin(),collection_.end(),
		 [&graph](auto& proxy){ proxy.graph_callback(graph); });
	GLOBAL_LOG(log_,trace) << "graph_callback() completed";
      }

    private:
      collection_type collection_;
      log_type log_ = "graph_model::MCMC::random_variable_proxy_collection";
    };

  } // namespace MCMC

} // namespace graph_model

#endif
