/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_DISTRIBUTION_MCMC_SAMPLER_COLLECTION_HPP
#define GRAPH_MODEL_DISTRIBUTION_MCMC_SAMPLER_COLLECTION_HPP 1

#include <algorithm>
#include <functional>
#include <vector>
#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>
#include <graph_model/MCMC/tags.hpp>
#include <graph_model/type/signal.hpp>
#include <graph_model/vertex/rvalue_reference_wrapper.hpp>
#include <graph_model/vertex_and_edge_list_graph.hpp>
#include <logging/channel_log.hpp>
#include <logging/global_log.hpp>
#include <mpl/enable_constructor.hpp>
#include <use_default.hpp>

namespace graph_model
{

  namespace MCMC
  {

    class sampler_collection
    {
      using self = sampler_collection;
      using signal_tag = tag::sampler_collection;
      using signal_type = type::signal_t<void(),use_default<>,signal_tag>;
      using log_type = logging::channel_log;

      using graph_signal_type = type::signal_t<void(vertex_and_edge_list_graph&),use_default<>,signal_tag>;

    public:
      template < typename ... Args
		 , typename = mpl::enable_constructor_t<self,Args...>
		 >
      explicit sampler_collection (Args&& ... args)
	: sampler_signal_(std::forward<Args>(args)...)
	, commit_signal_(std::forward<Args>(args)...)
	, graph_signal_(std::forward<Args>(args)...)
      {}

    public:
      auto& reset_sampler_invocation_strategy ()
      { 
	sampler_signal_.reset_invocation_strategy();
	return *this;
      }

      template < typename InvocationPolicy >
      auto& reset_sampler_invocation_strategy (InvocationPolicy&& invocation_policy)
      { 
	sampler_signal_.reset_invocation_strategy(std::forward<InvocationPolicy>(invocation_policy));
	return *this;
      }

      auto& reset_commit_invocation_strategy ()
      { 
	commit_signal_.reset_invocation_strategy();
	return *this;
      }

      template < typename InvocationPolicy >
      auto& reset_commit_invocation_strategy (InvocationPolicy&& invocation_policy)
      { 
	commit_signal_.reset_invocation_strategy(std::forward<InvocationPolicy>(invocation_policy));
	return *this;
      }

      auto& reset_graph_invocation_strategy ()
      { 
	graph_signal_.reset_invocation_strategy();
	return *this;
      }

      template < typename InvocationPolicy >
      auto& reset_graph_invocation_strategy (InvocationPolicy&& invocation_policy)
      { 
	graph_signal_.reset_invocation_strategy(std::forward<InvocationPolicy>(invocation_policy));
	return *this;
      }

      auto& reset_invocation_strategy ()
      {
	reset_sampler_invocation_strategy();
	reset_commit_invocation_strategy();
	reset_graph_invocation_strategy();
	return *this;
      }

      template < typename InvocationPolicy >
      auto& reset_invocation_strategy (InvocationPolicy&& invocation_policy)
      {
	reset_sampler_invocation_strategy(std::forward<InvocationPolicy>(invocation_policy));
	reset_commit_invocation_strategy(std::forward<InvocationPolicy>(invocation_policy));
	reset_graph_invocation_strategy(std::forward<InvocationPolicy>(invocation_policy));
	return *this;
      }

    public:
      void operator () ()
      {
	update();
	commit();
      }

      template < typename Sampler >
      auto& add_sampler (Sampler&& sampler)
      {
 	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "add_sampler<Sampler>(Sampler&&)";
	auto sampler_function = ref(std::forward<Sampler>(sampler));
	sampler_signal_.connect([sampler_function](){ sampler_function(); });
	commit_signal_.connect([sampler_function](){ sampler_function.get().commit(); });
	graph_signal_.connect([sampler_function](auto& graph){ sampler_function.get().graph_callback(graph); });
	return *this;
      }

      auto& update_graph ()
      {
 	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "update_graph() [" << this << "]: &graph_=" << &graph_;
	graph_signal_(graph_);
	GLOBAL_LOG(log_,trace) << "update_graph() [" << this << "]: &graph_=" << &graph_ << " completed";
	return *this;
      }

      auto& graph () { return graph_; }
      auto& graph () const { return graph_; }

    private:
      void update ()
      {
 	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "update()";
	sampler_signal_();
	GLOBAL_LOG(log_,trace) << "update() done";
      }

      void commit ()
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "commit()";
	commit_signal_();
	GLOBAL_LOG(log_,trace) << "commit() done";
      }

    private:
      signal_type sampler_signal_;
      signal_type commit_signal_;
      graph_signal_type graph_signal_;
      vertex_and_edge_list_graph graph_;
      mutable log_type log_ = "graph_model::MCMC::sampler_collection";
    };

    template < typename ... Args >
    auto make_sampler_collection (Args&& ... args)
    {
      using sampler_collection_type = sampler_collection;
      return sampler_collection_type(std::forward<Args>(args)...);
    }

  } // namespace MCMC

} // namespace graph_model

#endif
