/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_DISTRIBUTION_BETA_BETA_STRATEGY_HPP
#define GRAPH_MODEL_DISTRIBUTION_BETA_BETA_STRATEGY_HPP 1

#include <cmath>
#include <type_traits>
#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>
#include <boost/math/distributions/beta.hpp>
#include <boost/math/special_functions/beta.hpp>
#include <boost/random/beta_distribution.hpp>
#include <graph_model/distribution/beta/tags.hpp>
#include <graph_model/distribution/scalar_distribution_policy.hpp>
#include <graph_model/random/distribution.hpp>
#include <graph_model/traits/real.hpp>
#include <graph_model/vertex/expression_function.hpp>
// #include <graph_model/vertex/expression_terminal.hpp>
#include <graph_model/vertex/parameter_tags.hpp>
#include <graph_model/vertex/tagged_parameter.hpp>
#include <graph_model/vertex/vertex_expression.hpp>
#include <sample_space/is_possible.hpp>

#include <graph_model/distribution/parameter/alpha.hpp>
#include <graph_model/distribution/parameter/beta.hpp>
#include <graph_model/distribution/parameter/parameter_pack.hpp>
#include <graph_model/distribution/parameter/precision.hpp>
#include <graph_model/distribution/parameter/probability.hpp>
#include <graph_model/distribution/parameter/variate.hpp>

#include <logging/channel_log.hpp>
#include <logging/global_log.hpp>

namespace graph_model
{

  namespace distribution
  {

    namespace beta
    {

      /// \brief Generate a beta distribution
      ///
      /// This policy class models the Distribution Generator concept,
      /// which generates a distribution from its parameters.  A
      /// distribution models the corresponding concept in the Boost
      /// Math library.  Specifically, it can be evaluated by the \c
      /// pdf() function (and possibly the \c cdf() function if it is
      /// a scalar distribution).
      ///
      /// This policy generates a beta distribution from the Boost
      /// Math library.
      ///
      struct distribution_policy
      {
	template < typename Variate, typename Alpha, typename Beta >
	static auto apply (Alpha alpha, Beta beta)
	{
	  using variate_type
	    = std::common_type_t<
	      traits::real_t<Variate>
	    , traits::real_t<Alpha>
	    , traits::real_t<Beta>
	    >;
	  using beta_distribution = boost::math::beta_distribution<variate_type>;
	  return beta_distribution(alpha,beta);
	}
      };

      /// \brief Generate a beta random number generator
      ///
      /// This policy class models the RNG Generator concept, which
      /// generates a random number distribution from its parameters.
      /// A random number distribution models the corresponding
      /// concept in the Boost Random library.  Specifically, it can
      /// be evaluated to obtain a sample from the distribution.
      ///
      /// This policy generates a beta distribution from the Boost
      /// Random library.
      ///
      struct generator_policy
      {
	template < typename Alpha, typename Beta >
	static auto& apply ()
	{
	  using variate_type
	    = std::common_type_t<
	      traits::real_t<Alpha>
	    , traits::real_t<Beta>
	    >;
	  using beta_distribution
	    = random::distribution< boost::random::beta_distribution<variate_type> >;
	  return beta_distribution::apply();
	}
      };

      /// \brief Translate the parameter pack into alpha
      ///
      /// The parameters of the beta distribution may be specified in
      /// one of two ways:
      ///
      /// - alpha and beta
      /// - probability and precision
      ///
      /// However, the underlying distribution and generator are
      /// parameterized with alpha and beta.  Specializations of this
      /// class facilitate translation of a parameter pack holding
      /// either set of parameters into alpha.
      ///
      template < typename Tag >
      class alpha;

      template <>
      struct alpha < tag::alpha_beta >
      {
	template < typename ParameterPack, typename ... T >
	static auto apply (ParameterPack const& pack, T&& ... t)
	{ return pack.template value<parameter::tag::alpha>(std::forward<T>(t)...); }
      };

      template <>
      struct alpha < tag::probability_precision >
      {
	template < typename ParameterPack, typename ... T >
	static auto apply (ParameterPack const& pack, T&& ... t)
	{
	  auto probability = pack.template value<parameter::tag::probability>(std::forward<T>(t)...);
	  auto precision = pack.template value<parameter::tag::precision>(std::forward<T>(t)...);
	  auto alpha = probability * precision;
	  return alpha;
	}
      };

      /// \brief Translate the parameter pack into beta
      ///
      /// The parameters of the beta distribution may be specified in
      /// one of two ways:
      ///
      /// - alpha and beta
      /// - probability and precision
      ///
      /// However, the underlying distribution and generator are
      /// parameterized with alpha and beta.  Specializations of this
      /// class facilitate translation of a parameter pack holding
      /// either set of parameters into beta.
      ///
      template < typename Tag >
      class beta;

      template <>
      struct beta < tag::alpha_beta >
      {
	template < typename ParameterPack, typename ... T >
	static auto apply (ParameterPack const& pack, T&& ... t)
	{ return pack.template value<parameter::tag::beta>(std::forward<T>(t)...); }
      };

      template <>
      struct beta < tag::probability_precision >
      {
	template < typename ParameterPack, typename ... T >
	static auto apply (ParameterPack const& pack, T&& ... t)
	{
	  auto probability = pack.template value<parameter::tag::probability>(std::forward<T>(t)...);
	  auto precision = pack.template value<parameter::tag::precision>(std::forward<T>(t)...);
	  auto beta = (1 - probability) * precision;
	  return beta;
	}
      };

      /// \brief Strategy for beta distributions
      /// \tparam Tag Determine how to interpret the parameters
      /// \tparam ParameterPack The parameters
      ///
      /// A beta distribution is a distribution across a scalar
      /// variate, so this strategy models the Scalar Distribution
      /// concept, implemented as appropriate for a beta distribution.
      /// Generally, the implementation simply delegates to the
      /// appropriate components of the Boost Math or Random
      /// libraries.
      ///
      /// The parameters of a beta distrubtion may be specified in
      /// either of two different ways:
      ///
      /// - as alpha and beta, which is how the underlying Boost
      ///   libraries are parameterized, or
      /// - as the probability and precision.  In this case, alpha
      ///   equals probability * precision.
      ///
      /// Given that different use cases may naturally prefer one of
      /// these over the rest, both are supported.  The \c Tag
      /// template parameter determines how the parameters should be
      /// interpreted.  In either case, both alpha and beta are
      /// calculated before being passed to the underlying Boost
      /// library functions.
      ///
      template < typename Tag, typename ParameterPack >
      class beta_strategy
      {
	using policy_type = scalar_distribution_policy<distribution_policy,generator_policy>;
	using log_type = logging::channel_log;

      public:
	beta_strategy () = default;
	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	explicit beta_strategy (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack)
	  : pack_(pack)
	{}

      public:
	template < typename ... T >
	auto variate (T&& ... t) const
	{ return pack_.template value<parameter::tag::variate>(std::forward<T>(t)...); }

	template < typename ... T >
	auto alpha (T&& ... t) const
	{
	  return distribution::beta::alpha<Tag>::apply(pack_,std::forward<T>(t)...);
	}

	template < typename ... T >
	auto beta (T&& ... t) const
	{
	  return distribution::beta::beta<Tag>::apply(pack_,std::forward<T>(t)...);
	}

	template < typename ... T >
	auto probability (T&& ... t) const
	{
	  auto alpha = this->alpha(std::forward<T>(t)...);
	  auto beta = this->beta(std::forward<T>(t)...);
	  using real_type = traits::real_t<std::common_type_t<decltype(alpha),decltype(beta)>>;
	  return static_cast<real_type>(alpha) / (alpha + beta);
	}

	template < typename ... T >
	auto precision (T&& ... t) const
	{
	  auto alpha = this->alpha(std::forward<T>(t)...);
	  auto beta = this->beta(std::forward<T>(t)...);
	  return alpha + beta;
	}

      public:
	template < typename ... T >
	auto sample (T&& ... t) const
	{
	  using namespace boost::log::trivial;

	  auto alpha = this->alpha(std::forward<T>(t)...);
	  auto beta = this->beta(std::forward<T>(t)...);
	  auto result = sample(alpha,beta);

	  GLOBAL_LOG(log_,trace)
	    << "alpha=" << alpha << ", "
	    << "beta=" << beta << ", "
	    << "sample=" << result;

	  return result;
	}
	template < typename ... T >
	auto pdf (T&& ... t) const
	{
	  using namespace boost::log::trivial;
	  
	  auto variate = this->variate(std::forward<T>(t)...);
	  auto alpha = this->alpha(std::forward<T>(t)...);
	  auto beta = this->beta(std::forward<T>(t)...);
	  auto result = pdf(variate,alpha,beta);

	  GLOBAL_LOG(log_,trace)
	    << "variate=" << variate << ", "
	    << "alpha=" << alpha << ", "
	    << "beta=" << beta << ", "
	    << "pdf=" << result;

	  return result;
	}
	template < typename ... T >
	auto cdf (T&& ... t) const
	{
	  using namespace boost::log::trivial;

	  auto variate = this->variate(std::forward<T>(t)...);
	  auto alpha = this->alpha(std::forward<T>(t)...);
	  auto beta = this->beta(std::forward<T>(t)...);
	  auto result = cdf(variate,alpha,beta);

	  GLOBAL_LOG(log_,trace)
	    << "variate=" << variate << ", "
	    << "alpha=" << alpha << ", "
	    << "beta=" << beta << ", "
	    << "cdf=" << result;

	  return result;
	}
	template < typename Variate, typename ... T >
	auto cdf (Variate&& variate, T&& ... t) const
	{
	  return cdf
	    (vertex::make_expression_function(std::forward<Variate>(variate)),
	     alpha(std::forward<T>(t)...),
	     beta(std::forward<T>(t)...));
	}

      private:
	template < typename Alpha, typename Beta >
	auto sample (Alpha alpha, Beta beta) const
	{
	  using sample_space::is_possible;
	  assert(alpha > 0);
	  assert(beta > 0);
	  assert(is_possible(alpha));
	  assert(is_possible(beta));
	  return policy_type::sample(alpha,beta);
	}

	template < typename Variate, typename Alpha, typename Beta >
	auto pdf (Variate variate, Alpha alpha, Beta beta) const
	{
	  using namespace boost::log::trivial;
	  using sample_space::is_possible;

	  auto result =
	    alpha > 0
	    && beta > 0
	    && variate >= 0
	    && variate <= 1
	    && is_possible(variate)
	    && is_possible(alpha)
	    && is_possible(beta)
	    ? policy_type::eval_pdf(variate,alpha,beta)
	    : 0;

	  GLOBAL_LOG(log_,trace)
	    << "variate=" << variate << ", "
	    << "alpha=" << alpha << ", "
	    << "beta=" << beta << ", "
	    << "pdf=" << result;

	  return result;
	}

	template < typename Variate, typename Alpha, typename Beta >
	auto cdf (Variate variate, Alpha alpha, Beta beta) const
	{
	  using sample_space::is_possible;
	  return
	    alpha > 0
	    && beta > 0
	    && variate >= 0
	    && variate <= 1
	    && is_possible(variate)
	    && is_possible(alpha)
	    && is_possible(beta)
	    ? policy_type::eval_cdf(variate,alpha,beta)
	    : 0;
	}

	template < typename Expr, typename Alpha, typename Beta >
	auto cdf (vertex::expression_function<Expr>&& variate, Alpha alpha, Beta beta) const
	{ return cdf(variate(),alpha,beta); }

      private:
	ParameterPack pack_;
	mutable log_type log_ = "graph_model::distribution::beta";
      };

      template < typename Tag, typename ParameterPack >
      auto make_beta_strategy (ParameterPack parameter_pack)
      {
	using beta_strategy_type = beta_strategy<Tag,ParameterPack>;
	return beta_strategy_type(parameter_pack);
      };

    } // namespace beta

  } // namespace distribution

} // namespace graph_model

#endif
