/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file
/// \brief Construct a strategy for a dirichlet distribution
///

#ifndef GRAPH_MODEL_DISTRIBUTION_DIRICHLET_DIRICHLET_DISTRIBUTION_HPP
#define GRAPH_MODEL_DISTRIBUTION_DIRICHLET_DIRICHLET_DISTRIBUTION_HPP 1

#include <boost/hana/map.hpp>
#include <graph_model/distribution/dirichlet/dirichlet_strategy.hpp>
#include <graph_model/distribution/dirichlet/tags.hpp>
#include <graph_model/distribution/parameter/alpha.hpp>
#include <graph_model/distribution/parameter/mean.hpp>
#include <graph_model/distribution/parameter/named_parameter.hpp>
#include <graph_model/distribution/parameter/parameter.hpp>
#include <graph_model/distribution/parameter/parameter_pack.hpp>
#include <graph_model/distribution/parameter/precision.hpp>
#include <graph_model/distribution/parameter/variate.hpp>

namespace graph_model
{

  namespace distribution
  {

    namespace dirichlet
    {

      namespace detail
      {

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   >
	auto dirichlet_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>);

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto dirichlet_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 parameter::named_parameter<Key,Value>,
	 Parameters&& ...);

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto dirichlet_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 Value&&,
	 Parameters&& ...);

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	auto dirichlet_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>);

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto dirichlet_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 parameter::named_parameter<Key,Value>,
	 Parameters&& ...);

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto dirichlet_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 Value&&,
	 Parameters&& ...);

	template < typename Tag >
	struct retag
	{
	  template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		     , typename Value
		     , typename ... Parameters
		   >
	  static auto apply
	  (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	   parameter::named_parameter<Tag,Value> parameter,
	   Parameters&& ... parameters)
	  {
	    return dirichlet_distribution
	      (pack.add(parameter),std::forward<Parameters>(parameters)...);
	  }
	};

	template <>
	struct retag < parameter::tag::alpha >
	{
	  template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		     , typename Value
		     , typename ... Parameters
		   >
	  static auto apply
	  (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	   parameter::named_parameter<parameter::tag::alpha,Value> parameter,
	   Parameters&& ... parameters)
	  {
	    return dirichlet_distribution<tag::alpha>
	      (pack.add(parameter),std::forward<Parameters>(parameters)...);
	  }
	};

	template <>
	struct retag < parameter::tag::mean >
	{
	  template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		     , typename Value
		     , typename ... Parameters
		   >
	  static auto apply
	  (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	   parameter::named_parameter<parameter::tag::mean,Value> parameter,
	   Parameters&& ... parameters)
	  {
	    return dirichlet_distribution<tag::mean_precision>
	      (pack.add(parameter),std::forward<Parameters>(parameters)...);
	  }
	};

	template <>
	struct retag < parameter::tag::precision >
	{
	  template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		     , typename Value
		     , typename ... Parameters
		   >
	  static auto apply
	  (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	   parameter::named_parameter<parameter::tag::precision,Value> parameter,
	   Parameters&& ... parameters)
	  {
	    return dirichlet_distribution<tag::mean_precision>
	      (pack.add(parameter),std::forward<Parameters>(parameters)...);
	  }
	};

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   >
	auto dirichlet_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack)
	{ return make_dirichlet_strategy<Tag>(pack); }

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto dirichlet_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 parameter::named_parameter<Key,Value> parameter,
	 Parameters&& ... parameters)
	{
	  return dirichlet_distribution<Tag>
	    (pack.add(parameter),
	     std::forward<Parameters>(parameters)...);
	}

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto dirichlet_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 Value&& value,
	 Parameters&& ... parameters)
	{
	  return dirichlet_distribution<Tag>
	    (pack.add(std::forward<Value>(value)),
	     std::forward<Parameters>(parameters)...);
	}

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	auto dirichlet_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack)
	{ return dirichlet_distribution<tag::alpha>(pack); }

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto dirichlet_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 parameter::named_parameter<Key,Value> parameter,
	 Parameters&& ... parameters)
	{
	  return detail::retag<Key>::apply(pack,parameter,std::forward<Parameters>(parameters)...);
	}

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto dirichlet_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 Value&& value,
	 Parameters&& ... parameters)
	{
	  return dirichlet_distribution
	    (pack.add(std::forward<Value>(value)),
	     std::forward<Parameters>(parameters)...);
	}

      }	// namespace detail

      template < typename ... Parameters >
      auto dirichlet_distribution (Parameters&& ... parameters)
      {
	using parameter::make_parameter;
	using parameter::make_parameter_pack;
	return detail::dirichlet_distribution
	  (make_parameter_pack
	   (make_parameter<parameter::tag::alpha>(),
	    make_parameter<parameter::tag::mean>(),
	    make_parameter<parameter::tag::precision>(),
	    make_parameter<parameter::tag::variate>()),
	   std::forward<Parameters>(parameters)...);
      }

    } // namespace dirichlet

  } // namespace distribution

} // namespace graph_model

#endif
