/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_DISTRIBUTION_NORMAL_NORMAL_STRATEGY_HPP
#define GRAPH_MODEL_DISTRIBUTION_NORMAL_NORMAL_STRATEGY_HPP 1

#include <cmath>
#include <type_traits>
#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>
#include <boost/math/distributions/normal.hpp>
#include <boost/random/normal_distribution.hpp>
#include <graph_model/distribution/scalar_distribution_policy.hpp>
#include <graph_model/random/distribution.hpp>
#include <graph_model/traits/real.hpp>
#include <graph_model/vertex/expression_function.hpp>
// #include <graph_model/vertex/expression_terminal.hpp>
#include <graph_model/vertex/parameter_tags.hpp>
#include <graph_model/vertex/tagged_parameter.hpp>
#include <graph_model/vertex/vertex_expression.hpp>
#include <sample_space/is_possible.hpp>

#include <graph_model/distribution/parameter/dispersion.hpp>
#include <graph_model/distribution/parameter/mean.hpp>
#include <graph_model/distribution/parameter/parameter_pack.hpp>
#include <graph_model/distribution/parameter/precision.hpp>
#include <graph_model/distribution/parameter/standard_deviation.hpp>
#include <graph_model/distribution/parameter/variance.hpp>
#include <graph_model/distribution/parameter/variate.hpp>

#include <logging/channel_log.hpp>
#include <logging/global_log.hpp>

namespace graph_model
{

  namespace distribution
  {

    namespace normal
    {

      /// \brief Generate a normal distribution
      ///
      /// This policy class models the Distribution Generator concept,
      /// which generates a distribution from its parameters.  A
      /// distribution models the corresponding concept in the Boost
      /// Math library.  Specifically, it can be evaluated by the \c
      /// pdf() function (and possibly the \c cdf() function if it is
      /// a scalar distribution).
      ///
      /// This policy generates a normal distribution from the Boost
      /// Math library.
      ///
      struct distribution_policy
      {
	template < typename Variate, typename Mean, typename StandardDeviation >
	static auto apply (Mean mean, StandardDeviation standard_deviation)
	{
	  using variate_type
	    = std::common_type_t<
	      traits::real_t<Variate>
	    , traits::real_t<Mean>
	    , traits::real_t<StandardDeviation>
	    >;
	  using normal_distribution = boost::math::normal_distribution<variate_type>;
	  return normal_distribution(mean,standard_deviation);
	}
      };

      /// \brief Generate a normal random number generator
      ///
      /// This policy class models the RNG Generator concept, which
      /// generates a random number distribution from its parameters.
      /// A random number distribution models the corresponding
      /// concept in the Boost Random library.  Specifically, it can
      /// be evaluated to obtain a sample from the distribution.
      ///
      /// This policy generates a normal distribution from the Boost
      /// Random library.
      ///
      struct generator_policy
      {
	template < typename Mean, typename StandardDeviation >
	static auto& apply ()
	{
	  using variate_type
	    = std::common_type_t<
	      traits::real_t<Mean>
	    , traits::real_t<StandardDeviation>
	    >;
	  using normal_distribution
	    = random::distribution< boost::normal_distribution<variate_type> >;
	  return normal_distribution::apply();
	}
      };

      /// \brief Translate the dispersion parameter into the standard deviation
      ///
      /// The dispersion of a normal distribution may be specified in
      /// several different ways, as the standard deviation, as the
      /// variance, or as the precision.  However, the underlying
      /// distribution and generator require it to be specified as the
      /// standard deviation.  Specializations of this class for each
      /// of the three cases translates the value of the dispersion
      /// into the standard deviation.
      ///
      template < typename Tag >
      class dispersion;

      template <>
      struct dispersion < parameter::tag::standard_deviation >
      {
	template < typename ParameterPack, typename ... T >
	static auto apply (ParameterPack const& pack, T&& ... t)
	{
	  return pack.template value<parameter::tag::dispersion>(std::forward<T>(t)...);
	}
      };

      template <>
      struct dispersion < parameter::tag::precision >
      {
	template < typename ParameterPack, typename ... T >
	static auto apply (ParameterPack const& pack, T&& ... t)
	{
	  auto precision = pack.template value<parameter::tag::dispersion>(std::forward<T>(t)...);
	  using precision_type = decltype(precision);
	  using real_type = traits::real_t<precision_type>;
	  return one<real_type>() / precision;
	}

      private:
	template < typename T >
	static auto one () { return T(1); }
      };

      template <>
      struct dispersion < parameter::tag::variance >
      {
	template < typename ParameterPack, typename ... T >
	static auto apply (ParameterPack const& pack, T&& ... t)
	{
	  auto variance = pack.template value<parameter::tag::dispersion>(std::forward<T>(t)...);
	  return sqrt(variance);
	}
      };

      /// \brief Strategy for normal distributions
      /// \tparam Tag Determine how to interpret the dispersion
      /// \tparam Variate The expression for the (scalar) variate
      /// \tparam Mean The expression for the expected value of the distribution
      /// \tparam Dispersion The expression for the dispersion of the distribution (see below)
      ///
      /// A normal (or Gaussian) distribution is a distribution across
      /// a scalar variate, so this strategy models the Scalar
      /// Distribution concept, implemented as appropriate for a
      /// normal distribution.  Generally, the implementation simply
      /// delegates to the appropriate components of the Boost Math or
      /// Random libraries.
      ///
      /// The dispersion of a normal distrubtion may be specified in
      /// any of three different ways:
      ///
      /// - as the standard deviation, which is how the underlying
      ///   Boost libraries are parameterized,
      /// - as the variance, or
      /// - as the precision, i.e., inverse of the standard deviation.
      ///
      /// Given that different use cases may naturally prefer one of
      /// these over the rest, all three are supported.  The \c Tag
      /// template parameter determines how the dispersion should be
      /// interpreted.  In the case of either variance or precision,
      /// the value of the dispersion is converted into the
      /// corresponding value of the standard deviation before being
      /// passed to the underlying Boost library functions.
      ///
      /// The other three template parameters, \c Variate, \c Mean,
      /// and \c Dispersion, must be functor classes.  Their function
      /// call operator must evaluate the expression each represents.
      ///
      template < typename Tag, typename ParameterPack >
      class normal_strategy
      {
	// using variate_type = std::decay_t<Variate>;
	// using mean_type = std::decay_t<Mean>;
	// using dispersion_type = std::decay_t<Dispersion>;
	using policy_type = scalar_distribution_policy<distribution_policy,generator_policy>;
	using log_type = logging::channel_log;

      public:
	normal_strategy () = default;
	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	explicit normal_strategy (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack)
	  : pack_(pack)
	{}

      public:
	template < typename ... T >
	auto variate (T&& ... t) const
	{ return pack_.template value<parameter::tag::variate>(std::forward<T>(t)...); }
	template < typename ... T >
	auto mean (T&& ... t) const
	{ return pack_.template value<parameter::tag::mean>(std::forward<T>(t)...); }
	template < typename ... T >
	auto dispersion (T&& ... t) const
	{ return pack_.template value<parameter::tag::dispersion>(std::forward<T>(t)...); }

	template < typename ... T >
	auto standard_deviation (T&& ... t) const
	{
	  return normal::dispersion<Tag>::apply(pack_,std::forward<T>(t)...);
	}

      public:
	template < typename ... T >
	auto sample (T&& ... t) const
	{
	  using namespace boost::log::trivial;

	  auto m = mean(std::forward<T>(t)...);
	  auto sd = standard_deviation(std::forward<T>(t)...);
	  auto result = sample
	    (mean(std::forward<T>(t)...),
	     standard_deviation(std::forward<T>(t)...));

	  GLOBAL_LOG(log_,trace)
	    << "mean=" << m << ", "
	    << "standard_deviation=" << sd << ", "
	    << "sample=" << result;

	  return result;
	}
	template < typename ... T >
	auto pdf (T&& ... t) const
	{
	  using namespace boost::log::trivial;
	  
	  auto v = variate(std::forward<T>(t)...);
	  auto m = mean(std::forward<T>(t)...);
	  auto sd = standard_deviation(std::forward<T>(t)...);
	  auto result = pdf(v,m,sd);

	  GLOBAL_LOG(log_,trace)
	    << "variate=" << v << ", "
	    << "mean=" << m << ", "
	    << "standard_deviation=" << sd << ", "
	    << "pdf=" << result;

	  return result;
	}
	template < typename ... T >
	auto cdf (T&& ... t) const
	{
	  return cdf
	    (variate(std::forward<T>(t)...),
	     mean(std::forward<T>(t)...),
	     standard_deviation(std::forward<T>(t)...));
	}
	template < typename Variate, typename ... T >
	auto cdf (Variate&& variate, T&& ... t) const
	{
	  return cdf
	    (vertex::make_expression_function(std::forward<Variate>(variate)),
	     mean(std::forward<T>(t)...),
	     standard_deviation(std::forward<T>(t)...));
	}

      private:
	template < typename Mean, typename StandardDeviation >
	auto sample (Mean mean, StandardDeviation standard_deviation) const
	{
	  using sample_space::is_possible;
	  assert(standard_deviation > 0);
	  assert(is_possible(mean));
	  assert(is_possible(standard_deviation));
	  return policy_type::sample(mean,standard_deviation);
	}

	template < typename Variate, typename Mean, typename StandardDeviation >
	auto pdf (Variate variate, Mean mean, StandardDeviation standard_deviation) const
	{
	  using sample_space::is_possible;
	  return
	    standard_deviation > 0
	    && is_possible(mean)
	    && is_possible(standard_deviation)
	    ? policy_type::eval_pdf(variate,mean,standard_deviation)
	    : 0;
	}

	template < typename Variate, typename Mean, typename StandardDeviation >
	auto cdf (Variate variate, Mean mean, StandardDeviation standard_deviation) const
	{
	  using sample_space::is_possible;
	  return
	    standard_deviation > 0
	    && is_possible(variate)
	    && is_possible(mean)
	    && is_possible(standard_deviation)
	    ? policy_type::eval_cdf(variate,mean,standard_deviation)
	    : 0;
	}

	template < typename Expr, typename Mean, typename StandardDeviation >
	auto cdf (vertex::expression_function<Expr>&& variate, Mean mean, StandardDeviation standard_deviation) const
	{ return cdf(variate(),mean,standard_deviation); }

      private:
	ParameterPack pack_;
	mutable log_type log_ = "graph_model::distribution::normal";
      };

      template < typename Tag, typename ParameterPack >
      auto make_normal_strategy (ParameterPack parameter_pack)
      {
	using normal_strategy_type = normal_strategy<Tag,ParameterPack>;
	return normal_strategy_type(parameter_pack);
      };

    } // namespace normal

  } // namespace distribution

} // namespace graph_model

#endif
