/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_FUNCTIONAL_EQUAL_TO_HPP
#define GRAPH_MODEL_FUNCTIONAL_EQUAL_TO_HPP 1

#include <functional>
#include <graph_model/type_traits/has_equal_to.hpp>

namespace graph_model
{

  namespace functional
  {

    /// \brief Provide a customization point for \c std::equal_to
    /// \tparam T The type to compare
    ///
    /// The graph model library requires a specialization of \c
    /// std::equal_to for \c std::reference_wrapper.  Since it is
    /// unwise to provide such specializations within the \c std::
    /// namespace, this customization point is provided.  By default,
    /// it simply delegates to \c std::equal_t, but allows for
    /// specializations, including for types defined within the \c
    /// std:: namespace.
    ///
    template < typename T >
    struct equal_to
      : std::equal_to<T>
    {};

    /// \brief Specialization of \c equal_to for \c std::reference_wrapper
    ///
    template < typename T >
    struct equal_to< std::reference_wrapper<T> >
    {
      using first_argument_type = std::reference_wrapper<T>;
      using second_argument_type = std::reference_wrapper<T>;
      using result_type = bool;

      result_type operator () (first_argument_type const& left, second_argument_type const& right) const
      {
	using has_equal_to = type_traits::has_equal_to_t<first_argument_type>;
	return operator()(has_equal_to(),left,right);
      }

    private:
      result_type operator () (std::false_type, first_argument_type const& left, second_argument_type const& right) const
      { return &left.get() == &right.get(); }

      result_type operator () (std::true_type, first_argument_type const& left, second_argument_type const& right) const
      { return left.get() == right.get(); }
    };

  } // namespace functional

} // namespace graph_model

#endif
