/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_SIGNALS_SIGNAL_FWD_HPP
#define GRAPH_MODEL_SIGNALS_SIGNAL_FWD_HPP 1

#include <use_default.hpp>

namespace graph_model
{

  namespace signals
  {

    /// \brief A signal class
    /// \tparam Signature The function signature
    /// \tparam Combiner The type of combiner for merging results (if any)
    /// \tparam SignalTag A type identifying the signal
    /// \tparam Collection The type of collection for storing function objects
    ///
    /// This class is a minimal signal class.  It delegates its
    /// function call operator to a set of slots, each of which is a
    /// function object with the signature \tparam Signature.  The
    /// results (if any) are combined using a \tparam Combiner object
    /// (if any).  Unlike more general signal classes, however, it
    /// does not provide any connection management.  Rather, it is
    /// assumed that the set of slots cannot be changed.
    ///
    /// The behavior of \c signal may be customized in three
    /// orthogonal ways:
    ///
    /// - The result type of the signal depends on the \c Signature
    ///   and the \c Combiner.  If the \c Combiner is not void, it is
    ///   the result type, which allows user-control over retrieving
    ///   state from the combined slot function results.  Otherwise,
    ///   the result type is the result of \c Signature, and the value
    ///   is unspecified.
    ///
    /// - The algorithm used to invoke the functions is determined by
    ///   the invocation policy, which may be reset at runtime.  By
    ///   default, \c default_invocation_policy is used, which invokes
    ///   functions sequentially.
    ///
    /// - The collection of functions is managed by an object of type
    ///   \tparam Collection.  By default, this is \c
    ///   boost:poly_collection::base_collection, which manages a
    ///   polymorphic collection of objects.  Alternatively, \c
    ///   graph_model::utility::unique_base_collection will ensure
    ///   uniqueness among the polymorphic function objects.  In that
    ///   case, \c operator<() for each type of function object should
    ///   provide a strict weak ordering.
    ///
    /// Note that the order of invocation for functions in the
    /// collection is undefined and entirely under the control of the
    /// \tparam Collection.  In the case of collections derived from
    /// \c boost::poly_collection::base_collection different function
    /// objects of the same type will be grouped together, but neither
    /// the order of groups nor the order of function invocation
    /// within groups is defined.
    ///
    template < typename Signature
	       , typename Combiner = use_default<>
	       , typename SignalTag = use_default<>
	       , template < typename ... > class SlotCollection = use_default
	       >
    class signal;

  } // namespace signals

} // namespace graph_model

#endif
