/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_UTILITY_SEQUENTIAL_INVOCATION_POLICY_HPP
#define GRAPH_MODEL_UTILITY_SEQUENTIAL_INVOCATION_POLICY_HPP 1

#include <tuple>
#include <asynchronous/signals/invoke.hpp>
#include <graph_model/signals/type/invocation_result.hpp>

namespace graph_model
{

  namespace utility
  {

    /// \brief Sequential invocation policy for \c graph_model::utility::function_collection
    ///
    /// This class models the InvocationPolicy concept for \c
    /// graph_model::utility::function_collection, which requires the
    /// four function call operator overloads implemented here.  This
    /// policy executes each function sequentially.
    ///
    class sequential_invocation_policy
    {
      struct result_type
      {
	template < typename Iterator, typename ... Args >
	static auto apply (Iterator i, Args&& ... args)
	{ return (*i)(std::forward<Args>(args)...); }
      };

      struct result_is_void
      {
	template < typename Iterator, typename ... Args >
	static auto apply (Iterator i, Args&& ... args)
	{
	  using result_type = decltype(result_type::apply(i,std::forward<Args>(args)...));
	  return typename std::is_void<result_type>::type();
	}
      };

    public:
      /// \brief Invoke a set of functions sequentially
      /// \returns The result of invoking a set of functions
      ///
      /// The type of the result depends on the signature of the functions.
      ///
      /// | Signature       | Result type | Result value                            |
      /// |-----------------|-------------|-----------------------------------------|
      /// | void(Args...)   | void        | None                                    |
      /// | Result(Args...) | Result      | The result of the last function invoked |
      ///
      template < typename Iterator, typename ... Args >
      auto operator () (Iterator begin, Iterator end, Args&& ... args)
	-> signals::type::invocation_result_t<void,Iterator,Args...>
      {
	using result_is_void = decltype(result_is_void::apply(begin,std::forward<Args>(args)...));
	return invoke(result_is_void(),begin,end,std::forward<Args>(args)...);
      }

      template < typename Iterator, typename ... Args >
      auto operator () (Iterator begin, Iterator end, Args&& ... args) const
	-> signals::type::invocation_result_t<void,Iterator,Args...>
      {
	using result_is_void = decltype(result_is_void::apply(begin,std::forward<Args>(args)...));
	return invoke(result_is_void(),begin,end,std::forward<Args>(args)...);
      }

      /// \brief Invoke a set of functions sequentially with a combiner
      /// \returns \c Combiner, containing the result of combining the function results
      ///
      /// For each invoked function, the \c Combiner's function call
      /// operator will be called with the result of the function (if
      /// not void) or with no argument.
      ///
      template < typename Combiner, typename Iterator, typename ... Args >
      auto operator () (Combiner combiner, Iterator begin, Iterator end, Args&& ... args)
	-> signals::type::invocation_result_t<Combiner,Iterator,Args...>
      {
	using result_is_void = decltype(result_is_void::apply(begin,std::forward<Args>(args)...));
	return invoke(result_is_void(),combiner,begin,end,std::forward<Args>(args)...);
      }

      template < typename Combiner, typename Iterator, typename ... Args >
      auto operator () (Combiner combiner, Iterator begin, Iterator end, Args&& ... args) const
	-> signals::type::invocation_result_t<Combiner,Iterator,Args...>
      {
	using result_is_void = decltype(result_is_void::apply(begin,std::forward<Args>(args)...));
	return invoke(result_is_void(),combiner,begin,end,std::forward<Args>(args)...);
      }

    private:
      template < typename Iterator, typename ... Args >
      auto invoke (std::false_type, Iterator begin, Iterator end, Args&& ... args)
      {
	using result_type = decltype(result_type::apply(begin,std::forward<Args>(args)...));
	result_type last_result;
	std::for_each(begin,end,
		      [&last_result,tuple=std::forward_as_tuple<Args...>(std::forward<Args>(args)...)](auto f)
		      { last_result = asynchronous::signals::invoke::apply(f,tuple); });
	return last_result;
      }

      template < typename Iterator, typename ... Args >
      auto invoke (std::false_type, Iterator begin, Iterator end, Args&& ... args) const
      {
	using result_type = decltype(result_type::apply(begin,std::forward<Args>(args)...));
	result_type last_result;
	std::for_each(begin,end,
		      [&last_result,tuple=std::forward_as_tuple<Args...>(std::forward<Args>(args)...)](auto f)
		      { last_result = asynchronous::signals::invoke::apply(f,tuple); });
	return last_result;
      }

    private:
      template < typename Iterator, typename ... Args >
      auto invoke (std::true_type, Iterator begin, Iterator end, Args&& ... args)
      {
	std::for_each(begin,end,
		      [tuple=std::forward_as_tuple<Args...>(std::forward<Args>(args)...)](auto f)
		      { asynchronous::signals::invoke::apply(f,tuple); });
      }

      template < typename Iterator, typename ... Args >
      auto invoke (std::true_type, Iterator begin, Iterator end, Args&& ... args) const
      {
	std::for_each(begin,end,
		      [tuple=std::forward_as_tuple<Args...>(std::forward<Args>(args)...)](auto f)
		      { asynchronous::signals::invoke::apply(f,tuple); });
      }

    private:
      template < typename Combiner, typename Iterator, typename ... Args >
      auto invoke (std::false_type, Combiner combiner, Iterator begin, Iterator end, Args&& ... args)
      {
	std::for_each(begin,end,[&combiner,tuple=std::forward_as_tuple<Args...>(std::forward<Args>(args)...)](auto f)
		      { combiner(asynchronous::signals::invoke::apply(f,tuple)); });
	return combiner;
      }

      template < typename Combiner, typename Iterator, typename ... Args >
      auto invoke (std::false_type, Combiner combiner, Iterator begin, Iterator end, Args&& ... args) const
      {
	std::for_each(begin,end,[&combiner,tuple=std::forward_as_tuple<Args...>(std::forward<Args>(args)...)](auto f)
		      { combiner(asynchronous::signals::invoke::apply(f,tuple)); });
	return combiner;
      }

    private:
      template < typename Combiner, typename Iterator, typename ... Args >
      auto invoke (std::true_type, Combiner combiner, Iterator begin, Iterator end, Args&& ... args)
      {
	std::for_each(begin,end,[&combiner,tuple=std::forward_as_tuple<Args...>(std::forward<Args>(args)...)](auto f)
		      { asynchronous::signals::invoke::apply(f,tuple); combiner(); });
	return combiner;
      }

      template < typename Combiner, typename Iterator, typename ... Args >
      auto invoke (std::true_type, Combiner combiner, Iterator begin, Iterator end, Args&& ... args) const
      {
	std::for_each(begin,end,[&combiner,tuple=std::forward_as_tuple<Args...>(std::forward<Args>(args)...)](auto f)
		      { asynchronous::signals::invoke::apply(f,tuple); combiner(); });
	return combiner;
      }
    };

  } // namespace utility

} // namespace graph_model

#endif
