/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_ABS_HPP
#define GRAPH_MODEL_VERTEX_ABS_HPP 1

#include <cmath>
#include <graph_model/vertex/expression_function.hpp>
#include <graph_model/vertex/expression_terminal.hpp>
#include <mpl/enable_constructor.hpp>

namespace graph_model
{

  namespace vertex
  {

    /// \brief Absolute value function
    /// \tparam Expression The type of expression to evaluate as the argument
    ///
    template < typename Expression >
    class abs_function
    {
      using self = abs_function;

    public:
      template < typename Expr
		 , typename = mpl::enable_constructor_t<self,Expr>
		 >
      explicit abs_function (Expr&& expr)
	: expression_(std::forward<Expr>(expr))
      {}

      template < typename ... T >
      auto operator () (T&& ... t) const
      {
	using std::abs;
	return abs(expression_(std::forward<T>(t)...));
      }

    private:
      Expression expression_;
    };

    template < typename ... T, typename ... Args >
    auto transform_expression (expression_terminal<abs_function<T...>> expr, Args&& ... args)
    { return evaluate_terminal(expr,std::forward<Args>(args)...); }

    template < typename ... T, typename ... Args >
    auto transform_expression (expression_terminal<abs_function<T...>&> expr, Args&& ... args)
    { return evaluate_terminal(expr,std::forward<Args>(args)...); }

    /// \brief Construct a absolute value function
    ///
    /// This function constructs a abs_function object from an \c
    /// expression_function.
    ///
    template < typename Wrapper >
    auto abs (expression_function<Wrapper>&& function)
    {
      using function_type = expression_function<Wrapper>;
      using abs_function_type = abs_function<function_type>;
      return make_expression_function(abs_function_type(std::move(function)));
    }

    /// \brief Construct a absolute value function
    /// \tparam Expr The type of argument
    ///
    /// This function constructs an \c expression_function
    /// representing the absolute value function.
    ///
    template < typename Expr >
    auto abs (Expr&& expr)
    { return abs(make_expression_function(std::forward<Expr>(expr))); }

    template < typename Expression >
    auto abs (abs_function<Expression> const& expr) { return expr; }

    template < typename Expression >
    auto abs (abs_function<Expression>&& expr) { return std::move(expr); }

  } // namespace vertex

} // namespace graph_model

#endif
