/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_DISTRIBUTION_OBSERVER_BASE_HPP
#define GRAPH_MODEL_VERTEX_DISTRIBUTION_OBSERVER_BASE_HPP 1

#include <stdexcept>
#include <boost/exception/exception.hpp>
#include <boost/throw_exception.hpp>
#include <observer_ptr.hpp>
#include <type_traits/is_rvalue.hpp>

namespace graph_model
{

  namespace vertex
  {

    /// \brief Manage a distribution observer
    ///
    /// Some vertex classes delegate their member functions to an
    /// observed distribution.  This class manages the observer and
    /// allows reseting the observed distribution at runtime.  It is
    /// the caller's responsibility to set a distribution prior to
    /// use.
    ///
    template < typename Derived, typename Distribution >
    class distribution_observer_base
    {
      using observer_type = observer_ptr<Distribution const>;

      struct missing_distribution
	: std::logic_error
	, virtual boost::exception
      {
	missing_distribution ()
	  : std::logic_error("missing distribution")
	{}
      };

    public:			// XXX - for CRTP only
      auto& distribution () const
      {
	if (!observer_)
	  BOOST_THROW_EXCEPTION(missing_distribution());
	return *observer_;
      }

    public:
      auto& reset_distribution ()
      {
	if (observer_)
	  derived().remove_vertex(observer_);
	observer_.reset();
	return derived();
      }

    protected:
      template < typename E >
      auto reset_distribution (E const& distribution)
      {
	reset_distribution();
	auto observer = derived().add_adjacent_vertex(distribution);
	observer_ = observer;
	return observer;
      }

      template < typename E
		 , typename = std::enable_if_t<::type_traits::is_rvalue_v<E>>
		 >
      auto reset_distribution (E&& distribution)
      {
	reset_distribution();
	auto observer = derived().add_adjacent_vertex(std::move(distribution));
	observer_ = observer;
	return observer;
      }

    public:
      template < typename E >
      auto distributed_as (E&& distribution)
      { return reset_distribution(std::forward<E>(distribution)); }

    private:
      auto& derived () { return static_cast<Derived&>(*this); }
      auto& derived () const { return static_cast<Derived const&>(*this); }

    private:
      observer_type observer_;
    };

  } // namespace vertex

} // namespace graph_model

#endif
