/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_EXPRESSION_FUNCTION_HPP
#define GRAPH_MODEL_VERTEX_EXPRESSION_FUNCTION_HPP 1

#include <graph_model/vertex/expression_function_fwd.hpp>
#include <boost/operators.hpp>
#include <boost/yap/algorithm.hpp>
#include <graph_model/functional/equal_to.hpp>
#include <graph_model/functional/less.hpp>
#include <graph_model/vertex/expression_terminal.hpp>
#include <graph_model/vertex/rvalue_reference_wrapper.hpp>
#include <graph_model/vertex/vertex_expression.hpp>
#include <mpl/enable_constructor.hpp>

#include <iosfwd>
#include <debugging/class_name.hpp>
// #include <debugging/object_tracker_base.hpp>

namespace graph_model
{

  namespace vertex
  {

    /// \brief Encapsulate a wrapped expression as a function object
    /// \tparam Wrapper The type of wrapper around the expression to encapsulate
    ///
    /// This class is inspired by the Yap class of the same name.
    /// However, it encapsulates a wrapped expression instead of an
    /// expression so that lvalue and rvalue expressions can be
    /// handled uniformly.  Typically, the expression wrapper will be
    /// a \c std::reference_wrapper or an \c rvalue_reference_wrapper.
    ///
    template < typename Wrapper >
    class expression_function
      : boost::totally_ordered<expression_function<Wrapper>>
    // : public debugging::object_tracker_base<expression_function<ExprWrapper>>
    {
      using self = expression_function;

    public:
      expression_function () = default;

      template < typename W
		 , typename = mpl::enable_constructor_t<self,W>
		 >
      expression_function (W&& wrapper)
	: wrapper_(std::forward<W>(wrapper))
      {}

      auto operator () () const
      {
	return ::boost::yap::evaluate(expr());
      }

      auto& expr () const { return wrapper_.get(); }

      friend auto operator == (self left, self right)
      {
	using equal_to = functional::equal_to<Wrapper>;
	return equal_to()(left.wrapper_,right.wrapper_);
      }

      friend auto operator < (self left, self right)
      {
	using less = functional::less<Wrapper>;
	return less()(left.wrapper_,right.wrapper_);
      }

    private:      
      Wrapper wrapper_;
    };

    template < typename ... T, typename ... Args >
    auto transform_expression (expression_terminal<expression_function<T...>> expr, Args&& ... args)
    { return evaluate_terminal(expr,std::forward<Args>(args)...); }
					   
    template < typename ... T, typename ... Args >
    auto transform_expression (expression_terminal<expression_function<T...>&> expr, Args&& ... args)
    { return evaluate_terminal(expr,std::forward<Args>(args)...); }
					   
    template < typename Wrapper >
    std::ostream& operator << (std::ostream& os, expression_function<Wrapper>)
    { return os << debugging::full_class_name::apply<expression_function<Wrapper>>(); }

    /// \brief Construct an expression_function
    /// \tparam Expr The type of wrapped expression to be encapsulated
    ///
    /// This function constructs an expression_function object from an
    /// expression wrapped in a \c std::reference_wrapper object.
    ///
    template < typename Expr >
    auto make_expression_function (std::reference_wrapper<Expr> wrapper)
    {
      static_assert(boost::yap::is_expr<Expr>::value,
		    "make_expression_function() is only defined for expressions.");
      using wrapper_type = std::reference_wrapper<Expr>;
      using function_type = expression_function<wrapper_type>;
      return function_type(wrapper);
    }

    /// \brief Construct an expression_function
    /// \tparam Expr The type of wrapped expression to be encapsulated
    ///
    /// This function constructs an expression_function object from an
    /// expression wrapped in a \c rvalue_reference_wrapper object.
    ///
    template < typename Expr >
    auto make_expression_function (rvalue_reference_wrapper<Expr> const& wrapper)
    {
      static_assert(boost::yap::is_expr<Expr>::value,
		    "make_expression_function() is only defined for expressions.");
      using wrapper_type = rvalue_reference_wrapper<Expr>;
      using function_type = expression_function<wrapper_type>;
      return function_type(wrapper);
    }

    /// \brief Construct an expression_function
    /// \tparam Expr The type of wrapped expression to be encapsulated
    ///
    /// This function constructs an expression_function object from an
    /// expression wrapped in a \c rvalue_reference_wrapper object.
    ///
    template < typename Expr >
    auto make_expression_function (rvalue_reference_wrapper<Expr>&& wrapper)
    {
      static_assert(boost::yap::is_expr<Expr>::value,
		    "make_expression_function() is only defined for expressions.");
      using wrapper_type = rvalue_reference_wrapper<Expr>;
      using function_type = expression_function<wrapper_type>;
      return function_type(std::move(wrapper));
    }

    /// \brief Construct an expression_function
    /// \tparam Expr The type of expression encapsulated by the expression_function object
    ///
    /// This function constructs an expression_function object from an
    /// expression.  Note that \c Expr will be converted to an \c
    /// expression_function and therefore may be any type; it may also
    /// be either an lvalue or an rvalue.
    ///
    template < typename Expr >
    auto make_expression_function (Expr&& expr)
    {
      using std::cref;
      using ::graph_model::vertex::cref;
      using boost::yap::as_expr;
      return vertex::make_expression_function
	(cref(as_expr<vertex_expression>(std::forward<Expr>(expr))));
    }

    template < typename ExprWrapper >
    auto make_expression_function (expression_function<ExprWrapper> expr)
    { return expr; }

  } // namespace vertex

} // namespace graph_model

#endif
