/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_MARKOV_BLANKET_HPP
#define GRAPH_MODEL_VERTEX_MARKOV_BLANKET_HPP 1

#include <boost/yap/algorithm.hpp>
#include <graph_model/concept/reference_wrapper_concept.hpp>
#include <graph_model/vertex/rvalue_reference_wrapper.hpp>
#include <graph_model/vertex/vertex_expression.hpp>
#include <mpl/enable_constructor.hpp>

namespace graph_model
{

  namespace vertex
  {

    template < typename Wrapper >
    class markov_blanket_wrapper
    {
      BOOST_CONCEPT_ASSERT((concept::reference_wrapper_concept<Wrapper>));

      using self = markov_blanket_wrapper;

    public:
      template < typename W
		 , typename = mpl::enable_constructor_t<self,W>
		 >
      explicit markov_blanket_wrapper (W&& wrapper)
	: wrapper_(std::forward<W>(wrapper))
      {}

      template < typename ... T >
      auto operator () (T&& ... t) const
      {
	return wrapper_.get().markov_blanket(std::forward<T>(t)...);
      }

    private:
      Wrapper wrapper_;
    };

    template < typename Expr >
    auto make_markov_blanket_wrapper (std::reference_wrapper<Expr>&& wrapper)
    {
      using wrapper_type = std::reference_wrapper<Expr>;
      using function_type = markov_blanket_wrapper<wrapper_type>;
      return function_type(wrapper);
    }

    template < typename Expr >
    auto make_markov_blanket_wrapper (rvalue_reference_wrapper<Expr>&& wrapper)
    {
      using wrapper_type = rvalue_reference_wrapper<Expr>;
      using function_type = markov_blanket_wrapper<wrapper_type>;
      return function_type(wrapper);
    }

    template < typename Expr >
    auto make_markov_blanket_wrapper (Expr&& expr)
    {
      using std::cref;
      return make_markov_blanket_wrapper(cref(std::forward<Expr>(expr)));
    }

    template < typename Expr >
    auto markov_blanket (Expr&& expr)
    {
      using boost::yap::make_expression;
      using boost::yap::expr_kind;
      return make_expression<vertex_expression,expr_kind::call>
	(make_markov_blanket_wrapper(std::forward<Expr>(expr)));
    }

    template < typename Wrapper >
    auto markov_blanket (markov_blanket_wrapper<Wrapper> markov_blanket) { return markov_blanket; }

  } // namespace vertex

} // namespace graph_model

#endif
