/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_MEAN_FACTOR_HPP
#define GRAPH_MODEL_VERTEX_MEAN_FACTOR_HPP 1

#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>
#include <graph_model/vertex/mean_factor_fwd.hpp>
#include <asynchronous/signals/mean_combiner.hpp>
#include <graph_model/signals/tags.hpp>
#include <graph_model/type/signal.hpp>
#include <graph_model/vertex/expression_terminal.hpp>
#include <graph_model/vertex/expression_function.hpp>
#include <graph_model/vertex/factor_base.hpp>
#include <graph_model/vertex/factor_impl_base.hpp>
#include <graph_model/vertex/strategy/basic_factor.hpp>
#include <graph_model/vertex/strategy_owner_base.hpp>
#include <graph_model/vertex/tags.hpp>
#include <logging/channel_log.hpp>
#include <logging/global_log.hpp>
#include <mpl/enable_constructor.hpp>
#include <use_default.hpp>

namespace graph_model
{

  namespace vertex
  {

    /// \brief Accumulate the mean of a set of factors
    ///
    /// The \c mean_factor class is itself a factor that accumulates
    /// the mean of a set of other factors.
    ///
    template < typename Probability >
    class mean_factor
      : public factor_impl_base<mean_factor<Probability>,tag::factor,factor_base<Probability>>
      , public strategy_owner_base<mean_factor<Probability>,strategy::basic_factor_factory<Probability>>
    {
      using self = mean_factor;
      using factor_base = factor_base<Probability>;
      using base = factor_impl_base<self,tag::factor,factor_base>;
      using combiner_type = asynchronous::signals::mean_combiner<Probability>;
      using signal_tag = signals::tag::mean_factor;
      // using cutoff_policy = asynchronous::signals::constant_cutoff_policy;
      // using signal_type = boost::signals2::signal<Probability(),combiner_type>;
      // using cpu_load_policy = asynchronous::signals::default_save_cpu_load<10,1000,1000>;
      // using threadpool_policy = asynchronous::signals::shared_threadpool<cpu_load_policy>;
      using signal_type = ::graph_model::type::signal_t<Probability(),combiner_type,signal_tag>;
      using log_type = logging::channel_log;

    public:
      template < typename ... Args
		 , typename = mpl::enable_constructor_t<self,Args...>
		 >
      explicit mean_factor (Args&& ... args)
	: signal_(std::forward<Args>(args)...)
      {}

    public:
      template < typename Factor >
      auto& add_factor (Factor&& factor)
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "add_factor<Factor>(Factor&&) [" << this << "]";
	connect(std::forward<Factor>(factor));
	return *this;
      }

      template < typename Factor >
      auto connect (Factor&& factor)
      { return signal_.connect(make_expression_function(std::forward<Factor>(factor))); }

    private:
      virtual typename base::result_type eval_ () const override
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "eval_()const override [" << this << "]";
	auto result = this->pdf();
	GLOBAL_LOG(log_,trace) << "eval_()const override [" << this << "]: result=" << result;
	return result;
      }
      virtual typename base::probability_type pdf_ () const override // { return signal_(); }
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "pdf_()const override [" << this << "]";
	auto pdf = signal_().result();
	GLOBAL_LOG(log_,trace) << "pdf_()const override [" << this << "]: pdf=" << pdf;
	return pdf;
      }

    private:
      signal_type signal_;
      mutable log_type log_ = "graph_model::vertex::mean_factor";
    };

    template < typename ... T, typename ... Args >
    auto transform_expression (expression_terminal<mean_factor<T...>> expr, Args&& ... args)
    { return evaluate_terminal(expr,std::forward<Args>(args)...); }
					   
    template < typename ... T, typename ... Args >
    auto transform_expression (expression_terminal<mean_factor<T...>&> expr, Args&& ... args)
    { return evaluate_terminal(expr,std::forward<Args>(args)...); }

    template < typename Probability, typename ... Args >
    auto make_mean_factor (Args&& ... args)
    {
      using mean_factor_type = mean_factor<Probability>;
      return mean_factor_type(std::forward<Args>(args)...);
    }

  } // namespace vertex

} // namespace graph_model

#endif
