/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_POW_HPP
#define GRAPH_MODEL_VERTEX_POW_HPP 1

#include <cmath>
#include <graph_model/vertex/expression_function.hpp>
#include <graph_model/vertex/expression_terminal.hpp>

namespace graph_model
{

  namespace vertex
  {

    /// \brief Power function
    /// \tparam Expression The type of expression to evaluate as the argument
    ///
    template < typename Base, typename Exponent >
    class pow_function
    {
      using self = pow_function;

    public:
      template < typename B, typename E >
      explicit pow_function (B&& base, E&& exponent)
	: base_(std::forward<B>(base))
	, exponent_(std::forward<E>(exponent))
      {}

      template < typename ... T >
      auto operator () (T&& ... t) const
      {
	using std::pow;
	// std::cout << "graph_model::vertex::pow_function<Base,Exponent>::operator()()const" << std::endl;
	auto result = pow(base_(std::forward<T>(t)...),exponent_(std::forward<T>(t)...));
	// std::cout << "graph_model::vertex::pow_function<Base,Exponent>::operator()()const: base=" << base_() << ", exponent=" << exponent_() << ", result=" << result << std::endl;
	return result;
      }

    private:
      Base base_;
      Exponent exponent_;
    };

    template < typename ... T, typename ... Args >
    auto transform_expression (expression_terminal<pow_function<T...>> expr, Args&& ... args)
    { return evaluate_terminal(expr,std::forward<Args>(args)...); }

    template < typename ... T, typename ... Args >
    auto transform_expression (expression_terminal<pow_function<T...>&> expr, Args&& ... args)
    { return evaluate_terminal(expr,std::forward<Args>(args)...); }

    /// \brief Construct a power function
    ///
    /// This function constructs a pow_function object from an \c
    /// expression_function.
    ///
    template < typename BaseWrapper, typename ExponentWrapper >
    auto pow (expression_function<BaseWrapper>&& base, expression_function<ExponentWrapper>&& exponent)
    {
      using base_function_type = expression_function<BaseWrapper>;
      using exponent_function_type = expression_function<ExponentWrapper>;
      using pow_function_type = pow_function<base_function_type,exponent_function_type>;
      return make_expression_function(pow_function_type(std::move(base),std::move(exponent)));
    }

    /// \brief Construct a power function
    /// \tparam Base The type of the base
    /// \tparam Exponent The type of the exponent
    ///
    /// This function constructs an \c expression_function
    /// representing the power function.
    ///
    template < typename Base, typename Exponent >
    auto pow (Base&& base, Exponent&& exponent)
    {
      return pow
	(make_expression_function(std::forward<Base>(base)),
	 make_expression_function(std::forward<Exponent>(exponent)));
    }

    template < typename Base, typename Exponent >
    auto pow (pow_function<Base,Exponent> const& expr) { return expr; }

    template < typename Base, typename Exponent >
    auto pow (pow_function<Base,Exponent>&& expr) { return std::move(expr); }

  } // namespace vertex

} // namespace graph_model

#endif
