/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_RVALUE_REFERENCE_WRAPPER_HPP
#define GRAPH_MODEL_VERTEX_RVALUE_REFERENCE_WRAPPER_HPP 1

#include <functional>
#include <iostream>
#include <memory>
#include <type_traits>
#include <graph_model/functional/equal_to.hpp>
#include <graph_model/functional/less.hpp>
#include <type_traits/is_rvalue.hpp>

#include <debugging/stack_trace.hpp>

namespace graph_model
{

  namespace vertex
  {

    namespace detail
    {

      template < typename T >
      class basic_rvalue_reference_wrapper
      {
	using self = basic_rvalue_reference_wrapper;
	using value_type = std::decay_t<T>;
	using const_lvalue_reference = value_type const&;
	using mutable_rvalue_reference = value_type&&;
	using pointer = std::shared_ptr<value_type>;

      public:
	using type = T;

	basic_rvalue_reference_wrapper () = delete;
	explicit basic_rvalue_reference_wrapper (const_lvalue_reference t)
	  : p_(initialize(t))
	{}
	explicit basic_rvalue_reference_wrapper (mutable_rvalue_reference t)
	  : p_(initialize(std::move(t)))
	{}

	self& operator = (const_lvalue_reference t) { p_ = initialize(t); return *this; }
	self& operator = (mutable_rvalue_reference t) { p_ = initialize(std::move(t)); return *this; }

	T& get () const { is_ok(); return *p_; }

	operator T& () const { return get(); }

	template < typename ... Args >
	auto operator () (Args&& ... args) const
	// XXX - c++17 { return std::invoke(get(),std::forward<Args>(args)...); }
	{ return get()(std::forward<Args>(args)...); }

      private:
	static pointer initialize (const_lvalue_reference t)
	{ return std::make_shared<value_type>(t); }
	static pointer initialize (mutable_rvalue_reference t)
	{ return std::make_shared<value_type>(std::move(t)); }

	void is_ok () const
	{
	  if (!p_)
	    {
	      std::cout << __FILE__ << ":" << __LINE__ << ":  ERROR: rvalue_reference_wrapper is null" << std::endl;
	      std::cout << debugging::stack_trace<>() << std::endl;
	      assert(p_);
	    }
	}

      private:
	pointer p_;
      };

    } // namespace detail

    /// \brief Wrap an rvalue reference
    /// \tparam T The type of object to wrap
    ///
    /// An rvalue_reference_wrapper parallels \c
    /// std::reference_wrapper<T>: it is copyable and provides the
    /// same object access and function call interfaces.  However,
    /// unlike std::reference_wrapper<T> which explicitly disavows
    /// ownership of the wrapped object and cannot be constructed from
    /// an rvalue reference, an rvalue_reference_wrapper explicitly
    /// owns the wrapped object and can be constructed from an rvalue
    /// reference.
    ///
    /// Reference wrapper objects may be constructed using the ref()
    /// and cref() functions, which parallel the std::ref() and
    /// std::cref() functions.  When called with lvalue references,
    /// these functions delegate to their standard counterparts and
    /// therefore construct std::reference_wrapper<T> objects.  When
    /// called with rvalue references, however, they construct
    /// rvalue_reference_wrapper<T> objects.  Note that the latter
    /// overloads are deleted for the standard functions, but are
    /// useful in this context.
    ///
    template < typename T >
    class rvalue_reference_wrapper
      : public detail::basic_rvalue_reference_wrapper<T>
    {
      using base = detail::basic_rvalue_reference_wrapper<T>;

    public:
      rvalue_reference_wrapper () = delete;
      template < typename U >
      rvalue_reference_wrapper (U&& u)
	: base(std::forward<U>(u))
      {}
    };

    /// Construct a std::reference_wrapper<T> object from another one
    ///
    template < typename T >
    auto ref (std::reference_wrapper<T> const& wrapper) { return wrapper; }

    /// Construct a std::reference_wrapper<T> object from another one
    ///
    template < typename T >
    auto ref (std::reference_wrapper<T>&& wrapper) { return std::move(wrapper); }

    /// Construct a rvalue_reference_wrapper<T> object from an rvalue
    ///
    /// In the case of std::reference_wrapper<T>, this function is
    /// deleted.
    ///
    template < typename T
	       , typename = std::enable_if_t<::type_traits::is_rvalue_v<T>>
	       >
    auto ref (T&& t)
    {
      using reference_wrapper_type = rvalue_reference_wrapper<std::decay_t<T>>;
      return reference_wrapper_type(std::forward<T>(t));
    }

    /// Construct a rvalue_reference_wrapper<T> object from another one
    ///
    template < typename T >
    auto ref (rvalue_reference_wrapper<T> wrapper)
    { return wrapper; }

    /// Construct a std::reference_wrapper<T const> object from another one
    ///
    template < typename T >
    auto cref (std::reference_wrapper<T const> wrapper) { return wrapper; }

    /// Construct a rvalue_reference_wrapper<T const> object from an rvalue
    ///
    /// In the case of std::reference_wrapper<T const>, this function
    /// is deleted.
    ///
    template < typename T
	       , typename = std::enable_if_t<::type_traits::is_rvalue_v<T>>
	       >
		   auto cref (T&& t)
    {
      using reference_wrapper_type = rvalue_reference_wrapper<std::decay_t<T> const>;
      return reference_wrapper_type(std::forward<T>(t));
    }

    /// Construct a rvalue_reference_wrapper<T const> object from another one
    ///
    template < typename T >
    auto cref (rvalue_reference_wrapper<T const> wrapper)
    { return wrapper; }

  } // namespace vertex

  namespace functional
  {

    template < typename T >
    struct equal_to< vertex::rvalue_reference_wrapper<T> >
    {
      using first_argument_type = vertex::rvalue_reference_wrapper<T>;
      using second_argument_type = vertex::rvalue_reference_wrapper<T>;
      using result_type = bool;

      result_type operator () (first_argument_type const& left, second_argument_type const& right) const
      { return &left.get() == &right.get(); }
    };

    template < typename T >
    struct less< vertex::rvalue_reference_wrapper<T> >
    {
      using first_argument_type = vertex::rvalue_reference_wrapper<T>;
      using second_argument_type = vertex::rvalue_reference_wrapper<T>;
      using result_type = bool;

      result_type operator () (first_argument_type const& left, second_argument_type const& right) const
      { return &left.get() < &right.get(); }
    };

  } // namespace functional

} // namespace graph_model

#endif
