/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_STRATEGY_BASIC_FACTOR_HPP
#define GRAPH_MODEL_VERTEX_STRATEGY_BASIC_FACTOR_HPP 1

#include <cassert>
#include <memory>
#include <graph_model/concept/factor_policy_concept.hpp>
#include <graph_model/vertex/strategy/factor_base.hpp>
#include <graph_model/vertex/strategy/policy_owner_base.hpp>
#include <graph_model/vertex/trapping_eval_base.hpp>
#include <graph_model/vertex/trapping_pdf_base.hpp>

namespace graph_model
{

  namespace vertex
  {

    namespace strategy
    {

      template < typename Policy
		 , typename Probability
		 >
      class basic_factor
	: public factor_base<Probability>
	, policy_owner_base<basic_factor<Policy,Probability>,Policy>
      {
	using self = basic_factor;
	using base = factor_base<Probability>;
	using policy_owner_base = policy_owner_base<basic_factor<Policy,Probability>,Policy>;

	BOOST_CONCEPT_ASSERT((concept::factor_policy_concept<Policy,Probability>));

      public:
	basic_factor () = default;
	explicit basic_factor (Policy const& policy) : policy_owner_base(policy) {}
	explicit basic_factor (Policy&& policy) : policy_owner_base(std::move(policy)) {}

      private:
	typename base::result_type
	eval_ () const override { return this->pdf(); }
	typename base::probability_type
	pdf_ () const override { return this->policy().pdf(); }
      };

      template < typename Probability >
      class basic_factor< void,Probability >
	: public trapping_eval_base< basic_factor<void,Probability>,
	  trapping_pdf_base< basic_factor<void,Probability>,
	  factor_base<Probability
	  > > >
      {};

      template < typename Probability >
      auto make_basic_factor_policy_pointer ()
      {
	using basic_factor_type = basic_factor<void,Probability>;
	return std::make_unique<basic_factor_type>();
      }

      template < typename Probability, typename Policy >
      auto make_basic_factor_policy_pointer (Policy&& policy)
      {
	using policy_type = std::decay_t<Policy>;
	using basic_factor_type = basic_factor<policy_type,Probability>;
	return std::make_unique<basic_factor_type>(std::forward<Policy>(policy));
      }

      template < typename Probability >
      struct basic_factor_factory
      {
	using pointer = std::unique_ptr<factor_base<Probability>>;

	static pointer apply ()
	{ return make_basic_factor_policy_pointer<Probability>(); }

	template < typename Policy >
	static pointer apply (Policy&& policy)
	{ return make_basic_factor_policy_pointer<Probability>(std::forward<Policy>(policy)); }
      };

    } // namespace strategy

  } // namespace vertex

} // namespace graph_model

#endif
