/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_SUMMARIZED_SCALAR_RANDOM_VARIABLE_HPP
#define GRAPH_MODEL_VERTEX_SUMMARIZED_SCALAR_RANDOM_VARIABLE_HPP 1

#include <graph_model/vertex/summarized_scalar_random_variable_fwd.hpp>
#include <boost/yap/algorithm.hpp>
#include <graph_model/vertex/distribution_base.hpp>
#include <graph_model/vertex/expression_terminal.hpp>
#include <graph_model/vertex/scalar_distribution_base.hpp>
#include <graph_model/vertex/scalar_distribution_observer_base.hpp>
#include <graph_model/vertex/summarized_scalar_random_variable_base.hpp>
#include <graph_model/vertex/summarized_scalar_random_variable_impl_base.hpp>
#include <graph_model/vertex/tags.hpp>
#include <graph_model/vertex/vertex_expression.hpp>
#include <mpl/enable_constructor.hpp>
#include <use_default.hpp>

namespace graph_model
{

  namespace vertex
  {

    /// A concrete summarized scalar_random variable vertex
    ///
    /// A summarized scalar random variable accumulates its own statistics
    /// and hence obviates the need for a separate accumulator object.
    ///
    template < typename Variate, typename Probability, typename Stats >
    struct summarized_scalar_random_variable
      : summarized_scalar_random_variable_impl_base< summarized_scalar_random_variable<Variate,Probability,Stats>
						     , tag::summarized_scalar_random_variable
						     , Variate
						     , Probability
						     , Stats
						     , summarized_scalar_random_variable_base<Variate,Probability>
						     >
      , scalar_distribution_observer_base< summarized_scalar_random_variable<Variate,Probability,Stats>
					   , scalar_distribution_base<Variate,Probability>
					   >
    {};

    template < typename ... T, typename ... Args >
    auto transform_expression (expression_terminal<summarized_scalar_random_variable<T...>> expr, Args&& ... args)
    { return evaluate_terminal(expr,std::forward<Args>(args)...); }

    template < typename ... T, typename ... Args >
    auto transform_expression (expression_terminal<summarized_scalar_random_variable<T...>&> expr, Args&& ... args)
    { return evaluate_terminal(expr,std::forward<Args>(args)...); }
					   
    template < typename Variate, typename Probability, typename Stats, typename ... Args >
    auto make_summarized_scalar_random_variable (Args&& ... args)
    {
      using summarized_scalar_random_variable_type = summarized_scalar_random_variable<Variate,Probability,Stats>;
      return summarized_scalar_random_variable_type(std::forward<Args>(args)...);
    }

    template < typename Variate, typename Probability, typename ... Args >
    auto make_summarized_scalar_random_variable (Args&& ... args)
    {
      using summarized_scalar_random_variable_type = summarized_scalar_random_variable<Variate,Probability>;
      return summarized_scalar_random_variable_type(std::forward<Args>(args)...);
    }

    template < typename Probability, typename Stats, typename Variate, typename ... Args >
    auto make_summarized_scalar_random_variable (Variate variate, Args&& ... args)
    {
      using summarized_scalar_random_variable_type = summarized_scalar_random_variable<Variate,Probability,Stats>;
      summarized_scalar_random_variable_type rv(std::forward<Args>(args)...);
      rv.reset_variate(variate);
      return rv;
    }

    template < typename Probability, typename Variate, typename ... Args >
    auto make_summarized_scalar_random_variable (Variate variate, Args&& ... args)
    {
      using summarized_scalar_random_variable_type = summarized_scalar_random_variable<Variate,Probability>;
      summarized_scalar_random_variable_type rv(std::forward<Args>(args)...);
      rv.reset_variate(variate);
      return rv;
    }

    template < typename Variate, typename ... Args >
    auto make_summarized_scalar_random_variable (Args&& ... args)
    { return make_summarized_scalar_random_variable<Variate,double>(std::forward<Args>(args)...); }

    template < typename Variate, typename ... Args >
    auto make_summarized_scalar_random_variable (Variate variate, Args&& ... args)
    { return make_summarized_scalar_random_variable<double>(variate,std::forward<Args>(args)...); }

  } // namespace vertex

} // namespace graph_model

#endif
