/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_VARIATE_OWNER_BASE_HPP
#define GRAPH_MODEL_VERTEX_VARIATE_OWNER_BASE_HPP 1

#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>
#include <logging/channel_log.hpp>
#include <logging/global_log.hpp>
#include <observer_ptr.hpp>

#include <debugging/stack_trace.hpp>

namespace graph_model
{

  namespace vertex
  {

    /// \brief Manage variates
    ///
    /// This base class manages the values of variates, especially for
    /// MCMC algorithms.  For the purposes of iteration, three
    /// variates are managed:
    ///
    /// - Current variate: this variate is the current value for the
    ///   iteration.  Normally, this is the one the the \c variate()
    ///   member function should return.  It will be returned by
    ///   default or if the \c use_current_variate() member function
    ///   is called (by any thread) prior to calling the \c variate()
    ///   member function.
    ///
    /// - Proposed variate: this variate is the proposal (if any)
    ///   provided by the MCMC sampler.  It will be returned if the \c
    ///   use_proposed_variate() member function is called (by any
    ///   thread) prior to calling the \c variate() member function.
    ///
    /// - Next variate: this variate is what will become the value of
    ///   the current variate if it is committed with the \c commit()
    ///   member function.  It is normally updated using either the \c
    ///   accept() or \c reject() member functions.
    ///
    /// A typical MCMC algorithm will interact with this class in the
    /// following way.  First, a variate will be proposed and set with
    /// the \c reset_proposed_variate() member function.  Second, the
    /// Markov blanket of relevant random variables will be evaluated
    /// twice, once for the current variate and once for the proposed
    /// variate; the \c use_current_variate() and \c
    /// use_proposed_variate() member functions will be used to select
    /// the appropriate variate.  Third, the proposed variate will be
    /// accepted or rejected with the \c accept() or \c reject()
    /// member functions.  Finally, the decision will be committed
    /// with the \c commit() member function to complete the
    /// iteration.  The separation between the decision to accept or
    /// reject the proposal and committing the decision is needed to
    /// support parallelism.  In that case, the first part of the
    /// algorithm, i.e., proposal generation and accept/reject, must
    /// avoid conflicts among threads; however, the second part of the
    /// algorithm, i.e., committing the decision, can proceed
    /// independently as there are no possible conflicts.
    ///
    /// \note Thread safety: while retrieving the values of the
    /// variates is thread safe, all threads will receive the same
    /// (current or proposed) variate via the \c variate() member
    /// function.  Which is returned depends on which of the \c
    /// use_current_variate() or \c use_proposed_variate() member
    /// functions were called most recently by any thread.  It is up
    /// to the application to avoid conflicts among threads.
    ///
    template < typename Derived, typename Base >
    class variate_owner_base
      : public Base
    {
      using self = variate_owner_base;
      using base = Base;
      using log_type = logging::channel_log;

    public:
      auto& reset_variate (typename base::variate_type variate)
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "reset_variate(typename base::variate_type): identifier=" << derived().identifier() << ", variate=" << variate;
	reset_current_variate(variate);
	reset_proposed_variate(variate);
	reset_next_variate(variate);
	return derived();
      }
      auto& reset_current_variate (typename base::variate_type variate)
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "reset_current_variate(typename base::variate_type): identifier=" << derived().identifier() << ", variate=" << variate;
	m_current_variate_ = variate;
	return derived();
      }
      auto& reset_proposed_variate (typename base::variate_type variate)
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "reset_proposed_variate(typename base::variate_type): identifier=" << derived().identifier() << ", variate=" << variate;
	m_proposed_variate_ = variate;
	return derived();
      }
    private:
      void reset_next_variate (typename base::variate_type variate)
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "reset_next_variate(typename base::variate_type): identifier=" << derived().identifier() << ", variate=" << variate;
	m_next_variate_ = variate;
      }

    public:
      auto& variate ()
      {
	using namespace boost::log::trivial;
	auto& result = use_current_variate_ ? current_variate() : proposed_variate();
	GLOBAL_LOG(log_,trace) << "variate(): identifier=" << derived().identifier() << ", use_current_variate_=" << std::boolalpha << use_current_variate_ << ", result=" << result;
	return result;
      }
      auto& variate () const
      {
	using namespace boost::log::trivial;
	auto& result = use_current_variate_ ? current_variate() : proposed_variate();
	GLOBAL_LOG(log_,trace) << "variate()const: identifier=" << derived().identifier() << ", use_current_variate_=" << std::boolalpha << use_current_variate_ << ", result=" << result;
	return result;
      }

      auto& current_variate ()
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "current_variate(): identifier=" << derived().identifier() << ", m_current_variate_=" << m_current_variate_;
	return m_current_variate_;
      }
      auto& current_variate () const
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "current_variate()const: identifier=" << derived().identifier() << ", m_current_variate_=" << m_current_variate_;
	return m_current_variate_;
      }

      auto& proposed_variate ()
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "proposed_variate(): identifier=" << derived().identifier() << ", m_proposed_variate_=" << m_proposed_variate_;
	return m_proposed_variate_;
      }
      auto& proposed_variate () const
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "proposed_variate()const: identifier=" << derived().identifier() << ", m_proposed_variate_=" << m_proposed_variate_;
	return m_proposed_variate_;
      }

      void accept ()
      {
	using namespace boost::log::trivial;
	reset_next_variate(m_proposed_variate_);
	GLOBAL_LOG(log_,trace) << "accept(): identifier=" << derived().identifier() << ", m_next_variate_=" << m_next_variate_;
      }
      void reject ()
      {
	using namespace boost::log::trivial;
	reset_next_variate(m_current_variate_);
	GLOBAL_LOG(log_,trace) << "reject(): identifier=" << derived().identifier() << ", m_next_variate_=" << m_next_variate_;
      }
      void commit ()
      {
	using namespace boost::log::trivial;
	using std::swap;
	swap(m_next_variate_,m_current_variate_);
	GLOBAL_LOG(log_,trace) << "commit(): identifier=" << derived().identifier() << ", m_current_variate_=" << m_current_variate_;
	updated_variate_();
      }

      auto& use_current_variate ()
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "use_current_variate(): identifier=" << derived().identifier();
	use_current_variate_ = true;
	return derived();
      }
      auto& use_proposed_variate ()
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "use_proposed_variate(): identifier=" << derived().identifier();
	use_current_variate_ = false;
	return derived();
      }

    private:
      typename base::variate_type variate_ () const override { return variate(); }

    private:
      virtual void updated_variate_ () {}

    private:
      auto& derived () { return static_cast<Derived&>(*this); }
      auto& derived () const { return static_cast<Derived const&>(*this); }

    private:
      bool use_current_variate_ = true;
      typename base::variate_type m_current_variate_;
      typename base::variate_type m_proposed_variate_;
      typename base::variate_type m_next_variate_;
      mutable log_type log_ = "graph_model::vertex::variate_owner_base";
    };

  } // namespace vertex

} // namespace graph_model

#endif
