/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_VERTEX_EXPRESSION_HPP
#define GRAPH_MODEL_VERTEX_VERTEX_EXPRESSION_HPP 1

#include <graph_model/vertex/vertex_expression_fwd.hpp>

namespace graph_model
{

  namespace vertex
  {

    /// \brief Yap expression for vertices
    /// \tparam Kind The kind of expression
    /// \tparam Tuple The tuple of child expressions
    ///
    /// A \c vertex_expression encapsulates an expression of vertices
    /// for later evaluation by Yap customization points.  The class
    /// defines a set of operations using macros provided by the Yap
    /// library.
    ///
    template < boost::yap::expr_kind Kind, typename Tuple >
    class vertex_expression
    {
      using this_kind = boost::yap::expr_kind;

    public:
      // unary operators
      BOOST_YAP_USER_UNARY_OPERATOR_MEMBER(unary_plus, ::graph_model::vertex::vertex_expression)
      BOOST_YAP_USER_UNARY_OPERATOR_MEMBER(negate, ::graph_model::vertex::vertex_expression)

      // arithmetic operators
      BOOST_YAP_USER_BINARY_OPERATOR_MEMBER(plus, ::graph_model::vertex::vertex_expression)
      BOOST_YAP_USER_BINARY_OPERATOR_MEMBER(minus, ::graph_model::vertex::vertex_expression)
      BOOST_YAP_USER_BINARY_OPERATOR_MEMBER(multiplies, ::graph_model::vertex::vertex_expression)
      BOOST_YAP_USER_BINARY_OPERATOR_MEMBER(divides, ::graph_model::vertex::vertex_expression)
      BOOST_YAP_USER_BINARY_OPERATOR_MEMBER(modulus, ::graph_model::vertex::vertex_expression)

      // logical comparison operators
      BOOST_YAP_USER_BINARY_OPERATOR_MEMBER(equal_to, ::graph_model::vertex::vertex_expression)
      BOOST_YAP_USER_BINARY_OPERATOR_MEMBER(not_equal_to, ::graph_model::vertex::vertex_expression)
      BOOST_YAP_USER_BINARY_OPERATOR_MEMBER(less, ::graph_model::vertex::vertex_expression)
      BOOST_YAP_USER_BINARY_OPERATOR_MEMBER(less_equal, ::graph_model::vertex::vertex_expression)
      BOOST_YAP_USER_BINARY_OPERATOR_MEMBER(greater_equal, ::graph_model::vertex::vertex_expression)
      BOOST_YAP_USER_BINARY_OPERATOR_MEMBER(greater, ::graph_model::vertex::vertex_expression)

      // logical operators
      BOOST_YAP_USER_BINARY_OPERATOR_MEMBER(logical_and, ::graph_model::vertex::vertex_expression)
      BOOST_YAP_USER_BINARY_OPERATOR_MEMBER(logical_or, ::graph_model::vertex::vertex_expression)
      BOOST_YAP_USER_UNARY_OPERATOR_MEMBER(logical_not, ::graph_model::vertex::vertex_expression)
      
    public:			// XXX - for yap access
      static const this_kind kind = Kind;
      Tuple elements;
    };

    // arithmetic operators
    BOOST_YAP_USER_FREE_BINARY_OPERATOR(plus, ::graph_model::vertex::vertex_expression)
    BOOST_YAP_USER_FREE_BINARY_OPERATOR(minus, ::graph_model::vertex::vertex_expression)
    BOOST_YAP_USER_FREE_BINARY_OPERATOR(multiplies, ::graph_model::vertex::vertex_expression)
    BOOST_YAP_USER_FREE_BINARY_OPERATOR(divides, ::graph_model::vertex::vertex_expression)
    BOOST_YAP_USER_FREE_BINARY_OPERATOR(modulus, ::graph_model::vertex::vertex_expression)

    // logical comparison operators
    BOOST_YAP_USER_FREE_BINARY_OPERATOR(equal_to, ::graph_model::vertex::vertex_expression)
    BOOST_YAP_USER_FREE_BINARY_OPERATOR(not_equal_to, ::graph_model::vertex::vertex_expression)
    BOOST_YAP_USER_FREE_BINARY_OPERATOR(less, ::graph_model::vertex::vertex_expression)
    BOOST_YAP_USER_FREE_BINARY_OPERATOR(less_equal, ::graph_model::vertex::vertex_expression)
    BOOST_YAP_USER_FREE_BINARY_OPERATOR(greater_equal, ::graph_model::vertex::vertex_expression)
    BOOST_YAP_USER_FREE_BINARY_OPERATOR(greater, ::graph_model::vertex::vertex_expression)

    // logical operators
    BOOST_YAP_USER_FREE_BINARY_OPERATOR(logical_and, ::graph_model::vertex::vertex_expression)
    BOOST_YAP_USER_FREE_BINARY_OPERATOR(logical_or, ::graph_model::vertex::vertex_expression)

    // ternary operator
    BOOST_YAP_USER_EXPR_IF_ELSE(::graph_model::vertex::vertex_expression)

  } // namespace vertex

} // namespace graph_model

#endif
