/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef LOGGING_INITIALIZE_TRACE_LOG_HPP
#define LOGGING_INITIALIZE_TRACE_LOG_HPP 1

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/log/attributes.hpp>
#include <boost/log/trivial.hpp>
#include <boost/log/utility/setup/common_attributes.hpp>
#include <boost/log/utility/setup/console.hpp>
#include <boost/log/utility/setup/file.hpp>

namespace logging
{

  /// \brief Initialize a Boost.Log sink
  ///
  template < typename Severity = boost::log::trivial::severity_level >
  class initialize_trace_log
  {
    using self = initialize_trace_log;

  public:
    template < typename ... Args >
    static void apply (std::ostream& os, Args&& ... args)
    {
      auto sink = boost::log::add_console_log(os,std::forward<Args>(args)...);
      sink->set_formatter(trace_formatter());
      boost::log::add_common_attributes();
    }
    
    template < typename ... Args >
    static void apply (Args&& ... args)
    {
      auto sink = boost::log::add_file_log(std::forward<Args>(args)...);
      sink->set_formatter(trace_formatter());
      boost::log::add_common_attributes();
    }

  private:
    struct trace_formatter
    {
      template < typename Record, typename Ostream >
      void operator () (Record const& record, Ostream& os) const
      {
	using boost::log::extract;
	auto fullpath = extract<std::string>("File",record);
	auto line = extract<int>("Line",record);
	auto function = extract<std::string>("Function",record);
	auto timestamp = extract<boost::posix_time::ptime>("TimeStamp",record);
	auto severity = extract<Severity>("Severity",record);
	auto channel = extract<std::string>("Channel",record);
	auto message = record[boost::log::expressions::smessage];
 	os << fullpath << ":" << line << ": "
	  // << "[" << function << "] "
	   << "[" << timestamp << "] "
	   << "[" << severity << "] "
	   << "[" << channel << "] "
	   << message;
      }
    };
  };

} // namespace logging

#endif
