/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef MPL_ENABLE_CONSTRUCTOR_HPP
#define MPL_ENABLE_CONSTRUCTOR_HPP 1

#include <type_traits>

namespace mpl
{

  namespace detail
  {

    template < typename ... T >
    class enable_constructor;

    template < typename T >
    struct enable_constructor<T>
    {
      using type = std::enable_if< true >;
    };

    template < typename T, typename U >
    struct enable_constructor<T,U>
    {
      using type = std::enable_if< !std::is_base_of<std::decay_t<T>,std::decay_t<U>>::value >;
    };

    template < typename T, typename U, typename V, typename ... W >
    struct enable_constructor<T,U,V,W...>
    {
      using type = std::enable_if< true >;
    };

  } // namespace detail

  template < typename T, typename ... U >
  using enable_constructor = typename detail::enable_constructor<T,U...>::type;

  /// \brief Enable non copy constructor overloads
  ///
  /// Overload resolution for classes that use perfect forwarding for
  /// constructors can easily inadvertently select an overload in
  /// place of the (implicit) copy and move constructors.  Often this
  /// is a mistake.
  ///
  /// This type alias uses SFINAE to remove a template-generated
  /// constructor from the overload set do avoid conflicts with the
  /// copy and move constructors.  It should be used as follows:
  ///
  /// \code
  /// struct S
  /// {
  ///   template < typename ... T
  ///              , typename mpl::enable_constructor_t<S,T...>
  ///              >
  ///   explicit S (T&& ...) {}
  /// };
  /// \endcode
  ///
  /// Note that \c enable_constructor_t may be used with any number of
  /// template arguments greater than one.  For any number other than
  /// two, it has no effect because the function cannot be a
  /// substitute for the copy or move constructors.
  ///
  template < typename T, typename ... U >
  using enable_constructor_t = typename enable_constructor<T,U...>::type;

} // namespace mpl

#endif
