/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef SAMPLE_SPACE_MATH_CATEGORICAL_PDF_HPP
#define SAMPLE_SPACE_MATH_CATEGORICAL_PDF_HPP 1

#include <cmath>
#include <type_traits>
#include <boost/optional.hpp>
#include <boost/range/combine.hpp>
#include "sample_space/math/categorical_pdf_ratio.hpp"

namespace sample_space
{
  
  namespace math
  {

    template < typename P, typename N >
    class categorical_pdf
    {
      using self = categorical_pdf;

    private:
      P p_;
      N n_;

    private:
      template < typename Pair >
      static auto term (Pair pair)
      {
	using std::get;
	return term(get<0>(pair),get<1>(pair));
      }

      template < typename _P, typename _N >
      static auto term (_P p, _N n)
      {
	using std::pow;
	return pow(p,n);
      }

    public:
      using value_type = decltype(term(*begin(p_),*begin(n_)));
      
    public:
      categorical_pdf () = default;
      categorical_pdf (P p, N n) : p_(p), n_(n) {}

      operator value_type () const { return value(); }

      value_type value () const
      {
	using result_type = boost::optional<value_type>;
	
	assert(p_.size() == n_.size());
	
	result_type result;
	for (auto pair : boost::combine(p_,n_))
	  {
	    auto term = this->term(pair);
	    if (term == 0)
	      {
		result = 0;
		break;
	      }
	    if (result)
	      *result *= term;
	    else
	      result = term;
	  }
	return result ? *result : 0;
      }

      friend auto operator / (self left, self right)
      { return make_categorical_pdf_ratio(left,right); }
    };

    template < typename P, typename N >
    auto make_categorical_pdf (P p, N n)
    {
      using categorical_pdf_type = categorical_pdf<P,N>;
      return categorical_pdf_type(p,n);
    }

  } // namespace math

} // namespace sample_space

#endif
