/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef SAMPLE_SPACE_MATH_DIRICHLET_DISTRIBUTION_HPP
#define SAMPLE_SPACE_MATH_DIRICHLET_DISTRIBUTION_HPP 1

#include <cassert>
#include <boost/math/special_functions/gamma.hpp>
#include <boost/range/combine.hpp>
#include <always_false.hpp>
#include <sample_space/math/norm_1.hpp>
#include <traits/value.hpp>
#include <traits/variate.hpp>
#include <type_traits/is_map.hpp>

namespace sample_space
{
  
  namespace math
  {

    template < typename Alpha >
    struct dirichlet_distribution
    {
      dirichlet_distribution () = default;
      dirichlet_distribution (Alpha alpha) : alpha_(alpha) {}

      Alpha const& alpha () const { return alpha_; }

      template < typename Variate >
      auto pdf (Variate variate)
      {
	assert(alpha_.size() == variate.size());
	return pdf(variate,type_traits::is_map_t<Alpha>(),type_traits::is_map_t<Variate>());
      }

    private:
      template < typename Variate >
      auto pdf (Variate variate, std::false_type, std::false_type)
      {
	using alpha_type = traits::value_t<Alpha>;
	using tgamma_result_type = decltype(tgamma(alpha_type()));
      
	using boost::get;

	tgamma_result_type pdf = { 1 };
	tgamma_result_type numerator = { 1 };
	for (auto pair : boost::combine(variate,alpha_))
	  {
	    auto x = get<0>(pair);
	    auto alpha = get<1>(pair);
	    pdf *= pow(x,alpha-1);
	    numerator *= tgamma(alpha);
	  }
	auto denominator = tgamma(norm_1(alpha_));
	auto B = numerator / denominator;
	pdf /= B;
	return pdf;
      }

      template < typename Variate >
      auto pdf (Variate variate, std::false_type, std::true_type)
      {
	using alpha_type = traits::value_t<Alpha>;
	using tgamma_result_type = decltype(tgamma(alpha_type()));
      
	using boost::get;

	tgamma_result_type pdf = { 1 };
	tgamma_result_type numerator = { 1 };
	for (auto pair : boost::combine(variate,alpha_))
	  {
	    auto x = get<0>(pair);
	    auto alpha = get<1>(pair).second;
	    pdf *= pow(x,alpha-1);
	    numerator *= tgamma(alpha);
	  }
	auto denominator = tgamma(norm_1(alpha_));
	auto B = numerator / denominator;
	pdf /= B;
	return pdf;
      }

      template < typename Variate >
      auto pdf (Variate variate, std::true_type, std::false_type)
      {
	using alpha_type = traits::value_t<Alpha>;
	using tgamma_result_type = decltype(tgamma(alpha_type()));
      
	using boost::get;

	tgamma_result_type pdf = { 1 };
	tgamma_result_type numerator = { 1 };
	for (auto pair : boost::combine(variate,alpha_))
	  {
	    auto x = get<0>(pair).second;
	    auto alpha = get<1>(pair);
	    pdf *= pow(x,alpha-1);
	    numerator *= tgamma(alpha);
	  }
	auto denominator = tgamma(norm_1(alpha_));
	auto B = numerator / denominator;
	pdf /= B;
	return pdf;
      }

      template < typename Variate >
      auto pdf (Variate variate, std::true_type, std::true_type)
      {
	using alpha_type = traits::value_t<Alpha>;
	using tgamma_result_type = decltype(tgamma(alpha_type()));
      
	using boost::get;

	tgamma_result_type pdf = { 1 };
	tgamma_result_type numerator = { 1 };
	for (auto pair : alpha_)
	  {
	    auto x = variate.at(pair.first); // XXX - allow x=0 -> pdf=0?
	    auto alpha = pair.second;
	    pdf *= pow(x,alpha-1);
	    numerator *= tgamma(alpha);
	  }
	auto denominator = tgamma(norm_1(alpha_));
	auto B = numerator / denominator;
	pdf /= B;
	return pdf;
      }

    private:
      Alpha alpha_;
    };

    template < typename Alpha, typename Variate >
    auto pdf (dirichlet_distribution<Alpha> distribution, Variate variate)
    { return distribution.pdf(variate); }

  } // namespace math

} // namespace sample_space

#endif
