/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef SAMPLE_SPACE_MATH_MULTINOMIAL_COEFFICIENT_HPP
#define SAMPLE_SPACE_MATH_MULTINOMIAL_COEFFICIENT_HPP 1

#include <boost/math/special_functions/binomial.hpp>
#include <boost/operators.hpp>
#include <boost/optional.hpp>
#include <boost/range/algorithm/equal.hpp>
#include "sample_space/math/multinomial_coefficient_ratio.hpp"
#include "sample_space/math/norm_1.hpp"

namespace sample_space
{
  
  namespace math
  {

    template < typename Value, typename N >
    class multinomial_coefficient
      : boost::equality_comparable<multinomial_coefficient<Value,N>>
    {
      using self = multinomial_coefficient;

      static_assert(!std::is_integral<Value>::value,"");

    public:
      using value_type = Value;
      
    public:
      multinomial_coefficient () = default;
      explicit multinomial_coefficient (N n) : n_(n) {}

      operator Value () const { return value(); }

      Value value () const
      {
	using result_type = boost::optional<Value>;
	
	using std::begin;
	using std::endl;
	
	result_type result;
	auto first = begin(n_);
	auto last = end(n_);
	while (first != last)
	  {
	    auto binomial_term = binomial_partition<N>(first,last);
	    if (result)
	      *result *= binomial_term;
	    else
	      result = binomial_term;
	    ++first;
	  }
	return result ? *result : 0;
      }

      friend auto operator / (self left, self right)
      { return make_multinomial_coefficient_ratio(left,right); }

      friend bool operator == (self left, self right)
      {
	using boost::equal;
	return equal(left.n_,right.n_);
      }

    private:
      template < typename Range, typename Iterator >
      static Value binomial_partition (Iterator begin, Iterator end)
      {
	auto n = detail::norm_1<Range>::apply(begin,end);
	auto k = *begin;
	return binomial_coefficient(n,k);
      }
      
      template < typename _N, typename _K >
      static Value binomial_coefficient (_N n, _K k)
      { return binomial_coefficient(convert(n),convert(k)); }
      
      static Value binomial_coefficient (unsigned n, unsigned k)
      {
	using boost::math::binomial_coefficient;
	return binomial_coefficient<Value>(n,k);
      }

      // XXX - use Boost.Conversion?
      template < typename T >
      static unsigned convert (T t) { return static_cast<unsigned>(t); }
      
    private:
      N n_;
    };

    template < typename Value, typename N >
    auto make_multinomial_coefficient (N n)
    {
      using multinomial_coefficient_type = multinomial_coefficient<Value,N>;
      return multinomial_coefficient_type(n);
    }

  } // namespace math

} // namespace sample_space

#endif
