/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef SAMPLE_SPACE_MATH_MULTINOMIAL_DISTRIBUTION_HPP
#define SAMPLE_SPACE_MATH_MULTINOMIAL_DISTRIBUTION_HPP 1

#include <sample_space/math/multinomial_pdf.hpp>
#include <traits/value.hpp>

namespace sample_space
{
  
  namespace math
  {

    /// \brief Multinomial distribution
    /// \tparam Probability The type of result
    /// \tparam P The type of the probability vector
    ///
    /// This class models a statistical distribution as defined in the
    /// Boost Math library.  Specifically, a \c pdf() non-member
    /// function exists, which calculates the probability density
    /// function for a multinomial distribution.
    ///
    template < typename Probability, typename P >
    class multinomial_distribution
    {
      using value_type = traits::value_t<P>;

    public:
      multinomial_distribution () = default;
      explicit multinomial_distribution (P probability) : probability_(probability) {}
      multinomial_distribution (std::initializer_list<value_type> il) : probability_(il) {}

      Probability const& probability () const { return probability_; }

      template < typename Variate >
      auto pdf (Variate variate)
      {
	auto pdf = make_multinomial_pdf<Probability>(probability_,variate);
	return pdf.value();
      }

    private:
      P probability_;
    };

    template < typename Probability, typename P, typename Variate >
    auto pdf (multinomial_distribution<Probability,P> distribution, Variate variate)
    { return distribution.pdf(variate); }

  } // namespace math

} // namespace sample_space

#endif
