/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef SAMPLE_SPACE_NUMERIC_BASE_HPP
#define SAMPLE_SPACE_NUMERIC_BASE_HPP 1

#include <cmath>
#include <iosfwd>
#include <boost/math/constants/constants.hpp>
#include <boost/math/special_functions/log1p.hpp>
#include <boost/math/special_functions/expm1.hpp>

namespace sample_space
{

  namespace numeric
  {

    namespace base
    {

      class e
      {
	using self = e;

      public:
	template < typename T >
	static auto base () { return boost::math::constants::e<T>(); }

	template < typename Base, typename Magnitude >
	static auto convert_to (Magnitude magnitude)
	{ return convert_to(Base(),magnitude); }

	template < typename Base, typename Magnitude >
	static auto convert_from (Magnitude magnitude)
	{ return convert_from(Base(),magnitude); }

	template < typename Magnitude >
	static auto exp (Magnitude magnitude)
	{ using std::exp; return exp(magnitude); }

	template < typename Magnitude >
	static auto log (Magnitude magnitude)
	{ using std::abs; using std::log; return log(abs(magnitude)); }

	template < typename Sign, typename Magnitude >
	static auto& print (std::ostream& os, Sign sign, Magnitude magnitude)
	{ return os << sign << "exp(" << magnitude << ")"; }

	// calculate log(1 - exp(-|x|)) with high precision for negative x.
	// reference:
	//   title={Accurately computing log(1-exp(-|a|)) assesed by the Rmpfr package}
	//   author={Maechler, Martin}
	//   pdfurl={https://cran.r-project.org/web/packages/Rmpfr/vignettes/log1mexp-note.pdf}
	//   year={2012}
	// equation (10)
	//
	template < typename Magnitude >
	static auto log1pexp (Magnitude magnitude)
	{
	  using std::exp;
	  using boost::math::log1p;
	  return			// Maechler (2012), equation (10)
	      magnitude <= -37   ? exp(magnitude)
	    : magnitude <= +18   ? log1p(exp(magnitude))
	    : magnitude <= +33.3 ? magnitude + exp(-magnitude)
	    :                      magnitude;
	}

	// reference:
	//   title={Accurately computing log(1-exp(-|a|)) assesed by the Rmpfr package}
	//   author={Maechler, Martin}
	//   pdfurl={https://cran.r-project.org/web/packages/Rmpfr/vignettes/log1mexp-note.pdf}
	//   year={2012}
	//
	template < typename Magnitude >
	static auto log1mexp (Magnitude magnitude)
	{
	  using std::abs;
	  using std::exp;
	  using std::log;
	  using boost::math::expm1;
	  using boost::math::log1p;

	  if (magnitude >= 0 )
	    return(log1pexp(magnitude));
	  
	  return			// Maechler (2012), equation (7)
	      abs(magnitude) <= log(Magnitude(2)) ? log(-expm1(magnitude))
	    :                                       log1p(-exp(magnitude));
	}

      private:
	template < typename Magnitude >
	static auto convert_to (self, Magnitude magnitude)
	{ return magnitude; }

	template < typename Base, typename Magnitude >
	static auto convert_to (Base, Magnitude magnitude)
	{ return magnitude / log(Base::base()); }

	template < typename Magnitude >
	static auto convert_from (self, Magnitude magnitude)
	{ return magnitude; }

	template < typename Base, typename Magnitude >
	static auto convert_from (Base, Magnitude magnitude)
	{ return magnitude * log(Base::base()); }
      };

      class two
      {
	using self = two;

      public:
	template < typename T >
	static auto base () { return T(2); }

	template < typename Base, typename Magnitude >
	static auto convert_to (Magnitude magnitude)
	{ return convert_to(Base(),magnitude); }

	template < typename Base, typename Magnitude >
	static auto convert_from (Magnitude magnitude)
	{ return convert_from(Base(),magnitude); }

	template < typename Magnitude >
	static auto exp (Magnitude magnitude)
	{ using std::pow; return pow(base<Magnitude>(),magnitude); }

	template < typename Magnitude >
	static auto log (Magnitude magnitude)
	{
	  using std::abs;
	  using std::log;
	  return log(abs(magnitude))/log(base<Magnitude>());
	}

	template < typename Sign, typename Magnitude >
	static auto& print (std::ostream& os, Sign sign, Magnitude magnitude)
	{ return os << sign << "2^(" << magnitude << ")"; }

	template < typename Magnitude >
	static auto log1pexp (Magnitude magnitude)
	{ return convert_from<e>(e::log1pexp(convert_to<e>(magnitude))); }

	template < typename Magnitude >
	static auto log1mexp (Magnitude magnitude)
	{ return convert_from<e>(e::log1mexp(convert_to<e>(magnitude))); }

      private:
	template < typename Magnitude >
	static auto convert_to (self, Magnitude magnitude)
	{ return magnitude; }

	template < typename Base, typename Magnitude >
	static auto convert_to (Base, Magnitude magnitude)
	{ return magnitude / log(Base::base()); }

	template < typename Magnitude >
	static auto convert_from (self, Magnitude magnitude)
	{ return magnitude; }

	template < typename Base, typename Magnitude >
	static auto convert_from (Base, Magnitude magnitude)
	{ return magnitude * log(Base::base()); }
      };

      class ten
      {
	using self = ten;

      public:
	template < typename T >
	static auto base () { return T(10); }

	template < typename Base, typename Magnitude >
	static auto convert_to (Magnitude magnitude)
	{ return convert_to(Base(),magnitude); }

	template < typename Base, typename Magnitude >
	static auto convert_from (Magnitude magnitude)
	{ return convert_from(Base(),magnitude); }

	template < typename Magnitude >
	static auto exp (Magnitude magnitude)
	{ using std::pow; return pow(base<Magnitude>(),magnitude); }

	template < typename Magnitude >
	static auto log (Magnitude magnitude)
	{ using std::abs; using std::log10; return log10(abs(magnitude)); }

	template < typename Sign, typename Magnitude >
	static auto& print (std::ostream& os, Sign sign, Magnitude magnitude)
	{ return os << sign << "10^(" << magnitude << ")"; }

	template < typename Magnitude >
	static auto log1pexp (Magnitude magnitude)
	{ return convert_from<e>(e::log1pexp(convert_to<e>(magnitude))); }

	template < typename Magnitude >
	static auto log1mexp (Magnitude magnitude)
	{ return convert_from<e>(e::log1mexp(convert_to<e>(magnitude))); }

      private:
	template < typename Magnitude >
	static auto convert_to (self, Magnitude magnitude)
	{ return magnitude; }

	template < typename Base, typename Magnitude >
	static auto convert_to (Base, Magnitude magnitude)
	{ return magnitude / log(Base::base()); }

	template < typename Magnitude >
	static auto convert_from (self, Magnitude magnitude)
	{ return magnitude; }

	template < typename Base, typename Magnitude >
	static auto convert_from (Base, Magnitude magnitude)
	{ return magnitude * log(Base::base()); }
      };

    } // namespace base

  } // namespace numeric

} // namespace sample_space

#endif
