/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef SAMPLE_SPACE_NUMERIC_LOGNUMERIC_HPP
#define SAMPLE_SPACE_NUMERIC_LOGNUMERIC_HPP 1

#include <algorithm>
#include <cmath>
#include <type_traits>
#include <boost/math/constants/constants.hpp>
#include <boost/math/special_functions/log1p.hpp>
#include <boost/math/special_functions/expm1.hpp>
#include <boost/operators.hpp>
#include <sample_space/numeric/base.hpp>
#include <traits/probability.hpp>

namespace sample_space
{

  namespace numeric
  {

    template < typename T = double
	       , typename Base = base::e
	       >
    class lognumeric
      : boost::ordered_euclidean_ring_operators< lognumeric<T> >
    {
      using limits = std::numeric_limits<T>;
      static_assert(!limits::is_integer, "Integral types are not valid logarithmic types");

      using self = lognumeric;

      class sign_type
	: boost::totally_ordered<sign_type>
      {
	using self = sign_type;

	using type = enum { negative_ = -1, positive_ = +1 };

	sign_type (type t) : sign_(t) {}

      public:
	static auto positive () { return self(positive_); }
	static auto negative () { return self(negative_); }
	    
      public:
	sign_type () = default;
	explicit sign_type (T t)
	  : sign_(initialize(t))
	{}

	auto operator - () const { return negate(*this); }
	sign_type abs () { abs(*this); }

	bool is_nonnegative () const { return sign_ == positive_; }
	bool is_negative () const { return !is_nonnegative(); }

	friend bool operator == (self lhs, self rhs)
	{ return lhs.sign_ == rhs.sign_; }
	
	friend bool operator < (self lhs, self rhs)
	{ return lhs.sign_ < rhs.sign_; }

	friend std::ostream& operator << (std::ostream& os, self s)
	{ return os << (s.sign_ == positive_ ? "+" : "-"); }

      private:
	static auto initialize (T t)
	{
	  using std::copysign;
	  return copysign(1,t) > 0 ? positive_ : negative_;
	}

	static auto negate (sign_type s)
	{ s.sign_ = s.sign_ == positive_ ? negative_ : positive_; return s; }

	static auto abs (sign_type s) { s.sign_ = positive_; return s; }

      private:
	type sign_ = positive_;
      };

      class literal_value
      {
	using self = literal_value;

      public:
	explicit literal_value (T t) : t_(t) {}

	auto magnitude () const { return t_; }
	auto sign () const { return sign_; }

	auto positive () const { return positive(*this); }
	auto negative () const { return negative(*this); }

      private:
	static auto positive (self x)
	{ x.sign_ = sign_type::positive(); return x; }

	static auto negative (self x)
	{ x.sign_ = sign_type::negative(); return x; }

      private:
	T t_;
	sign_type sign_;
      };

    public:
      static auto literal (T t) { return literal_value(t); }

    private:
      lognumeric (T t, sign_type s)
	: magnitude_(t)
	, sign_(s)
      {}

    public:
      static constexpr self min () { return self(limits::max(),sign_type::negative()); }
      static constexpr self max () { return self(limits::max(),sign_type::positive()); }
      static constexpr self epsilon () { return self(limits::epsilon(),sign_type::positive()); }

    public:
      lognumeric () = default;

      lognumeric (T t)
	: magnitude_(Base::log(t))
	, sign_(t)
      {}

      lognumeric (literal_value t)
	: magnitude_(t.magnitude())
	, sign_(t.sign())
      {}

      template < typename U >
      lognumeric (lognumeric<U> x)
	: lognumeric(x.magnitude(x),x.sign(x))
      {}

    public:
      explicit operator T () const
      { return sign_.is_nonnegative() ? Base::exp(magnitude_) : -Base::exp(magnitude_); }

    public:
      self operator + () const { return *this; }
      self operator - () const { return negate(*this); }

    public:
      self& operator += (self other) { assign_add(other); return *this; }
      self& operator -= (self other) { assign_subtract(other); return *this; }
      self& operator *= (self other) { assign_multiply(other); return *this; }
      self& operator /= (self other) { assign_divide(other); return *this; }

    public:
      friend bool operator == (self lhs, self rhs)
      { return is_equal_to(lhs,rhs); }

      friend bool operator < (self lhs, self rhs)
      { return is_less_than(normalize(lhs),normalize(rhs)); }

    public:
      friend auto abs (self x) { return self(x.magnitude_,sign_type::positive()); }
      friend auto copysign (self x, self y) { return self(x.magnitude(),y.sign()); }
      friend auto exp (self x) { return lognumeric<T,base::e>(Base::template convert_to<base::e>(Base::exp(x.magnitude_)),x.sign_); }
      friend auto exp2 (self x) { return lognumeric<T,base::two>(Base::template convert_to<base::two>(Base::exp(x.magnitude_)),x.sign_); }
      friend auto exp10 (self x) { return lognumeric<T,base::ten>(Base::template convert_to<base::ten>(Base::exp(x.magnitude_)),x.sign_); }
      friend auto& max (self const& x, self const& y) { return x >= y ? x : y; }
      friend auto& min (self const& x, self const& y) { return x <= y ? x : y; }
      friend auto fmax (self x, self y) { return static_cast<double>(max(x,y)); }
      friend auto fmin (self x, self y) { return static_cast<double>(min(x,y)); }
      friend auto isfinite (self x) { using std::isfinite; return isfinite(x.magnitude_) || x.is_zero(); }
      friend auto isinf (self x) { return x.magnitude_ == sign_type::positive_infinity(); }
      friend auto isnan (self x) { using std::isnan; return isnan(x.magnitude_); }
      friend auto log (self x) { assert(x >= 0); return Base::template convert_to<base::e>(x.magnitude_); }
      friend auto log2 (self x) { assert(x >= 0); return Base::template convert_to<base::two>(x.magnitude_); }
      friend auto log10 (self x) { assert(x >= 0); return Base::template convert_to<base::ten>(x.magnitude_); }

      friend auto pow (self base, self exponent)
      {
	using std::exp;
	return pow(base,Base::exp(exponent.magnitude_),std::false_type());
      }
	
      template < typename U >
      friend auto pow (self base, U exponent)
      {
	using U_is_integer = std::integral_constant<bool,std::numeric_limits<U>::is_integer>;
	return pow(base, exponent, U_is_integer());
      }

      friend auto sqrt (self x)
      {
	assert(x.sign_.is_nonnegative());
	x.magnitude_ /= 2;
	return x;
      }

    public:
      friend auto cos (self x) { using std::cos; return cos(Base::exp(x.magnitude_)); }
      friend auto sin (self x) { using std::sin; return sin(Base::exp(x.magnitude_)); }
      friend auto tan (self x) { using std::tan; return tan(Base::exp(x.magnitude_)); }

    public:
      friend std::ostream& operator << (std::ostream& os, self x)
      { return Base::print(os,x.sign_,x.magnitude_); }

    private:
      static auto negate (self x) { x.sign_ = -x.sign_; return x; }

    private:
      void assign_add (self other)
      {
	if (sign_ == other.sign_)
	  magnitude_ = add_logs(magnitude_,other.magnitude_);
	else
	  {
	    if (magnitude_ == other.magnitude_)
	      set_to_zero();
	    else
	      {
		if (magnitude_ < other.magnitude_)
		  sign_ = -sign_;
		magnitude_ = subtract_logs(magnitude_,other.magnitude_);
	      }
	  }
      }
      void assign_subtract (self other) { assign_add(negate(other)); }
      void assign_multiply (self other)
      {
	magnitude_ = multiply_logs(magnitude_, other.magnitude_);
	sign_ = sign_ == other.sign_ ? sign_type::positive() : sign_type::negative();
      }
      void assign_divide (self other)
      {
	magnitude_ = divide_logs(magnitude_, other.magnitude_);
	sign_ = sign_ == other.sign_ ? sign_type::positive() : sign_type::negative();
      }

    private:
      static bool is_equal_to (self lhs, self rhs)
      {
	return (lhs.is_zero() && rhs.is_zero())
	  || (lhs.sign_ == rhs.sign_ && lhs.magnitude_ == rhs.magnitude_);
      }

      static bool is_less_than (self lhs, self rhs)
      {
	assert(lhs.is_normalized());
	assert(rhs.is_normalized());
	return
	  (lhs.sign_ < rhs.sign_)
	  || (lhs.sign_ == rhs.sign_ && lhs.sign_.is_negative() && lhs.magnitude_ > rhs.magnitude_)
	  || (lhs.sign_ == rhs.sign_ && lhs.sign_.is_nonnegative() && lhs.magnitude_ < rhs.magnitude_);
      }

    private:
      template < typename U >
      static auto pow (self base, U exponent, std::true_type)
      {
	base.magnitude_ *= exponent;
	base.sign_ = (exponent % 2) ? sign_type::negative() : sign_type::positive();
	return base;
      }
      template < typename U >
      static auto pow (self base, U exponent, std::false_type)
      {
	assert(base >= 0); // XXX - fractional exponentiation of negative base is undefined.
	base.magnitude_ *= exponent;
	return base;
      }

    private:
      static auto normalize (self x)
      {
	if (x.is_zero())
	  x.sign_ = sign_type::positive();
	return x;
      }

      void set_to_zero ()
      {
	magnitude_ = negative_infinity();
	sign_ = sign_type::positive();
      }

    private:			// classification
      bool is_zero () const { return magnitude_ == negative_infinity(); }
      bool is_normalized () const { return sign_ == sign_type::positive() || !is_zero(); }

    private:
      // addition: given a=log(x) and b=log(y), compute log(x+y).
      // reference: https://en.wikipedia.org/wiki/Log_probability
      // reference: http://gasstationwithoutpumps.wordpress.com/2014/05/06/sum-of-probabilities-in-log-prob-space/
      //
      static T add_logs(T a, T b)
      {
	return
	    a == b ? a + Base::log(T(2))
	  : a > b  ? a + Base::log1pexp(b - a)
	  :          b + Base::log1pexp(a - b);
      }

      // subtraction: given a=log(x) and b=log(y), compute log(x-y)
      //
      static T subtract_logs (T a, T b)
      {
	return
	    a == b ? negative_infinity()
	  : a > b  ? a + Base::log1mexp(b - a)
	  :          b + Base::log1mexp(a - b);
      }

      // multiply: given a=log(x) and b=log(y), compute log(x*y)
      static T multiply_logs (T a, T b) { return a + b; }

      // divide: given a=log(x) and b=log(y), compute log(x/y)
      static T divide_logs (T a, T b) { return a - b; }

    private:			// special values
      static auto positive_infinity () { return limits::infinity(); }
      static auto negative_infinity () { return -positive_infinity(); }

    private:
      sign_type sign_;
      T magnitude_ = negative_infinity();
    };

  } // namespace numeric

} // namespace sample_space

namespace traits
{

  namespace defaults
  {

    template < typename T, typename Base >
    struct probability < sample_space::numeric::lognumeric<T,Base> >
    {
      using type = sample_space::numeric::lognumeric<T,Base>;
    };

  } // namespace defaults

} // namespace traits

namespace std
{

  template<typename T>
  class numeric_limits<sample_space::numeric::lognumeric<T>>
    : public std::numeric_limits<T>
  {
    using lognumeric_type = sample_space::numeric::lognumeric<T>;

  public:
    static constexpr bool is_specialized = true;
    static constexpr bool is_signed = true;
    static constexpr bool has_dnorm_loss = false;
    static constexpr bool round_style = std::round_indeterminate;
    static constexpr bool is_iec559 = false; // logarthmic types are not standard floating point
    static constexpr bool is_modulo = false;
    static constexpr lognumeric_type min () { return lognumeric_type::min(); }
    static constexpr lognumeric_type max () { return lognumeric_type::max(); }
    static constexpr lognumeric_type epsilon () { return lognumeric_type::epsilon(); }
  };

} // namespace std

#endif
