/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef SAMPLE_SPACE_RANDOM_MULTINOMIAL_DISTRIBUTION_HPP
#define SAMPLE_SPACE_RANDOM_MULTINOMIAL_DISTRIBUTION_HPP 1

#include <algorithm>
#include <iterator>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/operators.hpp>
#include <boost/random/discrete_distribution.hpp>
#include <always_false.hpp>
#include <traits/value.hpp>
#include <type_traits/is_map.hpp>

namespace sample_space
{
  
namespace random
{

  template < typename Variate
	     , typename P = Variate
	     >
  class multinomial_distribution
    : boost::equality_comparable< multinomial_distribution<Variate,P> >
  {
    using self = multinomial_distribution;
    using integer_type = long;
    using probability_type = traits::value_t<P>;
    using discrete_distribution_type = boost::random::discrete_distribution<integer_type,probability_type>;

  public:
    using result_type = Variate;

  public:
    struct param_type
      : boost::equality_comparable< param_type >
    {
      friend class multinomial_distribution;

      param_type () = default;
      explicit param_type (P p) : p_(p) {}

      param_type (std::initializer_list<probability_type> il) : p_(il) {}

      template <typename Iterator >
      param_type (Iterator begin, Iterator end) : p_(begin,end) {}

      template < typename Range >
      explicit param_type (Range const& range) : p_(range) {}

      template < typename Function >
      param_type (std::size_t n, double min, double max, Function function)
	: p_(initialize(n,min,max,function))
      {}

      auto begin () { return begin(p_is_map()); }
      auto end () { return p_.end(p_is_map()); }

      auto begin () const { return begin(p_is_map()); }
      auto end () const { return end(p_is_map()); }

      friend bool operator == (param_type left, param_type right)
      { return left.p_ == right.p_; }

    private:
      using variate_is_map = type_traits::is_map_t<Variate>;
      using p_is_map = type_traits::is_map_t<P>;

      auto make_variate () const { return make_variate<void>(variate_is_map(),p_is_map()); }

      template < typename Dummy >
      auto make_variate (std::false_type, std::false_type) const
      { return Variate(p_.size(),0); }

      template < typename Dummy >
      auto make_variate (std::false_type, std::true_type) const
      { return Variate(p_.size(),0); }

      template < typename Dummy >
      auto make_variate (std::true_type, std::false_type) const
      {
	static_assert(always_false<Dummy>::value,
		      "mismatch between Variate and P; "
		      "if Variate has an index, so must P");
      }

      template < typename Dummy >
      auto make_variate (std::true_type, std::true_type) const
      {
	Variate variate;
	std::for_each(p_.begin(),p_.end(),[&variate](auto pair)
		      { variate.insert(variate.end(),std::make_pair(pair.first,0)); });
	return variate;
      }

    private:
      auto make_param () const
      {
	using param_type = typename discrete_distribution_type::param_type;
	return param_type(begin(),end());
      }
      
    private:
      auto begin (std::false_type) { return p_.begin(); }
      auto end (std::false_type) { return p_.end(); }

      auto begin (std::false_type) const { return p_.begin(); }
      auto end (std::false_type) const { return p_.end(); }

      static auto return_second () { return [](auto pair){ return pair.second; }; }

      auto begin (std::true_type) { return boost::make_transform_iterator(p_.begin(),return_second()); }
      auto end (std::true_type) { return boost::make_transform_iterator(p_.end(),return_second()); }

      auto begin (std::true_type) const { return boost::make_transform_iterator(p_.begin(),return_second()); }
      auto end (std::true_type) const { return boost::make_transform_iterator(p_.end(),return_second()); }

    private:
      template < typename Function >
      static auto initialize (std::size_t n, double min, double max, Function function)
      { return initialize(n,min,max,function,p_is_map()); }

      template < typename Function >
      static auto initialize (std::size_t n, double min, double max, Function function, std::false_type)
      {
	auto delta = (max - min) / n;
	auto half_delta = delta / 2;

	P p;
	for (std::size_t i = 0; i < n; ++i)
	  {
	    auto x = min + half_delta + i * delta;
	    p.insert(p.end(),function(x));
	  }
	return p;
      }

      template < typename Function >
      static auto initialize (std::size_t n, double min, double max, Function function, std::true_type)
      {
	auto delta = (max - min) / n;
	auto half_delta = delta / 2;

	P p;
	for (std::size_t i = 0; i < n; ++i)
	  {
	    auto x = min + half_delta + i * delta;
	    p.insert(p.end(),std::make_pair(x,function(x)));
	  }
	return p;
      }

    private:
      P p_;
    };

  public:
    multinomial_distribution () = default;
    explicit multinomial_distribution (P p) : param_(p) {}
    explicit multinomial_distribution (param_type param) : param_(param) {}

    multinomial_distribution (std::initializer_list<probability_type> p) : param_(p) {}

    template <typename Iterator >
    multinomial_distribution (Iterator begin, Iterator end) : param_(begin,end) {}

    template < typename Range >
    explicit multinomial_distribution (Range const& range) : param_(range) {}

    auto& param () const { return distribution_.param(); }
    void param (param_type param) { distribution_.param(param); }
    
    void reset () { distribution_.reset(); }

    template < typename Engine >
    result_type operator () (Engine& engine)
    { return generate_variate(engine,param_); }
    
    template < typename Engine >
    result_type operator () (Engine& engine, param_type param)
    { return generate_variate(engine,param); }
    
    friend bool operator == (self left, self right)
    { return left.distribution_ == right.distribution_ && left.param_ == right.param_; }

    friend std::ostream& operator << (std::ostream& os, self d)
    { return os << d.distribution_; }

    friend std::istream& operator > (std::istream& is, self& d)
    { return is >> d.distribution_; }

  private:
    template < typename Engine >
    result_type generate_variate (Engine& engine, param_type param)
    {
      Variate result = param.make_variate();
      auto n = distribution_(engine,param.make_param());
      auto i = result.begin();
      std::advance(i,n);
      ++(*i);

      return result;
    }
    
  private:
    discrete_distribution_type distribution_;
    param_type param_;
  };
  
} // namespace random

} // namespace sample_space

#endif
