/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef SAMPLE_SPACE_VECTOR_POLICY_EUCLIDEAN_VECTOR_SPACE_BASE_HPP
#define SAMPLE_SPACE_VECTOR_POLICY_EUCLIDEAN_VECTOR_SPACE_BASE_HPP 1

#include <algorithm>
#include <cmath>
#include <iosfwd>
#include <boost/operators.hpp>
#include <boost/range/algorithm/for_each.hpp>
#include <boost/range/combine.hpp>
#include <mpl/enable_constructor.hpp>
#include <sample_space/vector/mpl/vector_storage.hpp>
#include <sample_space/vector/tag/euclidean_basis.hpp>
#include <sample_space/vector/tag/euclidean_vector_space.hpp>
#include <sample_space/vector/type_traits/is_map.hpp>

namespace sample_space
{

  namespace vector
  {

    namespace policy
    {

      /// \brief A Euclidean vector space base class
      /// \tparam Derived The type of derived class for CRTP
      /// \tparam T The element in the vector space
      /// \tparam Index The type of index (if any)
      /// \tparam Tag The tag defining the type of numeric sample space
      /// \tparam Basis The basis for the vector space
      /// \tparam Storage The type of storage to use
      ///
      template < typename Derived
		 , typename T
		 , typename Index
		 , typename Tag
		 , typename Basis
		 , typename Storage
		 >
      class euclidean_vector_space_base;

      template < typename Derived
		 , typename T
		 , typename Index
		 , typename Storage
		 >
      class euclidean_vector_space_base<Derived,T,Index,tag::euclidean_vector_space,tag::euclidean_basis,Storage>
	: boost::totally_ordered< Derived
	, boost::additive< Derived
	, boost::integer_multiplicative<Derived,T
	> > >
      {
	using self = euclidean_vector_space_base;
	using vector_type = mpl::vector_storage_t<T,Index,Storage>;

      public:
	using size_type = typename vector_type::size_type;
	using iterator = typename vector_type::iterator;
	using const_iterator = typename vector_type::const_iterator;
	using value_type = typename vector_type::value_type;

      public:
	using is_map_t = type_traits::is_map_t<Index,Storage>;

      public:
	template < typename ... Args
		   , typename = ::mpl::enable_constructor_t<self,Args...>
		   >
	explicit euclidean_vector_space_base (Args&& ... args)
	  : vector_(std::forward<Args>(args)...)
	{}

	euclidean_vector_space_base (std::initializer_list<T> x)
	  : vector_(x)
	{}

	bool empty () const { return vector_.empty(); }
	size_type size () const { return vector_.size(); }

	iterator begin () {  return vector_.begin(); }
	iterator end () { return vector_.end(); }

	friend iterator begin (self& base) { return base.begin(); }
	friend iterator end (self& base) { return base.end(); }

	const_iterator begin () const {  return vector_.begin(); }
	const_iterator end () const { return vector_.end(); }

	friend const_iterator begin (self const& base) { return base.begin(); }
	friend const_iterator end (self const& base) { return base.end(); }

	template < typename Key >
	auto& at (Key key) { return vector_.at(key); }
	template < typename Key >
	auto& at (Key key) const { return vector_.at(key); }

	template < typename Key >
	auto& operator [] (Key key) { return vector_[key]; }
	template < typename Key >
	auto& operator [] (Key key) const { return vector_[key]; }

	void push_back (T t) { vector_.push_back(t); }

	template < typename Value >
	auto insert (const_iterator hint, Value value) { return vector_.insert(hint,value); }

	auto& zero () { vector_.zero(); return derived(); }
	auto& one () { vector_.one(); return derived(); }

      public:
	friend constexpr bool is_possible (Derived) { return true; }
      
      public:
	friend constexpr Derived operator + (Derived other) { return other; }

      public:
	friend constexpr Derived operator - (Derived other) { return negate(other); }

      private:
	static Derived negate (Derived other)
	{ return negate(other,type_traits::is_map_t<Index,Storage>()); }

	static Derived negate (Derived other, std::false_type)
	{
	  using boost::for_each;
	  for_each(other.vector_,[](auto& x){ x = -x; });
	  return other;
	}

	static Derived negate (Derived other, std::true_type)
	{
	  using boost::for_each;
	  for_each(other.vector_,[](auto& x){ x.second = -x.second; });
	  return other;
	}

      public:
	Derived& operator += (Derived other) { assign_add(other); return derived(); }

      private:
	void assign_add (Derived other)
	{ assign_add(other,type_traits::is_map_t<Index,Storage>()); }

	void assign_add (Derived other, std::false_type)
	{
	  using boost::combine;
	  assert_is_concordant(other);
	  for (auto pair : combine(vector_,other.vector_))
	    {
	      using boost::get;
	      get<0>(pair) += get<1>(pair);
	    }
	}

	void assign_add (Derived other, std::true_type)
	{
	  using iterator = typename vector_type::iterator;
	
	  for (auto pair : other.vector_)
	    {
	      iterator i;
	      bool inserted;
	      boost::tie(i,inserted) = vector_.insert(std::make_pair(pair.first,pair.second));
	      if (!inserted)
		i->second += pair.second;
	    }
	}

      public:
	Derived& operator -= (Derived other) { assign_subtract(other); return derived(); }

      private:
	void assign_subtract (Derived other)
	{ assign_subtract(other,type_traits::is_map_t<Index,Storage>()); }

	void assign_subtract (Derived other, std::false_type)
	{
	  using boost::combine;
	  assert_is_concordant(other);
	  for (auto pair : combine(vector_,other.vector_))
	    {
	      using boost::get;
	      get<0>(pair) -= get<1>(pair);
	    }
	}

	void assign_subtract (Derived other, std::true_type)
	{
	  using iterator = typename vector_type::iterator;
	
	  for (auto pair : other.vector_)
	    {
	      iterator i;
	      bool inserted;
	      boost::tie(i,inserted) = vector_.insert(std::make_pair(pair.first,-pair.second));
	      if (!inserted)
		i->second -= pair.second;
	    }
	}

      public:
	Derived& operator *= (T other) { assign_multiply(other); return derived(); }

      private:
	void assign_multiply (T other)
	{ assign_multiply(other,type_traits::is_map_t<Index,Storage>()); }

	void assign_multiply (T other, std::false_type)
	{
	  for (auto& x : vector_)
	    x *= other;
	}

	void assign_multiply (T other, std::true_type)
	{
	  for (auto& x : vector_)
	    x.second *= other;
	}

      public:
	Derived& operator /= (T other) { assign_divide(other); return derived(); }

      private:
	void assign_divide (T other)
	{ assign_divide(other,type_traits::is_map_t<Index,Storage>()); }

	void assign_divide (T other, std::false_type)
	{
	  for (auto& x : vector_)
	    x /= other;
	}

	void assign_divide (T other, std::true_type)
	{
	  for (auto& x : vector_)
	    x.second /= other;
	}

      public:
	Derived& operator %= (T other) { assign_modulo(other); return derived(); }

      private:
	void assign_modulo (T other)
	{ assign_modulo(other,type_traits::is_map_t<Index,Storage>()); }

	void assign_modulo (T other, std::false_type)
	{
	  for (auto& x : vector_)
	    x %= other;
	}

	void assign_modulo (T other, std::true_type)
	{
	  for (auto& x : vector_)
	    x.second %= other;
	}

      public:
	friend bool operator == (Derived left, Derived right)
	{ return equal(left,right); }
	
      private:
	static bool equal (Derived left, Derived right)
	{ return equal(left,right,type_traits::is_map_t<Index,Storage>()); }
	  
	static bool equal (Derived left, Derived right, std::false_type)
	{
	  using boost::equal;
	  return equal(left.vector_,right.vector_);
	}
	  
	static bool equal (Derived left, Derived right, std::true_type)
	{
	  using boost::equal;
	  return equal(left.vector_,right.vector_);
	}
	  
      public:
	friend bool operator < (Derived left, Derived right)
	{ return less_than(left,right); }
	
      private:
	static bool less_than (Derived left, Derived right)
	{ return less_than(left,right,type_traits::is_map_t<Index,Storage>()); }
	
	static bool less_than (Derived left, Derived right, std::false_type)
	{
	  using std::lexicographical_compare;
	  return lexicographical_compare
	    (left.vector_.begin(),left.vector_.end(),
	     right.vector_.begin(),right.vector_.end());
	}
	
	static bool less_than (Derived left, Derived right, std::true_type)
	{
	  using std::lexicographical_compare;
	  return lexicographical_compare
	    (left.vector_.begin(),left.vector_.end(),
	     right.vector_.begin(),right.vector_.end());
	}

      public:
	friend T dot_product (Derived first, Derived second)
	{ return dot_product(first,second,type_traits::is_map_t<Index,Storage>()); }

      private:
	static T dot_product (Derived first, Derived second, std::false_type)
	{
	  using boost::combine;
	  T result = 0;
	  for (auto pair : combine(first.vector_,second.vector_))
	    {
	      using boost::get;
	      result += get<0>(pair) * get<1>(pair);
	    }
	  return result;
	}

	static T dot_product (Derived first, Derived second, std::true_type)
	{
	  using boost::combine;
	  T result = 0;
	  for (auto pair : combine(first.vector_,second.vector_))
	    {
	      using boost::get;
	      result += get<0>(pair).second * get<1>(pair).second;
	    }
	  return result;
	}

      public:
	friend Derived hadamard_product (Derived first, Derived second)
	{ return hadamard_product(first,second,type_traits::is_map_t<Index,Storage>()); }

      private:
	static Derived hadamard_product (Derived first, Derived second, std::false_type)
	{
	  using boost::combine;
	  for (auto& pair : combine(first.vector_,second.vector_))
	    {
	      using boost::get;
	      get<0>(pair) *= get<1>(pair);
	    }
	  return first;
	}

	static Derived hadamard_product (Derived first, Derived second, std::true_type)
	{
	  using boost::combine;
	  for (auto& pair : combine(first.vector_,second.vector_))
	    {
	      using boost::get;
	      get<0>(pair).second *= get<1>(pair).second;
	    }
	  return first;
	}

      public:
	friend T norm_2_compare (Derived other) { return self::norm_2_compare(other); }
	
      private:
	static T norm_2_compare (Derived other)
	{ return norm_2_compare(other,type_traits::is_map_t<Index,Storage>()); }
	
	static T norm_2_compare (Derived other, std::false_type)
	{
	  using std::abs;
	  using boost::for_each;
	  
	  T result = 0;
	  for_each(other.vector_,[&result](auto x){ result += x*x; });
	  return result;
	}
	
	static T norm_2_compare (Derived other, std::true_type)
	{
	  using std::abs;
	  using boost::for_each;
	  
	  T result = 0;
	  for_each(other.vector_,[&result](auto pair){ result += pair.second*pair.second; });
	  return result;
	}

      public:
	friend T norm_2 (Derived other)
	{ using std::sqrt; return sqrt(norm_2_compare(other)); }

      public:
	friend std::ostream& operator << (std::ostream& os, Derived other)
	{ print(os,other); return os; }

      private:
	static void print (std::ostream& os, Derived other)
	{ print(os,other,type_traits::is_map_t<Index,Storage>()); }
	
	static void print (std::ostream& os, Derived other, std::false_type)
	{
	  using boost::for_each;
	  bool first = true;
	  os << '[';
	  for_each(other.vector_,[&os,&first](auto x)
		   { os << (first ? "" : ",") << x; first=false; });
	  os << ']';
	}

	static void print (std::ostream& os, Derived other, std::true_type)
	{
	  using boost::for_each;
	  bool first = true;
	  os << '[';
	  for_each(other.vector_,[&os,&first](auto x)
		   { os << (first ? "" : ",") << '(' << x.first << ',' << x.second << ')'; first=false; });
	  os << ']';
	}

      private:
	void assert_is_concordant (Derived other)
	{ assert(other.size() == vector_.size()); }

      private:
	Derived& derived () { return static_cast<Derived&>(*this); }
	Derived const& derived () const { return static_cast<Derived const&>(*this); }

      private:
	vector_type vector_;
      };

    } // namespace policy

  } // namespace vector
  
} // namespace sample_space

#endif
