/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef SAMPLE_SPACE_VECTOR_POLICY_VECTOR_STORAGE_POLICY_HPP
#define SAMPLE_SPACE_VECTOR_POLICY_VECTOR_STORAGE_POLICY_HPP 1

#include <algorithm>
#include <map>
#include <vector>
#include <mpl/enable_constructor.hpp>
#include <traits/vector.hpp>

namespace sample_space
{

  namespace vector
  {

    namespace policy
    {

      /// \brief Storage policy for variates
      /// \tparam T The element type
      /// \tparam Index The index type (see below)
      ///
      /// This class defines the policy used for storing vector-based
      /// or map-based variates.  The distinction between the two
      /// types of storage is based upon the \c Index template
      /// parameter.  A void \c Index selects vector-based storage;
      /// otherwise map-based storage is used with \c Index being the
      /// key type for the map and \c T being the mapped type.
      ///
      template < typename T, typename Index >
      class vector_storage_policy;

      /// \brief Vector-based storage policy
      ///
      template < typename T >
      class vector_storage_policy<T,void>
      {
	using self = vector_storage_policy;
	using vector_type = std::vector<T>;

      public:
	using size_type = typename vector_type::size_type;
	using iterator = typename vector_type::iterator;
	using const_iterator = typename vector_type::const_iterator;
	using value_type = typename vector_type::value_type;
	using index_type = void;
      
      public:
	template < typename ... Args
		   , typename = ::mpl::enable_constructor_t<self,Args...>
		   >
	explicit vector_storage_policy (Args&& ... args)
	: vector_(std::forward<Args>(args)...)
	{}

	vector_storage_policy (std::initializer_list<T> x)
	  : vector_(x)
	{}

	bool empty () const { return vector_.empty(); }
	size_type size () const { return vector_.size(); }

	iterator begin () { return vector_.begin(); }
	iterator end () { return vector_.end(); }

	friend iterator begin (self& base) { return base.begin(); }
	friend iterator end (self& base) { return base.end(); }
	
	const_iterator begin () const { return vector_.begin(); }
	const_iterator end () const { return vector_.end(); }

	friend const_iterator begin (self const& base) { return base.begin(); }
	friend const_iterator end (self const& base) { return base.end(); }
	
	T& operator [] (size_type n) { return vector_[n]; }
	T const& operator [] (size_type n) const { return vector_[n]; }

	T& at (size_type n) { return vector_[n]; }
	T const& at (size_type n) const { return vector_[n]; }

	void push_back (T t) { vector_.push_back(t); }

	auto insert (const_iterator hint, T t) { return vector_.insert(hint,t); }

	auto& zero () { vector_.assign(size(),0); return *this; }
	auto& one () { vector_.assign(size(),1); return *this; }

      private:
	vector_type vector_;
      };

      /// \brief Map-based storage policy
      ///
      template < typename T, typename Index >
      class vector_storage_policy
      {
	using self = vector_storage_policy;
	using map_type = std::map<Index,T>;

      public:
	using size_type = typename map_type::size_type;
	using iterator = typename map_type::iterator;
	using const_iterator = typename map_type::const_iterator;
	using value_type = typename map_type::mapped_type;
	using index_type = typename map_type::key_type;

      public:
	template < typename ... Args
		   , typename = ::mpl::enable_constructor_t<self,Args...>
		   >
	vector_storage_policy (Args&& ... args)
	  : map_(std::forward<Args>(args)...)
	{}

	vector_storage_policy (std::initializer_list<typename map_type::value_type> x)
	  : map_(x)
	{}

	bool empty () const { return map_.empty(); }
	size_type size () const { return map_.size(); }

	iterator begin () { return map_.begin(); }
	iterator end () { return map_.end(); }

	friend iterator begin (self& base) { return base.begin(); }
	friend iterator end (self& base) { return base.end(); }

	const_iterator begin () const { return map_.begin(); }
	const_iterator end () const { return map_.end(); }

	friend const_iterator begin (self const& base) { return base.begin(); }
	friend const_iterator end (self const& base) { return base.end(); }

	T& operator [] (Index i) { return map_[i]; }
	T const& operator [] (Index i) const { return map_[i]; }

	T& at (Index i) { return map_[i]; }
	T const& at (Index i) const { return map_[i]; }

	std::pair<iterator,bool> insert (typename map_type::value_type pair)
	{ return map_.insert(pair); }

	auto insert (const_iterator hint, typename map_type::value_type pair)
	{ return map_.insert(hint,pair); }

	auto& zero () { std::for_each(begin(),end(),[](auto& pair){ pair.second = 0; }); return *this; }
	auto& one () { std::for_each(begin(),end(),[](auto& pair){ pair.second = 1; }); return *this; }

      private:
	map_type map_;
      };

    } // namespace policy

  } // namespace vector
  
} // namespace sample_space

#endif
