/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#define BOOST_TEST_MODULE dataframe_sequence_proxy

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include "dataframe/sequence_proxy.hpp"
#include <boost/mpl/vector.hpp>
#include <boost/range/algorithm/equal.hpp>

using test_types
= boost::mpl::vector<
  std::vector<int>
  , std::vector<long>
  , std::vector<long long>
  , std::vector<float>
  , std::vector<double>
  , std::vector<long double>
  >;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_typedef,Sequence,test_types)
{
  using sequence_proxy_type = dataframe::sequence_proxy<Sequence>;
  using const_sequence_proxy_type = dataframe::sequence_proxy<Sequence const>;
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_default_constructor,Sequence,test_types)
{
  using sequence_proxy_type = dataframe::sequence_proxy<Sequence>;
  using const_sequence_proxy_type = dataframe::sequence_proxy<Sequence const>;

  sequence_proxy_type sequence_proxy;
  BOOST_CHECK(sequence_proxy.empty());
  BOOST_CHECK_EQUAL(sequence_proxy.size(),0);
  const_sequence_proxy_type const_sequence_proxy;
  BOOST_CHECK(const_sequence_proxy.empty());
  BOOST_CHECK_EQUAL(const_sequence_proxy.size(),0);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_sequence_proxy,Sequence,test_types)
{
  using sequence_proxy_type = dataframe::sequence_proxy<Sequence>;

  using boost::equal;
  
  Sequence sequence = { 42, 17 };
  sequence_proxy_type sequence_proxy(sequence);
  BOOST_CHECK(!sequence_proxy.empty());
  BOOST_CHECK_EQUAL(sequence_proxy.size(),2);
  BOOST_CHECK(equal(sequence_proxy,sequence));
  sequence_proxy = { 13 };
  BOOST_CHECK(!sequence_proxy.empty());
  BOOST_CHECK_EQUAL(sequence_proxy.size(),1);
  BOOST_CHECK(equal(sequence_proxy,Sequence({13})));
  sequence_proxy = { 17 };
  BOOST_CHECK(!sequence_proxy.empty());
  BOOST_CHECK_EQUAL(sequence_proxy.size(),1);
  BOOST_CHECK(equal(sequence_proxy,Sequence(1,17)));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_const_sequence_proxy,Sequence,test_types)
{
  using sequence_proxy_type = dataframe::sequence_proxy<Sequence>;
  using const_sequence_proxy_type = dataframe::sequence_proxy<Sequence const>;

  using boost::equal;
  
  Sequence sequence = { 42, 17 };
  const_sequence_proxy_type sequence_proxy(sequence);
  BOOST_CHECK(!sequence_proxy.empty());
  BOOST_CHECK_EQUAL(sequence_proxy.size(),2);
  BOOST_CHECK(equal(sequence_proxy,Sequence({42,17})));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_sequence_proxy,Sequence,test_types)
{
  using sequence_proxy_type = dataframe::sequence_proxy<Sequence>;
  using value_type = typename Sequence::value_type;

  using dataframe::make_sequence_proxy;
  
  Sequence sequence = { 42, 17 };
  auto sequence_proxy = make_sequence_proxy<value_type>(sequence);
  static_assert(std::is_same<decltype(sequence_proxy),sequence_proxy_type>::value,"");
  BOOST_CHECK_EQUAL(sequence_proxy.at(0),42);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_sequence_proxy_2,Sequence,test_types)
{
  using sequence_proxy_type = dataframe::sequence_proxy<Sequence,std::string>;

  using dataframe::make_sequence_proxy;
  
  Sequence sequence = { 42, 17 };
  auto sequence_proxy = make_sequence_proxy<std::string>(sequence);
  static_assert(std::is_same<decltype(sequence_proxy),sequence_proxy_type>::value,"");
  BOOST_CHECK_EQUAL(sequence_proxy.at(0),"42");
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_const_sequence_proxy,Sequence,test_types)
{
  using const_sequence_proxy_type = dataframe::sequence_proxy<Sequence const>;
  using value_type = typename Sequence::value_type;

  using dataframe::make_sequence_proxy;
  
  Sequence const sequence = { 42, 17 };
  auto const_sequence_proxy = make_sequence_proxy<value_type>(sequence);
  static_assert(std::is_same<decltype(const_sequence_proxy),const_sequence_proxy_type>::value,"");
  BOOST_CHECK_EQUAL(const_sequence_proxy.at(0),42);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_const_sequence_proxy_2,Sequence,test_types)
{
  using const_sequence_proxy_type = dataframe::sequence_proxy<Sequence const,std::string>;

  using dataframe::make_sequence_proxy;
  
  Sequence const sequence = { 42, 17 };
  auto const_sequence_proxy = make_sequence_proxy<std::string>(sequence);
  static_assert(std::is_same<decltype(const_sequence_proxy),const_sequence_proxy_type>::value,"");
  BOOST_CHECK_EQUAL(const_sequence_proxy.at(0),"42");
}
