/*
 * Copyright (c) 2017 Gregory Penn.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#define BOOST_TEST_MODULE gdal

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;
#include <iostream>
#include <iomanip>
#include <cmath>
#include <limits>

#include <boost/mpl/vector.hpp>
#include <iostream>
#include <ogr_core.h>
#include <gdal/affine_transform.hpp>
#include <gdal/geographic_srs.hpp>
#include <gdal/grid_srs.hpp>
#include <eigen3/Eigen/Dense>
#include <boost/foreach.hpp>


using std::cout;
using std::endl;
using namespace gdal;


BOOST_AUTO_TEST_CASE(geographic_default_constructor)
{
  affine_transform<geographic_srs> transformer;
}

BOOST_AUTO_TEST_CASE(grid_default_constructor)
{
  affine_transform<grid_srs> transformer;
}

BOOST_AUTO_TEST_CASE(geographic_arg_constructor)
{
  using transformer = affine_transform<geographic_srs>;
  using grid_coords = std::array<int,2>;
  using geog_coords = std::array<double,2>;
  using matrix_type = Eigen::Matrix<double, 2, 2>;
  using gridpt = point<grid_srs>;
  using geogpt = point<geographic_srs>;

  auto srs  = make_geographic_srs(3857);
  auto grid = grid_srs(2,3);

  auto coeffs = std::array<double, 6>({ -11866500, 1000, 0, 3790500, 0, -500 });

  transformer srs_to_grid(srs, grid, coeffs);
  BOOST_CHECK_EQUAL(srs_to_grid.srs(), srs);
  BOOST_CHECK_EQUAL(srs_to_grid.grid(), grid);

  struct comp
  {
    bool operator () (gridpt const& a, gridpt const& b) const
      {
        bool result;
        if (a.y() < b.y()) result = true;
        else if ((a.y() == b.y()) && (a.x() < b.x())) result = true;
        else result = false;
        return result;
      }
  };

  gridpt gridpt00( 0, 0, grid );
  geogpt geogpt00( -11866031.303, 3790264.897, srs );
  BOOST_CHECK_EQUAL(srs_to_grid(geogpt00), gridpt00);

  gridpt gridpt10( 1, 0, grid );
  geogpt geogpt10( -11864970.721, 3790324.171, srs );
  BOOST_CHECK_EQUAL(srs_to_grid(geogpt10), gridpt10);

  gridpt gridpt01( 0, 1, grid );
  geogpt geogpt01( -11866274.750, 3789701.794, srs );
  BOOST_CHECK_EQUAL(srs_to_grid(geogpt01), gridpt01);

  gridpt gridpt11( 1, 1, grid );
  geogpt geogpt11( -11864600.259, 3789917.721, srs );
  BOOST_CHECK_EQUAL(srs_to_grid(geogpt11), gridpt11);

  gridpt gridpt02( 0, 2, grid );
  geogpt geogpt02( -11866399.649, 3789060.364, srs );
  BOOST_CHECK_EQUAL(srs_to_grid(geogpt02), gridpt02);

  gridpt gridpt12( 1, 2, grid );
  geogpt geogpt12( -11864589.674, 3789054.013, srs );
  BOOST_CHECK_EQUAL(srs_to_grid(geogpt12), gridpt12);

}

BOOST_AUTO_TEST_CASE(affine_transform_grid_constructor)
{
  auto srs  = make_geographic_srs(3857);
  auto grid = grid_srs(2,3);
  auto coeffs = std::array<double, 6>({ -11866500, 1000, 0, 3790500, 0, -500 });


  affine_transform<grid_srs> transformer(srs, grid, coeffs);
  Eigen::Matrix<double, 2, 1> origin;
  Eigen::Matrix<double, 2, 2> matrix;

  origin << coeffs[0], coeffs[3];
  matrix << coeffs[1], coeffs[2], coeffs[4], coeffs[5];

  BOOST_CHECK_EQUAL(origin, transformer.origin());
  BOOST_CHECK_EQUAL(matrix, transformer.matrix());
}

BOOST_AUTO_TEST_CASE(transform_grid_to_geographic)
{
  std::cout << "====> geotransform" << std::endl;

  using grid_coords = std::array<int,2>;
  using geog_coords = std::array<double,2>;

  auto srs  = make_geographic_srs(3857);
  auto grid = grid_srs(2,3);
  auto coeffs = std::array<double, 6>({ -11866500, 1000, 0, 3790500, 0, -500 });

  affine_transform<grid_srs> transformer(srs, grid, coeffs);
  BOOST_CHECK_EQUAL(transformer.srs(), srs);
  BOOST_CHECK_EQUAL(transformer.grid(), grid);

  std::cout << "origin:\n" << transformer.origin()
            << "\nmatrix:\n" << transformer.matrix()
            << std::endl;

  std::map<grid_coords, geog_coords> coords_map;
  coords_map[grid_coords{ 0, 0 }] = geog_coords{ -11866500.0, 3790500.0 };
  coords_map[grid_coords{ 1, 0 }] = geog_coords{ -11865500.0, 3790500.0 };
  coords_map[grid_coords{ 0, 1 }] = geog_coords{ -11866500.0, 3790000.0 };
  coords_map[grid_coords{ 1, 1 }] = geog_coords{ -11865500.0, 3790000.0 };
  coords_map[grid_coords{ 0, 2 }] = geog_coords{ -11866500.0, 3789500.0 };
  coords_map[grid_coords{ 1, 2 }] = geog_coords{ -11865500.0, 3789500.0 };

  for (auto& kv : coords_map)
  {
//    Eigen::Matrix<double, 2, 1> ;

//    location << kv.first[0],  kv.first[1];
//    expected << kv.second[0], kv.second[1];
//    Eigen::Matrix<double, 2, 1> returned = transformer(location);

    point<grid_srs> location(kv.first[0], kv.first[1]);
    point<geographic_srs> expected(kv.second[0], kv.second[1], srs);
    point<geographic_srs> returned(transformer(location));

    std::cout << "returned:\n"
              << std::fixed
              << std::setprecision(2)
              << returned
              << std::endl;
    BOOST_CHECK_EQUAL(returned, expected);
  }
}
