/*
 * Copyright (c) 2017 Gregory Penn.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#define BOOST_TEST_MODULE gdal

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <iostream>
#include <memory>
#include <gdal/rasterdataset.hpp>

using namespace gdal;

#define rasterfile "../data/test_raster.tif"

struct obj
{
  obj () = default;
  obj (int refs)
    : count(refs)
    {}

  ~obj () { f() = 0; }

  void Reference () { count = count + 1; }
  void Dereference () { count = count - 1; }

  int GetRefCount() { return count; }

  static int& f () { static int i(1); return i; }

  int count = 1;
};

BOOST_AUTO_TEST_CASE(ref_manager_default_constructor)
{
  auto p = ref_manager();
}

BOOST_AUTO_TEST_CASE(ref_manager_lambda_constructor)
{
  obj* p;
  p = new obj;

  ref_manager rm([p] () { delete p; }
                 , [p] () { p->Reference(); }
                 , [p] () { p->Dereference(); }
                 , [p] () { return p->GetRefCount();});
  BOOST_CHECK_EQUAL(rm.use_count(), 1);
}

BOOST_AUTO_TEST_CASE(ref_manager_destructor)
{
  {
    obj* p;
    p = new obj;
    obj::f() = 1;
    ref_manager rm([p] () { delete p; }
                   , [p] () { p->Reference(); }
                   , [p] () { p->Dereference(); }
                   , [p] () { return p->GetRefCount();});
    BOOST_CHECK_EQUAL(obj::f(), 1);

    {
      auto rm2 = rm;
      BOOST_CHECK_EQUAL(rm.use_count(), 2);
    }
    BOOST_CHECK_EQUAL(p->GetRefCount(), 1);
    BOOST_CHECK_EQUAL(obj::f(), 1);
  }
  BOOST_CHECK_EQUAL(obj::f(), 0);
}

BOOST_AUTO_TEST_CASE(ref_manager_copy_constructor)
{
  obj* p;
  p = new obj;
  obj::f() = 1;
  ref_manager rm([p] () { delete p; }
                 , [p] () { p->Reference(); }
                 , [p] () { p->Dereference(); }
                 , [p] () { return p->GetRefCount();});

  BOOST_CHECK_EQUAL(rm.use_count(), 1);

  auto rm2 = rm;
  BOOST_CHECK_EQUAL(rm.use_count(), 2);
}

BOOST_AUTO_TEST_CASE(ref_manager_move_constructor)
{
  obj* p;
  p = new obj;
  obj::f() = 1;
  ref_manager rm([p] () { delete p; }
                 , [p] () { p->Reference(); }
                 , [p] () { p->Dereference(); }
                 , [p] () { return p->GetRefCount();});

  BOOST_CHECK_EQUAL(rm.use_count(), 1);

  auto rm2 = std::move(rm);
  BOOST_CHECK_EQUAL(rm2.use_count(), 1);
}
