/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_MODULE graph_model_distribution_uniform_uniform_distribution

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <graph_model/distribution/uniform/uniform_distribution.hpp>
#include <graph_model/distribution/parameter/lower_bound.hpp>
#include <graph_model/distribution/parameter/upper_bound.hpp>
#include <graph_model/distribution/parameter/variate.hpp>
#include <graph_model/vertex/lit.hpp>

#include <boost/mpl/vector.hpp>

using test_types = boost::mpl::vector<float,double,long double>;

#if XXX_TOO_MANY_ARGUMENTS
BOOST_AUTO_TEST_CASE(test_too_many_arguments)
{
  using graph_model::distribution::uniform::uniform_distribution;
  auto distribution = uniform_distribution(1,2,3,4);
}
#endif

BOOST_AUTO_TEST_CASE_TEMPLATE(test_default_constructor,T,test_types)
{
  using graph_model::distribution::uniform::uniform_distribution;
  using graph_model::distribution::parameter::lower_bound;
  using graph_model::distribution::parameter::upper_bound;
  using graph_model::distribution::parameter::variate;
  {
    auto distribution = uniform_distribution();
    BOOST_CHECK_EQUAL(distribution.lower_bound(),T(0));
    BOOST_CHECK_EQUAL(distribution.upper_bound(),T(1));
    BOOST_CHECK_EQUAL(distribution.variate(),T(0.5));
    BOOST_CHECK_EQUAL(distribution.pdf(),T(1));
  }
  {
    auto distribution = uniform_distribution(T(-1));
    BOOST_CHECK_EQUAL(distribution.lower_bound(),T(-1));
    BOOST_CHECK_EQUAL(distribution.upper_bound(),T(1));
    BOOST_CHECK_EQUAL(distribution.variate(),T(0.5));
    BOOST_CHECK_EQUAL(distribution.pdf(),T(0.5));
  }
  {
    auto distribution = uniform_distribution(T(-2),T(+2));
    BOOST_CHECK_EQUAL(distribution.lower_bound(),T(-2));
    BOOST_CHECK_EQUAL(distribution.upper_bound(),T(+2));
    BOOST_CHECK_EQUAL(distribution.variate(),T(0.5));
    BOOST_CHECK_EQUAL(distribution.pdf(),T(0.25));
  }
  {
    auto distribution = uniform_distribution(variate=T(0.5),upper_bound=T(1),lower_bound=T(-2));
    BOOST_CHECK_EQUAL(distribution.lower_bound(),T(-2));
    BOOST_CHECK_EQUAL(distribution.upper_bound(),T(1));
    BOOST_CHECK_EQUAL(distribution.variate(),T(0.5));
    BOOST_CHECK_EQUAL(distribution.pdf(),T(1)/T(3));
  }
  {
    auto distribution = uniform_distribution(lower_bound=T(-2),T(+2));
    BOOST_CHECK_EQUAL(distribution.lower_bound(),T(-2));
    BOOST_CHECK_EQUAL(distribution.upper_bound(),T(+2));
    BOOST_CHECK_EQUAL(distribution.variate(),T(0.5));
    BOOST_CHECK_EQUAL(distribution.pdf(),T(0.25));
  }
  {
    auto distribution = uniform_distribution(T(+2),lower_bound=T(-2));
    BOOST_CHECK_EQUAL(distribution.lower_bound(),T(-2));
    BOOST_CHECK_EQUAL(distribution.upper_bound(),T(+2));
    BOOST_CHECK_EQUAL(distribution.variate(),T(0.5));
    BOOST_CHECK_EQUAL(distribution.pdf(),T(0.25));
  }
  {
    auto distribution = uniform_distribution(lower_bound=T(0),upper_bound=T(42),variate=T(7));
    BOOST_CHECK_EQUAL(distribution.lower_bound(),T(0));
    BOOST_CHECK_EQUAL(distribution.upper_bound(),T(42));
    BOOST_CHECK_EQUAL(distribution.variate(),T(7));
    BOOST_CHECK_EQUAL(distribution.pdf(),T(1)/T(42));
  }
}

BOOST_AUTO_TEST_CASE(test_constructor)
{
  using graph_model::distribution::uniform::uniform_distribution;
  {
    auto distribution = uniform_distribution();
    BOOST_CHECK_EQUAL(distribution.lower_bound(),0);
    BOOST_CHECK_EQUAL(distribution.upper_bound(),1);
    BOOST_CHECK_EQUAL(distribution.variate(),0.5);
    BOOST_CHECK_EQUAL(distribution.pdf(),1);
  }
  {
    auto distribution = uniform_distribution(-1);
    BOOST_CHECK_EQUAL(distribution.lower_bound(),-1);
    BOOST_CHECK_EQUAL(distribution.upper_bound(),1);
    BOOST_CHECK_EQUAL(distribution.variate(),0.5);
    BOOST_CHECK_EQUAL(distribution.pdf(),0.5);
  }
  {
    auto distribution = uniform_distribution(0.25,0.5);
    BOOST_CHECK_EQUAL(distribution.lower_bound(),0.25);
    BOOST_CHECK_EQUAL(distribution.upper_bound(),0.5);
    BOOST_CHECK_EQUAL(distribution.variate(),0.5);
    BOOST_CHECK_EQUAL(distribution.pdf(),4);
  }
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_constructor_lit,T,test_types)
{
  using graph_model::distribution::uniform::uniform_distribution;
  using graph_model::vertex::lit;

  auto lower_bound = lit(T(0.25));
  auto upper_bound = lit(T(0.50));
  auto distribution = uniform_distribution(lower_bound,upper_bound);
  BOOST_CHECK_EQUAL(distribution.lower_bound(),T(0.25));
  BOOST_CHECK_EQUAL(distribution.upper_bound(),T(0.50));
  BOOST_CHECK_EQUAL(distribution.variate(),T(0.5));
  BOOST_CHECK_EQUAL(distribution.pdf(),T(4));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_constructor_expr,T,test_types)
{
  using graph_model::distribution::uniform::uniform_distribution;
  using graph_model::vertex::lit;

  auto variate = T(0.25);
  auto lower_bound = lit(0) + T(0.25);
  auto upper_bound = lit(0) + T(0.50);
  auto distribution = uniform_distribution(lower_bound,upper_bound,variate);
  BOOST_CHECK_EQUAL(distribution.lower_bound(),T(0.25));
  BOOST_CHECK_EQUAL(distribution.upper_bound(),T(0.50));
  BOOST_CHECK_EQUAL(distribution.variate(),T(0.25));
  BOOST_CHECK_EQUAL(distribution.pdf(),T(4));
}

BOOST_AUTO_TEST_CASE(test_constructor_tagged_parameters)
{
  using std::sqrt;

  using graph_model::distribution::uniform::uniform_distribution;
  using graph_model::distribution::parameter::lower_bound;
  using graph_model::distribution::parameter::upper_bound;
  using graph_model::distribution::parameter::variate;
  using graph_model::vertex::lit;

  auto v = 0.25;
  auto lower = lit(0) + 0.25;
  auto upper = lit(0) + 0.50;

  {
    auto distribution = uniform_distribution(variate=v,upper_bound=upper,lower_bound=lower);
    BOOST_CHECK_EQUAL(distribution.lower_bound(),0.25);
    BOOST_CHECK_EQUAL(distribution.upper_bound(),0.50);
    BOOST_CHECK_EQUAL(distribution.variate(),0.25);
    BOOST_CHECK_EQUAL(distribution.pdf(),4);
  }
  {
    auto distribution = uniform_distribution(lower,v,upper_bound=upper);
    BOOST_CHECK_EQUAL(distribution.lower_bound(),0.25);
    BOOST_CHECK_EQUAL(distribution.upper_bound(),0.5);
    BOOST_CHECK_EQUAL(distribution.variate(),0.25);
    BOOST_CHECK_EQUAL(distribution.pdf(),4);
  }
  {
    auto distribution = uniform_distribution(variate=v,lower);
    BOOST_CHECK_EQUAL(distribution.lower_bound(),0.25);
    BOOST_CHECK_EQUAL(distribution.upper_bound(),1);
    BOOST_CHECK_EQUAL(distribution.variate(),0.25);
    BOOST_CHECK_EQUAL(distribution.pdf(),4.0/3.0);
  }
}

BOOST_AUTO_TEST_CASE(test_constructor_tagged_parameter_pack)
{
  using graph_model::distribution::uniform::uniform_distribution;
  using graph_model::distribution::parameter::lower_bound;
  using graph_model::distribution::parameter::upper_bound;
  using graph_model::distribution::parameter::variate;
  using graph_model::vertex::lit;

  auto v = 0.25;
  auto lower = lit(0) + 0.25;
  auto upper = lit(0) + 0.50;

  {
    auto distribution = uniform_distribution(variate=v,lower_bound=lower,upper_bound=upper);
    BOOST_CHECK_EQUAL(distribution.lower_bound(),0.25);
    BOOST_CHECK_EQUAL(distribution.upper_bound(),0.50);
    BOOST_CHECK_EQUAL(distribution.variate(),0.25);
    BOOST_CHECK_EQUAL(distribution.pdf(),4);
  }
  {
    auto distribution = uniform_distribution(variate=v,lower);
    BOOST_CHECK_EQUAL(distribution.lower_bound(),0.25);
    BOOST_CHECK_EQUAL(distribution.upper_bound(),1);
    BOOST_CHECK_EQUAL(distribution.variate(),0.25);
    BOOST_CHECK_EQUAL(distribution.pdf(),4.0/3.0);
  }
  {
    auto distribution = uniform_distribution(lower_bound=lower,upper,v);
    BOOST_CHECK_EQUAL(distribution.lower_bound(),0.25);
    BOOST_CHECK_EQUAL(distribution.upper_bound(),0.50);
    BOOST_CHECK_EQUAL(distribution.variate(),0.25);
    BOOST_CHECK_EQUAL(distribution.pdf(),4);
  }
  {
    auto distribution = uniform_distribution(lower,upper,v);
    BOOST_CHECK_EQUAL(distribution.lower_bound(),0.25);
    BOOST_CHECK_EQUAL(distribution.upper_bound(),0.50);
    BOOST_CHECK_EQUAL(distribution.variate(),0.25);
    BOOST_CHECK_EQUAL(distribution.pdf(),4);
  }
}
