/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_MODULE graph_model_examples_bernoulli_grid

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <asynchronous/signals/type/default_signal_executor_is_sequential.hpp>

#include <iostream>
#include <map>

#include <boost/filesystem/fstream.hpp>
#include <boost/gil/algorithm.hpp>
#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>
#include <boost/range/empty.hpp>

#include <asynchronous/signals/signal_diagnostics.hpp>

#include <dataframe/dataframe.hpp>
#include <dataframe/element.hpp>
#include <dataframe/read_dataframe.hpp>

#include <graph_model/MCMC/mcmc_signal_policy.hpp>
#include <graph_model/MCMC/metropolis_sampler.hpp>
#include <graph_model/MCMC/sampler_collection.hpp>
#include <graph_model/distribution/bernoulli/bernoulli_distribution.hpp>
#include <graph_model/distribution/beta/beta_distribution.hpp>
#include <graph_model/distribution/normal/normal_distribution.hpp>
#include <graph_model/graph/graphviz.hpp>
#include <graph_model/graph/lattice.hpp>
#include <graph_model/graph_model.hpp>
#include <graph_model/signals/tags.hpp>
#include <graph_model/vertex/expression.hpp>
#include <graph_model/vertex/factor.hpp>
#include <graph_model/vertex/lit.hpp>
#include <graph_model/vertex/markov_blanket.hpp>
#include <graph_model/vertex/mean_factor.hpp>
#include <graph_model/vertex/pdf.hpp>
#include <graph_model/vertex/sample.hpp>
#include <graph_model/vertex/strategy/make_strategy.hpp>
#include <graph_model/vertex/summarized_random_variable.hpp>

#include <identifiers/x.hpp>
#include <mcmc/mcmc_tally.hpp>

#include <logging/initialize.hpp>

#include <boost/mpl/pair.hpp>
#include <boost/mpl/vector.hpp>

// Use a lognumeric type to represent probabilities
//
#if USE_LOGNUMERIC_PROBABILITY_TYPE
#include <sample_space/numeric/lognumeric.hpp>
#endif  // USE_LOGNUMERIC_PROBABILITY_TYPE

template < typename T >
auto X (boost::array<T,1> array)
{ return array[0]; }

template < typename T >
auto Y (boost::array<T,1> array)
{ return 0; }

template < typename T >
auto X (boost::array<T,2> array)
{ return array[0]; }

template < typename T >
auto Y (boost::array<T,2> array)
{ return array[1]; }

namespace boost
{

  // vertex descriptors
  template < typename T, std::size_t N >
  std::ostream& operator << (std::ostream& os, array<T,N> array)
  { return os << "(" << X(array) << "," << Y(array) << ")"; }

} // namespace boost

using test_types = boost::mpl::vector<
  boost::mpl::pair<double,long double>
  >;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_graph_model,Pair,test_types)
{
  using variate_type = typename Pair::first;
#if USE_LOGNUMERIC_PROBABILITY_TYPE
  using probability_type = sample_space::numeric::lognumeric<typename Pair::second>;
#else  // USE_LOGNUMERIC_PROBABILITY_TYPE
  using probability_type = typename Pair::second;
#endif  // USE_LOGNUMERIC_PROBABILITY_TYPE

  using ifstream_type = boost::filesystem::ifstream;
  using ofstream_type = boost::filesystem::ofstream;

  using factor_type = graph_model::vertex::factor<probability_type>;
  using factor_map_type = std::map<int,factor_type>;

  using expression_type = graph_model::vertex::expression<variate_type>;
  using data_map_type = std::map<int,expression_type>;

  using mean_variate_type = graph_model::vertex::mean_factor<variate_type>;
  using mean_probability_type = graph_model::vertex::mean_factor<probability_type>;

  using rv_type = graph_model::vertex::summarized_random_variable<variate_type,probability_type>;

  using iteration = graph_model::signals::tag::iteration;
  using accept = graph_model::signals::tag::accept;
  using reject = graph_model::signals::tag::reject;

  using graph_model::vertex::lit;

  using dataframe::make_dataframe;
  using dataframe::make_element;

  using graph_model::MCMC::make_mcmc_signal_policy;
  using graph_model::MCMC::metropolis_sampler;
  using graph_model::MCMC::make_sampler_collection;
  using graph_model::distribution::bernoulli::bernoulli_distribution;
  using graph_model::distribution::beta::beta_distribution;
  using graph_model::distribution::normal::normal_distribution;
  using graph_model::make_incidence_graph;
  using graph_model::make_vertex_and_edge_list_graph;
  using graph_model::vertex::make_mean_factor;
  using graph_model::vertex::make_summarized_random_variable;
  using graph_model::vertex::pdf;

  using graph_model::vertex::strategy::make_strategy;

  namespace parameter = graph_model::distribution::parameter;
  using graph_model::distribution::parameter::mean;
  using graph_model::distribution::parameter::standard_deviation;
  using graph_model::distribution::parameter::probability;
  using graph_model::distribution::parameter::variate;

  using graph_model::graph::make_lattice;
  using graph_model::graph::make_lattice_vertex_properties;

  using identifiers::x;
  using mcmc::make_mcmc_tally;

  using severity_level = boost::log::trivial::severity_level;
  using logger_type = boost::log::sources::severity_logger<severity_level>;

  namespace accumulators = boost::accumulators;

  logging::initialize();
  logger_type debug_log;

#if GRID_WIDTH > 1 && GRID_HEIGHT > 1
  auto lattice = make_lattice(GRID_WIDTH,GRID_HEIGHT);
  assert(lattice.dimensions() == 2);
  assert(num_vertices(lattice) == lattice.length(0) * lattice.length(1));
  assert(num_edges(lattice) == 2 * (lattice.length(0) * (lattice.length(1) - 1) + lattice.length(1) * (lattice.length(0) - 1)));
  GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "make_lattice: " << lattice.length(0) << "x" << lattice.length(1) << ", num_vertices=" << num_vertices(lattice) << ", num_edges=" << num_edges(lattice);
#elif GRID_WIDTH > 1
  auto lattice = make_lattice(GRID_WIDTH);
  assert(lattice.dimensions() == 1);
  assert(num_vertices(lattice) == lattice.length(0));
  assert(num_edges(lattice) == 2 * (lattice.length(0) - 1));
  GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "make_lattice: " << lattice.length(0) << "x" << 1 << ", num_vertices=" << num_vertices(lattice) << ", num_edges=" << num_edges(lattice);
#elif GRID_HEIGHT > 1
  auto lattice = make_lattice(GRID_HEIGHT);
  assert(lattice.dimensions() == 1);
  assert(num_vertices(lattice) == lattice.length(0));
  assert(num_edges(lattice) == 2 * (lattice.length(0) - 1));
  GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "make_lattice: " << 1 << "x" << lattice.length(0) << ", num_vertices=" << num_vertices(lattice) << ", num_edges=" << num_edges(lattice);
#else
#error invalid grid: GRID_WIDTHxGRID_HEIGHT
#endif

  // construct the MCMC components
  auto mcmc_tally = make_mcmc_tally();
  auto samplers = make_sampler_collection();

  // random variables, factors, and expressions
  auto m = make_summarized_random_variable<variate_type,probability_type>();

  auto precision = make_lattice_vertex_properties<rv_type>(lattice);
  auto p = make_lattice_vertex_properties<rv_type>(lattice);
  auto beta = make_lattice_vertex_properties<factor_type>(lattice);

  auto mean_probability = make_lattice_vertex_properties<mean_probability_type>(lattice);
  auto mean_precision = make_lattice_vertex_properties<mean_variate_type>(lattice);

  data_map_type data;
  factor_map_type bernoulli;

  m
    .reset_identifier("m")
    .reset_variate(0.1);

  {
    GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "constructing lattice ...";
    for (auto vertex_range = vertices(lattice); !boost::empty(vertex_range); ++vertex_range.first)
      {
        auto u = *vertex_range.first;
        GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "vertex " << u << ", out_degree=" << out_degree(u,lattice);

	// XXX - define identifiers and variates
        p[get(boost::vertex_index,lattice,u)].reset_identifier("Bernoulli p ",u);
#if defined(INITIAL_P)
	p[get(boost::vertex_index,lattice,u)].reset_variate(INITIAL_P);
#else  // defined(INITIAL_P)
	p[get(boost::vertex_index,lattice,u)].reset_variate(0.5);
#endif	// defined(INITIAL_P)

        beta[get(boost::vertex_index,lattice,u)]
          .reset_identifier("Beta factor ",u);
        mean_probability[get(boost::vertex_index,lattice,u)]
          .reset_identifier("Beta mean probability ",u);
        mean_precision[get(boost::vertex_index,lattice,u)]
          .reset_identifier("Beta mean precision ",u);
        precision[get(boost::vertex_index,lattice,u)]
          .reset_identifier("Beta background precision ",u)
          .reset_variate(INITIAL_PRECISION);

	// XXX - define graph edges
        p[get(boost::vertex_index,lattice,u)]
          .add_vertex(mean_probability[get(boost::vertex_index,lattice,u)]);
        beta[get(boost::vertex_index,lattice,u)]
          .reset_strategy
          (beta_distribution
           (parameter::variate=p[get(boost::vertex_index,lattice,u)],
            parameter::probability=mean_probability[get(boost::vertex_index,lattice,u)],
            parameter::precision=mean_precision[get(boost::vertex_index,lattice,u)]))
	  .markov_blanket(p[get(boost::vertex_index,lattice,u)]);
        mean_probability[get(boost::vertex_index,lattice,u)]
          .add_factor((lit(1) - m) * p[get(boost::vertex_index,lattice,u)])
          .add_vertex(beta[get(boost::vertex_index,lattice,u)]);
        mean_precision[get(boost::vertex_index,lattice,u)]
          .add_factor((lit(1) - m) * precision[get(boost::vertex_index,lattice,u)])
          .add_vertex(beta[get(boost::vertex_index,lattice,u)]);
        precision[get(boost::vertex_index,lattice,u)]
          .add_vertex(mean_precision[get(boost::vertex_index,lattice,u)]);
        m
          .add_vertex(mean_probability[get(boost::vertex_index,lattice,u)])
          .add_vertex(mean_precision[get(boost::vertex_index,lattice,u)]);
        for (auto adjacency_range = adjacent_vertices(u,lattice); !boost::empty(adjacency_range); ++adjacency_range.first)
          {
            auto v = *adjacency_range.first;
            GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "edge " << u << " -> " << v;
            p[get(boost::vertex_index,lattice,v)]
              .add_vertex(mean_probability[get(boost::vertex_index,lattice,u)]);
            mean_probability[get(boost::vertex_index,lattice,u)]
              .add_factor(m / lit(out_degree(u,lattice)) * p[get(boost::vertex_index,lattice,v)]);
            mean_precision[get(boost::vertex_index,lattice,u)]
              .add_factor(m / lit(out_degree(u,lattice)) * precision[get(boost::vertex_index,lattice,v)]);
            precision[get(boost::vertex_index,lattice,u)]
              .add_vertex(mean_precision[get(boost::vertex_index,lattice,v)]);
          }

	// XXX - define samplers
        GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "vertex " << u << ": samplers";
        samplers.add_sampler
          (metropolis_sampler
	   (markov_blanket(p[get(boost::vertex_index,lattice,u)]),
	    make_mcmc_signal_policy()
	    .template connect_signal<accept>(mcmc_tally.make_acceptor())
	    .template connect_signal<reject>(mcmc_tally.make_rejector()))
           .add_rv(p[get(boost::vertex_index,lattice,u)],
                   lit(p[get(boost::vertex_index,lattice,u)]) + sample(normal_distribution(0,0.1))));
      }
  }

  // construct a dataframe of observations
  ifstream_type is("samples.dat");
  auto dataframe = make_dataframe();
  is >> dataframe;

  // construct random variables for the observations and corresponding factors
  for (auto row : dataframe)
    {
      GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "row " << row;
      using lattice_type = decltype(lattice);
      using graph_traits = boost::graph_traits<lattice_type>;
      using vertex_descriptor = typename graph_traits::vertex_descriptor;
#if GRID_WIDTH > 1 && GRID_HEIGHT > 1
      vertex_descriptor u { row.template at<long>("x").value(), row.template at<long>("y").value() };
#elif GRID_WIDTH > 1
      vertex_descriptor u { row.template at<long>("x").value() };
#elif GRID_HEIGHT > 1
      vertex_descriptor u { row.template at<long>("y").value() };
#else
#error invalid grid: GRID_WIDTHxGRID_HEIGHT
#endif
      data[row]
	.reset_identifier(x(row))
	.reset_strategy(make_element<variate_type>(dataframe,row,"samples"))
	.add_vertex(bernoulli[row]);
      bernoulli[row]
	.reset_identifier("Bernoulli factor ",row)
        .reset_strategy(bernoulli_distribution
			(parameter::variate=data[row],
			 parameter::probability=p[get(boost::vertex_index,lattice,u)]))
	.markov_blanket(p[get(boost::vertex_index,lattice,u)]);
    }

  // convert the implicit incidence graph(s) into a vertex and edge list graph
  auto incidence_graph = make_incidence_graph();
  auto graph = make_vertex_and_edge_list_graph(incidence_graph,m);
  for (auto vertex_range = vertices(lattice); !boost::empty(vertex_range); ++vertex_range.first)
    {
      auto u = *vertex_range.first;
      graph.add_connected_component(incidence_graph,p[get(boost::vertex_index,lattice,u)]);
      graph.add_connected_component(incidence_graph,beta[get(boost::vertex_index,lattice,u)]);
      graph.add_connected_component(incidence_graph,mean_probability[get(boost::vertex_index,lattice,u)]);
      graph.add_connected_component(incidence_graph,mean_precision[get(boost::vertex_index,lattice,u)]);
      graph.add_connected_component(incidence_graph,precision[get(boost::vertex_index,lattice,u)]);
    }
  for (auto row : dataframe)
    graph.add_connected_component(incidence_graph,data[row]);
  std::cout << graph << std::endl;
  graph_model::graph::write_graphviz("graph.dot",graph);

  // construct an output stream for the MCMC samples and generate samples
  ofstream_type ofstream("mcmc.dat");
  ofstream << "i\tx\ty\tMCMC_sample\tmean" << std::endl;
  for (auto vertex_range = vertices(lattice); !boost::empty(vertex_range); ++vertex_range.first)
    {
      auto u = *vertex_range.first;
      ofstream << 0 << "\t" << X(u) << "\t" << Y(u) << "\t"
	       << p[get(boost::vertex_index,lattice,u)].variate() << "\t"
	       << accumulators::mean(p[get(boost::vertex_index,lattice,u)].accumulator())
	       << std::endl;
    }
  GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "MCMC samples:";
  for (int i = 0; i < MCMC_SAMPLE_SIZE; ++i)
    {
      GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "MCMC sample: " << i;
      samplers();
      for (auto vertex_range = vertices(lattice); !boost::empty(vertex_range); ++vertex_range.first)
	{
	  auto u = *vertex_range.first;
	  ofstream << i+1 << "\t" << X(u) << "\t" << Y(u) << "\t"
		   << p[get(boost::vertex_index,lattice,u)].variate() << "\t"
		   << accumulators::mean(p[get(boost::vertex_index,lattice,u)].accumulator())
		   << std::endl;
	}
    }

  // report summary statistics
  // std::cout << "estimated mu:   " << accumulators::mean(accumulator) << std::endl;
  // std::cout << "estimated mean: " << 1 + accumulators::mean(accumulator) << std::endl;
  std::cout << "MCMC tally: " << mcmc_tally << std::endl;
  std::cout << std::endl;
}
