/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_MODULE graph_model_examples_dirichlet_multinomial_sample

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <asynchronous/signals/type/default_signal_executor_is_sequential.hpp>

#include <iostream>
#include <map>

#include <boost/filesystem/fstream.hpp>
#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>

#include <asynchronous/signals/signal_diagnostics.hpp>

#include <dataframe/dataframe.hpp>
#include <dataframe/element.hpp>
#include <dataframe/read_dataframe.hpp>

#include <graph_model/MCMC/mcmc_signal_policy.hpp>
#include <graph_model/MCMC/metropolis_sampler.hpp>
#include <graph_model/accumulators/accumulator_set.hpp>
#include <graph_model/distribution/dirichlet/dirichlet_distribution.hpp>
#include <graph_model/distribution/multinomial/multinomial_distribution.hpp>
#include <graph_model/graph/graphviz.hpp>
#include <graph_model/graph_model.hpp>
#include <graph_model/signals/tags.hpp>
#include <graph_model/vertex/distribution.hpp>
#include <graph_model/vertex/expression.hpp>
#include <graph_model/vertex/factor.hpp>
#include <graph_model/vertex/lit.hpp>
#include <graph_model/vertex/markov_blanket.hpp>
#include <graph_model/vertex/random_variable.hpp>
#include <graph_model/vertex/sample.hpp>

#include <sample_space/vector/simplex.hpp>
#include <sample_space/vector/vector.hpp>

#include <identifiers/x.hpp>
#include <mcmc/mcmc_tally.hpp>

#include <logging/initialize.hpp>

#include <boost/mpl/pair.hpp>
#include <boost/mpl/vector.hpp>

// Use a lognumeric type to represent probabilities
//
#if USE_LOGNUMERIC_PROBABILITY_TYPE
#include <sample_space/numeric/lognumeric.hpp>
#endif	// USE_LOGNUMERIC_PROBABILITY_TYPE

using test_types = boost::mpl::vector<
  boost::mpl::pair<double,double>
  >;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_graph_model,Pair,test_types)
{
  using variate_type = typename Pair::first;
  using probability_type = typename Pair::second;

  using variate_vector_type = sample_space::vector::vector<variate_type>;
  using probability_vector_type = sample_space::vector::vector<probability_type>;

  using ifstream_type = boost::filesystem::ifstream;
  using ofstream_type = boost::filesystem::ofstream;

  using factor_type = graph_model::vertex::factor<probability_type>;
  using factor_map_type = std::map<int,factor_type>;

  using rv_type = graph_model::vertex::random_variable<variate_vector_type,probability_type>;

  using expression_type = variate_vector_type;
  using data_map_type = std::map<int,rv_type>;

  using iteration = graph_model::signals::tag::iteration;
  using accept = graph_model::signals::tag::accept;
  using reject = graph_model::signals::tag::reject;

  using severity_level = boost::log::trivial::severity_level;
  using logger_type = boost::log::sources::severity_logger<severity_level>;

  using graph_model::vertex::lit;

  using dataframe::make_dataframe;
  using dataframe::make_element;

  using graph_model::MCMC::make_mcmc_signal_policy;
  using graph_model::MCMC::metropolis_sampler;
  using graph_model::accumulators::make_accumulator_set;
  using graph_model::distribution::dirichlet::dirichlet_distribution;
  using graph_model::distribution::multinomial::multinomial_distribution;
  using graph_model::make_incidence_graph;
  using graph_model::make_vertex_and_edge_list_graph;
  using graph_model::vertex::make_distribution;
  using graph_model::vertex::make_factor;
  using graph_model::vertex::make_random_variable;

  using namespace graph_model::distribution;
  using graph_model::distribution::parameter::probability;
  using graph_model::distribution::parameter::variate;

  using identifiers::x;
  using mcmc::make_mcmc_tally;

  namespace accumulators = boost::accumulators;

  logging::initialize();
  logger_type debug_log;

  // construct the random variable to estimate
  auto p = make_random_variable<probability_vector_type,probability_type>();
  p.reset_identifier("p");

  // construct prior
  auto alpha = make_random_variable<variate_vector_type,probability_type>();
  auto dirichlet_factor = make_distribution<variate_vector_type,probability_type>();
  alpha
    .reset_identifier("alpha")
    .reset_variate({1,1})
    .distributed_as(dirichlet_factor);
  dirichlet_factor
    .reset_identifier("dirichlet")
    .reset_strategy(dirichlet_distribution(variate=p,parameter::alpha=alpha))
    .markov_blanket(p,alpha);

  // construct a dataframe of observations
  ifstream_type is("samples.dat");
  auto dataframe = make_dataframe();
  is >> dataframe;

  // construct random variables for the observations and corresponding factors
  data_map_type data;
  factor_map_type factor;
  for (auto row : dataframe)
    {
      data[row].reset_identifier(x(row));
      if (make_element<int>(dataframe,row,"samples")() == 0)
	data[row].reset_variate({1,0});
      else
	data[row].reset_variate({0,1});
      factor[row]
	.reset_identifier("multinomial ",row)
	.reset_strategy(multinomial_distribution(variate=data[row],probability=p))
	.markov_blanket(data[row],p);
    }

  // convert the implicit incidence graph(s) into a vertex and edge list graph
  auto incidence_graph = make_incidence_graph();
  auto graph = make_vertex_and_edge_list_graph(incidence_graph,p);
  graph.add_connected_component(incidence_graph,alpha);
  for (auto row : dataframe)
    graph.add_connected_component(incidence_graph,data[row]);
  std::cout << graph << std::endl;
  graph_model::graph::write_graphviz("graph.dot",graph);

  // construct the MCMC components
  auto mcmc_tally = make_mcmc_tally();
  auto p_sampler
    = metropolis_sampler
    (markov_blanket(p),
     make_mcmc_signal_policy()
     .template connect_signal<iteration>
     ([]{ static std::size_t i = 0; if (i == 0) std::cout << "iteration: "; if (i % 100 == 0) std::cout << " " << i << std::flush; ++i; })
     .template connect_signal<accept>(mcmc_tally.make_acceptor())
     .template connect_signal<reject>(mcmc_tally.make_rejector()))
    .add_rv(p,sample(alpha));
  GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "p_sampler.initialize()";
  p_sampler.initialize();
  p_sampler.commit();

  // construct an output stream for the MCMC samples and generate samples
  ofstream_type ofstream("mcmc.dat");
  ofstream << "i\tMCMC_sample\tmean" << std::endl;
  ofstream << 0 << "\t" << p.variate()[1] << "\t" << 0 /* accumulators::mean(p.accumulator()) */ << std::endl;
  GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "MCMC samples:";
  for (int i = 0; i < MCMC_SAMPLE_SIZE; ++i)
    {
      GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "MCMC sample: " << i;
      p_sampler.update();
      p_sampler.commit();
      ofstream << i+1 << "\t" << p.variate()[1] << "\t" << 0 /* accumulators::mean(p.accumulator()) */ << std::endl;
    }
  std::cout << std::endl;

  // report summary statistics
  // std::cout << "estimated p:   " << accumulators::mean(p.accumulator()) << std::endl;
  // std::cout << "estimated mean: " << accumulators::mean(p.accumulator()) << std::endl;
  std::cout << "MCMC tally: " << mcmc_tally << std::endl;
  std::cout << std::endl;
}
