/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_MODULE graph_model_graph_lattice

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <graph_model/graph/lattice.hpp>

#include <boost/mpl/vector.hpp>

struct Copyable {};
struct Moveable
{
  Moveable () = default;
  Moveable (Moveable&&) = default;
  auto& operator = (Moveable&&) { return *this; }
};

using copyable_test_types = boost::mpl::vector<int,long,long long,float,double,long double,Copyable>;
using moveable_test_types = boost::mpl::vector<int,long,long long,float,double,long double,Copyable,Moveable>;

BOOST_AUTO_TEST_CASE(test_make_lattice)
{
  using graph_model::graph::make_lattice;

  int m = 3;
  int n = 3;
  auto lattice = make_lattice(m,n);
  BOOST_CHECK_EQUAL(num_vertices(lattice),m*n);
  BOOST_CHECK_EQUAL(num_edges(lattice),2*(m*(n-1)+n*(m-1)));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_lattice_vertex_properties_copyable_0,T,copyable_test_types)
{
  using graph_model::graph::make_lattice_properties;

  auto property_map = make_lattice_properties<T>();
  BOOST_CHECK_EQUAL(std::distance(property_map.storage_begin(),property_map.storage_end()),0);
  property_map[0] = T();
  BOOST_CHECK_EQUAL(std::distance(property_map.storage_begin(),property_map.storage_end()),1);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_lattice_vertex_properties_copyable_1,T,copyable_test_types)
{
  using graph_model::graph::make_lattice_properties;

  int n = 3;
  auto property_map = make_lattice_properties(n,T());
  BOOST_CHECK_EQUAL(std::distance(property_map.storage_begin(),property_map.storage_end()),n);
  property_map[0] = T();
  BOOST_CHECK_EQUAL(std::distance(property_map.storage_begin(),property_map.storage_end()),n);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_lattice_vertex_properties_copyable_2,T,copyable_test_types)
{
  using graph_model::graph::make_lattice;
  using graph_model::graph::make_lattice_vertex_properties;

  int m = 3;
  int n = 3;
  auto lattice = make_lattice(m,n);
  auto property_map = make_lattice_vertex_properties(lattice,T());
  BOOST_CHECK_EQUAL(std::distance(property_map.storage_begin(),property_map.storage_end()),m*n);
  property_map[0] = T();
  BOOST_CHECK_EQUAL(std::distance(property_map.storage_begin(),property_map.storage_end()),m*n);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_lattice_vertex_properties_moveable_1,T,moveable_test_types)
{
  using graph_model::graph::make_lattice_properties;

  int n = 3;
  auto property_map = make_lattice_properties<T>(n);
  property_map[0] = T();
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_lattice_vertex_properties_moveable_2,T,moveable_test_types)
{
  using graph_model::graph::make_lattice;
  using graph_model::graph::make_lattice_vertex_properties;

  int m = 3;
  int n = 3;
  auto lattice = make_lattice(m,n);
  auto property_map = make_lattice_vertex_properties<T>(lattice);
  property_map[0] = T();
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_lattice_edge_properties_copyable_0,T,copyable_test_types)
{
  using graph_model::graph::make_lattice_properties;

  auto property_map = make_lattice_properties<T>();
  BOOST_CHECK_EQUAL(std::distance(property_map.storage_begin(),property_map.storage_end()),0);
  property_map[0] = T();
  BOOST_CHECK_EQUAL(std::distance(property_map.storage_begin(),property_map.storage_end()),1);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_lattice_edge_properties_copyable_1,T,copyable_test_types)
{
  using graph_model::graph::make_lattice_properties;

  int n = 3;
  auto property_map = make_lattice_properties(n,T());
  BOOST_CHECK_EQUAL(std::distance(property_map.storage_begin(),property_map.storage_end()),n);
  property_map[0] = T();
  BOOST_CHECK_EQUAL(std::distance(property_map.storage_begin(),property_map.storage_end()),n);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_lattice_edge_properties_copyable_2,T,copyable_test_types)
{
  using graph_model::graph::make_lattice;
  using graph_model::graph::make_lattice_edge_properties;

  int m = 3;
  int n = 3;
  auto lattice = make_lattice(m,n);
  auto property_map = make_lattice_edge_properties(lattice,T());
  BOOST_CHECK_EQUAL(std::distance(property_map.storage_begin(),property_map.storage_end()),2*(m*(n-1)+n*(m-1)));
  property_map[0] = T();
  BOOST_CHECK_EQUAL(std::distance(property_map.storage_begin(),property_map.storage_end()),2*(m*(n-1)+n*(m-1)));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_lattice_edge_properties_moveable_1,T,moveable_test_types)
{
  using graph_model::graph::make_lattice_properties;

  int n = 3;
  auto property_map = make_lattice_properties<T>(n);
  property_map[0] = T();
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_lattice_edge_properties_moveable_2,T,moveable_test_types)
{
  using graph_model::graph::make_lattice;
  using graph_model::graph::make_lattice_edge_properties;

  int m = 3;
  int n = 3;
  auto lattice = make_lattice(m,n);
  auto property_map = make_lattice_edge_properties<T>(lattice);
  property_map[0] = T();
}

