/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#define BOOST_TEST_MODULE sample_space_vector_default_vector

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <sample_space/vector/vector.hpp>
#include <sample_space/math/norm_1.hpp>

#include <boost/accumulators/accumulators.hpp>
#include <boost/accumulators/statistics/stats.hpp>
#include <boost/accumulators/statistics/mean.hpp>
#include <boost/mpl/vector.hpp>

using test_types = boost::mpl::vector<char,int,long,unsigned,unsigned long,float,double,long double>;
using signed_test_types = boost::mpl::vector<signed char,int,long,float,double,long double>;
using integer_test_types = boost::mpl::vector<char,int,long,unsigned,unsigned long>;
using float_test_types = boost::mpl::vector<float,double,long double>;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_typedef,T,test_types)
{
  using euclidean_vector_space = sample_space::vector::tag::euclidean_vector_space;
  using euclidean_basis = sample_space::vector::tag::euclidean_basis;
  using array_storage = sample_space::vector::tag::array<3>;
  using vector_storage = sample_space::vector::tag::vector;

  { using vector_type = sample_space::vector::vector<T>; }
  { using vector_type = sample_space::vector::vector<T,void>; }
  { using vector_type = sample_space::vector::vector<T,void,euclidean_vector_space>; }
  { using vector_type = sample_space::vector::vector<T,void,euclidean_vector_space,euclidean_basis>; }
  { using vector_type = sample_space::vector::vector<T,void,use_default<>,euclidean_basis>; }
  { using vector_type = sample_space::vector::vector<T,void,euclidean_vector_space,euclidean_basis,array_storage>; }
  { using vector_type = sample_space::vector::vector<T,void,euclidean_vector_space,euclidean_basis,vector_storage>; }
  { using vector_type = sample_space::vector::vector<T,int,euclidean_vector_space,euclidean_basis,vector_storage>; }
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_default_constructor,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;
  
  vector_type vector;
  BOOST_CHECK(vector.empty());
  BOOST_CHECK_EQUAL(vector.size(),0);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_constructor,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;

  vector_type vector (42);
  BOOST_CHECK(!vector.empty());
  BOOST_CHECK_EQUAL(vector.size(),42);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_constructor_2,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;

  vector_type vector (13,42);
  BOOST_CHECK(!vector.empty());
  BOOST_CHECK_EQUAL(vector.size(),13);
  BOOST_CHECK_EQUAL(*vector.begin(),42);
  BOOST_CHECK(is_possible(vector));
}

BOOST_AUTO_TEST_CASE(test_constructor_3)
{
  using vector_type = sample_space::vector::vector<int>;

  vector_type vector = { 1, 2, 3 };
  BOOST_CHECK(!vector.empty());
  BOOST_CHECK_EQUAL(vector.size(),3);
  BOOST_CHECK_EQUAL(*vector.begin(),1);
  BOOST_CHECK(is_possible(vector));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_copy_constructor,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;

  vector_type v0;
  vector_type v(v0);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_move_constructor,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;

  vector_type v0;
  vector_type v(std::move(v0));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_copy_assignment,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;

  vector_type v0;
  vector_type v1;
  v1 = v0;
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_move_assignment,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;

  vector_type v0;
  vector_type v1;
  v1 = std::move(v0);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_at,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;
  using map_type = sample_space::vector::vector<T,int>;

  {
    vector_type vector (3);
    vector.at(2) = 2;
    vector.at(1) = 1;
    vector.at(0) = 0;
    BOOST_CHECK_EQUAL(vector.at(0),0);
    BOOST_CHECK_EQUAL(vector.at(1),1);
    BOOST_CHECK_EQUAL(vector.at(2),2);
  }

  {
    map_type map;
    map.at(2) = 2;
    map.at(1) = 1;
    map.at(0) = 0;
    BOOST_CHECK_EQUAL(map.at(0),0);
    BOOST_CHECK_EQUAL(map.at(1),1);
    BOOST_CHECK_EQUAL(map.at(2),2);
  }
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_equal,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;

  vector_type vector (13,42);
  BOOST_CHECK_EQUAL(vector,vector);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_less_than,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;

  vector_type v1 (13,42);
  vector_type v2 (14,42);
  BOOST_CHECK(v1 < v2);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_add,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;
  using map_type = sample_space::vector::vector<T,int>;

  {
    vector_type v1 (13,1);
    vector_type v2 (13,2);
    vector_type expected (13,3);
    BOOST_CHECK_EQUAL(v1 + v2, expected);
  }

  {
    map_type m1; m1.at(0) = 1; m1.at(1) = 2;
    map_type m2;               m2.at(1) = 3; m2.at(2) = 3;
    map_type expected; expected.at(0) = 1; expected.at(1) = 5; expected.at(2) = 3;
    BOOST_CHECK_EQUAL(m1 + m2, expected);
  }
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_subtract,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;

  vector_type v1 (13,1);
  vector_type v2 (13,2);
  vector_type expected (13,1);
  BOOST_CHECK_EQUAL(v2 - v1, expected);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_multiply,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;
  using map_type = sample_space::vector::vector<T,int>;

  {
    vector_type v (13,42);
    T x = 2;
    vector_type expected (13,42 * 2);
    BOOST_CHECK_EQUAL(v * x, expected);
    BOOST_CHECK_EQUAL(x * v, expected);
  }
  {
    map_type m; m.at(0) = 1; m.at(1) = 2;
    map_type expected; expected.at(0) = 2; expected.at(1) = 4;
    BOOST_CHECK_EQUAL(2 * m, expected);
  }
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_divide,T,test_types)
{
  using vector_type = sample_space::vector::vector<T>;

  vector_type v (13,42);
  T x = 7;
  vector_type expected (13,42 / 7);
  BOOST_CHECK_EQUAL(v / x, expected);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_modulo,T,integer_test_types)
{
  using vector_type = sample_space::vector::vector<T>;

  vector_type v (13,42);
  T x = 10;
  vector_type expected (13,42 % 10);
  BOOST_CHECK_EQUAL(v % x, expected);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_norm_1,T,signed_test_types)
{
  using vector_type = sample_space::vector::vector<T>;
  using map_type = sample_space::vector::vector<T,int>;

  using std::abs;
  using sample_space::math::norm_1;

  {
    vector_type vector (3);
    vector.at(0) = 0;
    vector.at(1) = 1;
    vector.at(2) = -3;
    BOOST_CHECK_EQUAL(norm_1(vector),abs(T(0))+abs(T(1))+abs(T(3)));
  }

  {
    map_type map;
    map.at(0) = 0;
    map.at(1) = 1;
    map.at(2) = -3;
    BOOST_CHECK_EQUAL(norm_1(map),abs(T(0))+abs(T(1))+abs(T(3)));
  }
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_norm_2,T,signed_test_types)
{
  using vector_type = sample_space::vector::vector<T>;
  using map_type = sample_space::vector::vector<T,int>;

  using std::sqrt;

  {
    vector_type vector (3);
    vector.at(0) = 0;
    vector.at(1) = 1;
    vector.at(2) = -3;
    BOOST_CHECK_EQUAL(norm_2(vector),T(sqrt(T(0)*T(0)+T(1)*T(1)+T(-3)*T(-3))));
  }

  {
    map_type map;
    map.at(0) = 0;
    map.at(1) = 1;
    map.at(2) = -3;
    BOOST_CHECK_EQUAL(norm_2(map),T(sqrt(T(0)*T(0)+T(1)*T(1)+T(-3)*T(-3))));
  }
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_accumulators,T,float_test_types)
{
  using vector_type = sample_space::vector::vector<T>;
  using count_tag = boost::accumulators::tag::count;
  using mean_tag = boost::accumulators::tag::mean;
  using sum_tag = boost::accumulators::tag::sum;
  using stats = boost::accumulators::stats<count_tag,sum_tag,mean_tag>;
  using accumulator_set_type = boost::accumulators::accumulator_set< vector_type,stats >;

  using boost::accumulators::count;
  using boost::accumulators::mean;
  using boost::accumulators::sum;

  vector_type zero = { 0, 0, 0 };
  vector_type one = { 1, 1, 1 };
  vector_type two = { 2, 2, 2 };
  vector_type three = { 3, 3, 3 };
  vector_type six = { 6, 6, 6 };

  accumulator_set_type accumulator_set (zero);
  accumulator_set(one);
  accumulator_set(two);
  accumulator_set(three);
  BOOST_CHECK_EQUAL(count(accumulator_set),3);
  BOOST_CHECK_EQUAL(sum(accumulator_set),six);
  BOOST_CHECK_EQUAL(mean(accumulator_set),two);
}
