/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU General Public
 *     License as published by the Free Software Foundation, either
 *     version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#define BOOST_TEST_MODULE sample_space_vector_dirichlet_distribution

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#include <boost/mpl/vector.hpp>
#endif

template < typename ... T > class XXX;

#include "sample_space/math/dirichlet_distribution.hpp"
#include "sample_space/random/dirichlet_distribution.hpp"
#include "sample_space/vector/vector.hpp"
#include <cmath>
#include <limits>
#include <boost/random/mersenne_twister.hpp>
#include <boost/tuple/tuple.hpp>

template < typename T >
boost::tuple<T,T,T>
generate_expectation (T alpha1, T alpha2, T alpha3)
{
  typedef boost::mt19937 rng_type;
  typedef boost::random::gamma_distribution<T> gamma_type;
  typedef typename gamma_type::param_type param_type;

  rng_type rng;
  gamma_type gamma;
  auto x1 = gamma(rng,param_type(alpha1));
  auto x2 = gamma(rng,param_type(alpha2));
  auto x3 = gamma(rng,param_type(alpha3));
  auto sum = x1 + x2 + x3;
  x1 /= sum;
  x2 /= sum;
  x3 /= sum;
  return boost::make_tuple(x1,x2,x3);
}

template < typename Alpha, typename Variate >
auto pdf (Alpha alpha, Variate variate)
{
  typedef sample_space::math::dirichlet_distribution<Alpha> dirichlet_distribution_type;

  dirichlet_distribution_type dirichlet_distribution (alpha);
  return pdf (dirichlet_distribution,variate);
}

typedef boost::mpl::vector<float,double,long double> test_types;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_dirichlet_distribution_vector,T,test_types)
{
  typedef sample_space::vector::tag::simplex simplex;
  typedef sample_space::vector::vector<T> vector_type;
  typedef sample_space::vector::vector<T,void,simplex> simplex_type;

  typedef sample_space::random::dirichlet_distribution<simplex_type,vector_type> dirichlet_distribution_type;
  typedef boost::mt19937 rng_type;

  using std::abs;
  using boost::get;
  
  vector_type alpha (3);
  alpha.at(2) = 1;
  alpha.at(1) = 1;
  alpha.at(0) = 1;

  dirichlet_distribution_type dirichlet_distribution(alpha);
  rng_type rng;
  auto x = dirichlet_distribution(rng);
  auto expected = generate_expectation(alpha.at(0),alpha.at(1),alpha.at(2));
  
  BOOST_CHECK_EQUAL(alpha.at(0),1);
  BOOST_CHECK_EQUAL(alpha.at(1),1);
  BOOST_CHECK_EQUAL(alpha.at(2),1);
  BOOST_CHECK(is_possible(alpha));

  BOOST_CHECK_EQUAL(x.at(0),get<0>(expected));
  BOOST_CHECK_EQUAL(x.at(1),get<1>(expected));
  BOOST_CHECK_EQUAL(x.at(2),get<2>(expected));
  BOOST_CHECK(abs(x.at(0)+x.at(1)+x.at(2)-1)
	      <= std::numeric_limits<T>::round_error());
  BOOST_CHECK(is_possible(x));

  BOOST_CHECK_EQUAL(pdf(alpha,x),2);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_dirichlet_distribution_map,T,test_types)
{
  typedef sample_space::vector::tag::simplex simplex;
  typedef sample_space::vector::vector<T,int> map_type;
  typedef sample_space::vector::vector<T,int,simplex> simplex_type;

  typedef sample_space::random::dirichlet_distribution<simplex_type,map_type> dirichlet_distribution_type;
  typedef boost::mt19937 rng_type;

  using std::abs;
  using boost::get;
  
  map_type alpha;
  alpha.at(0) = 1;
  alpha.at(1) = 1;
  alpha.at(2) = 1;

  dirichlet_distribution_type dirichlet_distribution(alpha);  
  rng_type rng;
  auto x = dirichlet_distribution(rng);
  auto expected = generate_expectation(alpha.at(0),alpha.at(1),alpha.at(2));
  
  BOOST_CHECK_EQUAL(alpha.at(0),1);
  BOOST_CHECK_EQUAL(alpha.at(1),1);
  BOOST_CHECK_EQUAL(alpha.at(2),1);
  BOOST_CHECK(is_possible(alpha));

  BOOST_CHECK_EQUAL(x.at(0),get<0>(expected));
  BOOST_CHECK_EQUAL(x.at(1),get<1>(expected));
  BOOST_CHECK_EQUAL(x.at(2),get<2>(expected));
  BOOST_CHECK(abs(x.at(0)+x.at(1)+x.at(2)-1)
	      <= std::numeric_limits<T>::round_error());
  BOOST_CHECK(is_possible(x));

  BOOST_CHECK_EQUAL(pdf(alpha,x),2);
}
